/*
Copyright (C)2008-2013, IBM, Inc. All rights reserved.
IBM, Inc. owns the copyright and other intellectual
property rights in this software. Duplication or use of the
Software is not permitted except as expressly provided in a
written agreement between your company and IBM, Inc.
*/

/* See bf_ipv6.h for a full explanation of what this code does. */

/****************************************************************
 * Initialization
 ****************************************************************/

#include "bf_agent.h"
#include "bf_poll.h"
#include <stdarg.h>


#ifdef BF_IPV6_UNIT_TEST
#  undef BFCONF_EXISTS
#  define BFCONF_EXISTS(x) (1)
#endif /* BF_IPV6_UNIT_TEST */

/*
 * To get extensive debugging information from the replacement functions in
 * this file, compile with DEBUG_BF_IPV6 defined.
 */

#undef D
#ifdef DEBUG_BF_IPV6
#  define D(x) x
# else
#  define D(x) /**/
#endif


/* Make sure we can get to h_errno */
#ifndef HAVE_H_ERRNO
extern int h_errno;
#endif /* HAVE_H_ERRNO */




/****************************************************************
 * inet_addr
 ****************************************************************/

#ifndef HAVE_INET_ADDR
/* Honestly, if you don't have this, we're probably missing your whole
 * network stack, but here goes nothing...
 */
in_addr_t bf_inet_addr(const char *cp) {
	struct in_addr ret;
	int ok = inet_aton(cp, &ret);
	D(activity_write(BFHERE "bf_inet_addr: [%s] ok=%d", cp, ok));
	return ok ? ret.s_addr : INADDR_NONE;
}
#endif /* !HAVE_INET_ADDR */



/****************************************************************
 * inet_aton
 ****************************************************************/

#ifndef HAVE_INET_ATON
 /* INADDR_NONE is a valid reponse when the input is "255.255.255.255",
 * but inet_addr uses it to indicate failure.  The inet_aton function
 * was invented to fix this, but it is somewhat new, so you may not
 * have it (Windows doesn't, for example).  You should be using
 * inet_pton instead, anyway.
 */
int bf_inet_aton(const char *cp, struct in_addr *inp)
{
	const char *p = cp;
	static const char *DOTS = "...";
	uint32_t addr = 0;
	unsigned int x;
	int i;

	for (i=0; i<4; ++i) {
		if (!isdigit((int)*p)) {
			D(BFTRACE("bf_inet_aton: octet started with non-digit"));
			return 0;
		}
		x = 0;
		do {
			x = x*10 + *p++ - '0';
			if (x > 255) {
				D(BFTRACE("bf_inet_aton: octet exceeded 255"));
				return 0;
			}
		} while (isdigit((int)*p));
		if (*p++ != DOTS[i]) {
			D(BFTRACE("bf_inet_aton: bad octet termination"));
			return 0;
		}
		addr = (addr << 8) + x;
	}

	D(activity_write(BFHERE "bf_inet_aton: [%s] ok", cp));
	inp->s_addr = htonl(addr);
	return 1;
}
#endif /* !HAVE_INET_ATON */



/****************************************************************
 * inet_ntoa
 ****************************************************************/

#ifndef HAVE_INET_NTOA
/* Straightforward dotted decimal renderer for an IPv4 address.
 * You should use inet_ntop instead.
 */
char *bf_inet_ntoa(struct in_addr in)
{
	static char buf[16];
	uint32_t addr;

	addr = ntohl(in.s_addr);
	snprintf( buf, sizeof(buf), "%d.%d.%d.%d",
			(addr >> 24) & 0xFF,
			(addr >> 16) & 0xFF,
			(addr >> 8) & 0xFF,
			(addr) & 0xFF );
	D(activity_write(BFHERE "bf_inet_ntoa: -> [%s]", buf));
	return buf;
}

#endif /* !HAVE_INET_NTOA */



/****************************************************************
 * gai_strerror
 ****************************************************************/

#ifndef HAVE_GAI_STRERROR
#  undef XEAI
#  define XEAI(err) case err: return #err
/* This implementation just returns stringified versions of the error
 * symbol.
 */
const char *bf_gai_strerror(int err) {
	switch (err) {
		case 0: return "0";
		XEAI(EAI_ADDRFAMILY);
		XEAI(EAI_AGAIN);
		XEAI(EAI_BADFLAGS);
		XEAI(EAI_FAIL);
		XEAI(EAI_FAMILY);
		XEAI(EAI_MEMORY);
		XEAI(EAI_NODATA);
		XEAI(EAI_NONAME);
		XEAI(EAI_OVERFLOW);
		XEAI(EAI_SERVICE);
		XEAI(EAI_SOCKTYPE);
		XEAI(EAI_SYSTEM);
	}
	return "?";
}
#  undef XEAI
#endif /* !HAVE_GAI_STRERROR */



/****************************************************************
 * getnameinfo
 ****************************************************************/

#ifndef HAVE_GETNAMEINFO

/* host half of getnameinfo */
static int bf_gni_host(
		const struct in_addr *addr,
		char *host,
		size_t hostlen,
		int flags)
{
	struct hostent *he = NULL;
	const char *n;

	/* Unless they told us to give them the address as octets, try to find
	 * something better...
	 */
	if (! (flags & NI_NUMERICHOST)) {
		he = gethostbyaddr(addr, sizeof(struct in_addr), AF_INET);
	}

	if (NULL != he) {
		/* Got an actual hostname! */
		n = he->h_name;
	} else if (flags & NI_NAMEREQD) {
		/* Didn't find a hostname, and that's what the caller wants. */
		D(BFTRACE("bf_getnameinfo: host not found"));
		return EAI_NONAME;
	} else {
		/* Falling back on octets */
		n = inet_ntoa(*addr);
	}

	/* Make sure we have room, then save the result. */
	if (strlen(n) >= hostlen) {
		D(BFTRACE("bf_getnameinfo: host buffer overflow"));
		return EAI_OVERFLOW;
	}
	strncpy(host, n, hostlen);

	/* Success! */
	D(activity_write(BFHERE "bf_getnameinfo: host -> [%s]", host));
	return 0;
}


/* service half of getnameinfo */
static int bf_gni_serv(in_port_t port, char *serv, size_t servlen, int flags)
{
	struct servent *se = NULL;
	const char *s;
	char sbuf[16];

	if (! (flags & NI_NUMERICSERV)) {
		se = getservbyport(port, (flags & NI_DGRAM) ? "udp" : "tcp");
	}

	if (NULL != se) {
		s = se->s_name;
	} else {
		snprintf(sbuf, sizeof(sbuf), "%d", ntohs(port));
		s = sbuf;
	}

	if (strlen(s) >= servlen) {
		D(BFTRACE("bf_getnameinfo: service buffer overflow"));
		return EAI_OVERFLOW;
	}
	strncpy(serv, s, servlen);
	D(activity_write(BFHERE "bf_getnameinfo: service -> [%s]", serv));
	return 0;
}


int bf_getnameinfo(
		const struct sockaddr *sa, socklen_t salen,
		char *host, size_t hostlen,
		char *serv, size_t servlen,
		int flags)
{
	const struct sockaddr_in *sin = (const struct sockaddr_in *)sa;
	const char *n, *s;
	char sbuf[8];
	int saved_errno = errno;
	int saved_h_errno = h_errno;
	int ret = 0;

	/* You have to request something... */
	if (NULL == host && NULL == serv) { return EAI_NONAME; }

	/* We only know how to handle sockets with AF_INET addresses, and
	 * the length of the sock address better match what we expect out
	 * of AF_INET, too.
	 */
	if (AF_INET != sin->sin_family) { return EAI_FAMILY; }
	if (sizeof(struct sockaddr_in) != salen) { return EAI_FAMILY; }

	/* Try to resolve the address first. */
	if (NULL != host) {
		ret = bf_gni_host(&sin->sin_addr, host, hostlen, flags);
	}

	/* If that was okay, then try to resolve the service.  If we
	 * can't, that's fine; we'll just use the port number.
	 */
	if (NULL != serv && 0 == ret) {
		ret = bf_gni_serv(sin->sin_port, serv, servlen, flags);
	}

	D(activity_write( BFHERE "bf_getnameinfo: ret=%d [%s]",
				ret, gai_strerror(ret) ));
	errno = saved_errno;
	h_errno = saved_h_errno;
	return ret;
}

#endif /* !HAVE_GETNAMEINFO */



/****************************************************************
 * getaddrinfo
 ****************************************************************/

#ifndef HAVE_GETADDRINFO

/* service half of getaddrinfo */
static int bf_gai_serv(
		const char *service,
		const struct bf_addrinfo *hints,
		in_port_t *port)
{
	const char *p = service;
	unsigned int x = 0;

	/* Try to interpret the service as a port number */
	while (isdigit((int)(*p))) {
		x = x*10 + *p++ - '0';
		if (x > 65535) {
			D(BFTRACE("bf_getaddrinfo: service exceeded 65535"));
			return EAI_SERVICE;
		}
	}

	if ('\0' == *p) {
		/* Valid port number. */
		*port = htons((uint16_t)x);
	} else if (hints->ai_flags & AI_NUMERICSERV) {
		/* Not a port number, but AI_NUMERICSERV was given. */
		D(activity_write( BFHERE
				"bf_getaddrinfo: AI_NUMERICSERV, but service=[%s]",
				service ));
		return EAI_NONAME;
	} else {
		/* Not a port number; try to look up service. */
		const char *proto = (hints->ai_socktype == SOCK_DGRAM) ? "udp" : "tcp";
		struct servent *se = getservbyname(service, proto);
		if (NULL == se) {
			/* No such service */
			D(activity_write( BFHERE
					"bf_getaddrinfo: service not found: [%s]",
					service ));
			return EAI_SERVICE;
		}
		*port = se->s_port;
	}

	D(activity_write(BFHERE
			"bf_getaddrinfo: service=[%d]",
			(int)ntohs(*port) ));
	return 0;
}


/* host half of getaddrinfo */
static int bf_gai_node(
		const char *node,
		in_port_t port,
		const struct bf_addrinfo *hints,
		struct bf_addrinfo **res)
{
	const char *cname = NULL;
	char *cname_copy = NULL;
	struct hostent *he;
	struct in_addr iaddr;
	struct in_addr *iaddr_ptr = &iaddr;
	struct in_addr **iaddrs = &iaddr_ptr;
	struct sockaddr_in *sa, *sa_head = NULL;
	struct bf_addrinfo *ai, *ai_head = NULL;
	size_t count = 1;
	int i;

	if (NULL == node) {
		/* No node name given.  Use 0.0.0.0 or 127.0.0.1 depending on whether
		 * the request indicated that it intends to use this for a passive
		 * connection or not.
		 */
		if (hints->ai_flags & AI_PASSIVE) {
			D(BFTRACE("bf_getaddrinfo: NULL + AI_PASSIVE -> INADDR_ANY"));
			iaddr.s_addr = htonl(INADDR_ANY);
		} else {
			D(BFTRACE("bf_getaddrinfo: NULL -> INADDR_LOOPBACK"));
			iaddr.s_addr = htonl(INADDR_LOOPBACK);
		}
	} else if (inet_aton(node, &iaddr)) {
		/* The node name was a valid IPv4 address.  Try to find its CNAME, if
		 * the caller requested it.
		 */
		D(activity_write( BFHERE
				"bf_getaddrinfo: resolved as octets: node=[%s]",
				node ));
		if ( hints->ai_flags & AI_CANONNAME &&
				!(hints->ai_flags & AI_NUMERICHOST) )
		{
			he = gethostbyaddr(&iaddr, sizeof(iaddr), AF_INET);
			cname = (NULL != he) ? he->h_name : node;
			D(activity_write(BFHERE "bf_getaddrinfo: cname -> [%s]", cname));
		}
	} else if (hints->ai_flags & AI_NUMERICHOST) {
		/* The node name was given; it is not a valid IPv4 address; but
		 * the caller hinted that it was.
		 */
		D(activity_write(BFHERE
					"bf_getaddrinfo: AI_NUMERICHOST, but node=[%s]",
					node));
		return EAI_NONAME;
	} else {
		/* Do a host lookup on the node name and use the address list
		 * that gets returned.  Also grab the CNAME, if one was requested.
		 */
		h_errno = 0;
		he = gethostbyname(node);
		if (NULL == he) {
			/* Host lookup failed. */
			D(BFTRACE("bf_getaddrinfo: host lookup failed"));
			switch (h_errno) {
				case HOST_NOT_FOUND:
					return EAI_NONAME;
#ifdef NO_ADDRESS
#  if NO_ADDRESS != NO_DATA
				case NO_ADDRESS:
#  endif /* NO_ADDRESS != NO_DATA */
#endif /* NO_ADDRESS */
				case NO_DATA:
					return EAI_NODATA;
				case TRY_AGAIN:
					return EAI_AGAIN;
				default:
					return EAI_FAIL;
			}
		}

		/* Make sure it really did give us a valid result, and note
		 * the CNAME if there is one.
		 */
		iaddrs = (struct in_addr **)he->h_addr_list;
		if (NULL == he->h_addr_list[0]) {
			/* PARANOID */
			D(BFTRACE("bf_getaddrinfo: gethostbyname returned empty list"));
			return EAI_NODATA;
		}
		while (NULL != he->h_addr_list[count]) { ++count; }
		if (hints->ai_flags & AI_CANONNAME) { cname = he->h_name; }
		D(activity_write( BFHERE
				"bf_getaddrinfo: count=%d, cname=[%s]",
				(int)count,
				(NULL != cname) ? cname : "NULL" ));
	}

	/* At this point, we have
	 * - iaddrs:  array of (struct in_addr *) items
	 * - count :  number of entries in iaddrs
	 * - cname :  the canonical name, or NULL
	 *
	 * To reduce the work for alloc, we grab the memory for all items of
	 * the same type at once.  That is, we allocate a single array and
	 * link together the elements of that array as opposed to building
	 * a linked list of individually allocated items.
	 */
	ai_head = calloc(count, sizeof(struct addrinfo));
	if (NULL == ai_head) { return EAI_MEMORY; }
	sa_head = calloc(count, sizeof(struct sockaddr_in));
	if (NULL == sa_head) {
		free(ai_head);
		return EAI_MEMORY;
	}
	if (NULL != cname) {
		cname_copy = strdup(cname);
		if (NULL == cname_copy) {
			free(ai_head);
			free(sa_head);
			return EAI_MEMORY;
		}
	}

	/* Now that we've successfully allocated all the memory that we need,
	 * populate the values to return to the caller.
	 */
	ai = ai_head;
	sa = sa_head;
	for (i=0; i<count; ++i, ++ai, ++sa) {
		ai->ai_family = PF_INET;
		ai->ai_socktype = hints->ai_socktype;
		ai->ai_protocol = hints->ai_protocol;
		ai->ai_addr = (struct sockaddr *)sa;
		ai->ai_addrlen = sizeof(struct sockaddr_in);
		ai->ai_next = &ai[1];

		sa->sin_family = AF_INET;
		sa->sin_port = port;
		memcpy(&sa->sin_addr, iaddrs[i], sizeof(struct in_addr));
	}

	/* Special case for the first entry */
	ai_head[0].ai_canonname = cname_copy;

	/* Special case for the last entry */
	ai_head[count-1].ai_next = NULL;

	/* Success! */
	*res = ai_head;
	return 0;
}

/* The settings we use if NULL is passed as getaddrinfo's hints param. */
const struct bf_addrinfo DEFAULT_HINTS = {
	0,          /* ai_flags */
	PF_UNSPEC,  /* ai_family */
	0,          /* ai_socktype */
	0,          /* ai_protocol */
	0,          /* ai_addrlen */
	NULL,       /* ai_addr */
	NULL,       /* ai_canonname */
	NULL        /* ai_next */
};

int bf_getaddrinfo(
		const char *node,
		const char *service,
		const struct bf_addrinfo *hints,
		struct bf_addrinfo **res)
{
	int saved_errno = errno;
	int saved_h_errno = h_errno;
	int ret = 0;
	struct sockaddr_in sa;
	in_port_t port;

	/* Treat empty strings the same way we treat NULL */
	if (NULL != node && '\0' == *node) { node = NULL; }
	if (NULL != service && '\0' == *service) { service = NULL; }

	/* The caller needs to actually ask for something... */
	if (NULL == node && NULL == service) {
		D(BFTRACE("bf_getaddrinfo: node and service were both empty/NULL"));
		return EAI_NONAME;
	}

	/* Make sure the hints are sane. */
	if (NULL != hints) {
		/* Since we're providing glue to IPv4 calls, only INET and UNSPEC
		 * can be provided.
		 */
		if (PF_INET != hints->ai_family && PF_UNSPEC != hints->ai_family) {
			return EAI_FAMILY;
		}

		/* Socket type has to be stream (TCP), datagram (UDP), or unspecified
		 * (for which we'll try TCP).
		 */
		switch (hints->ai_socktype) {
			case 0:
			case SOCK_STREAM:
			case SOCK_DGRAM:
				break;
			default:
				D(BFTRACE("bf_getaddrinfo: unsupported socket type"));
				return EAI_SOCKTYPE;
		}
	} else {
		D(BFTRACE("bf_getaddrinfo: using default hints"));
		hints = &DEFAULT_HINTS;
	}

	/* Resolve the service port number */
	port = htons(0);
	if (NULL != service) {
		ret = bf_gai_serv(service, hints, &port);
	}

	/* Resolve the socket address.  NULL is ok. */
	if (0 == ret) {
		ret = bf_gai_node(node, port, hints, res);
	}

	D(activity_write( BFHERE "bf_getaddrinfo: ret=%d [%s]",
				ret, gai_strerror(ret) ));
	errno = saved_errno;
	h_errno = saved_h_errno;
	return ret;
}


void bf_freeaddrinfo(struct bf_addrinfo *res) {
	if (NULL != res) {
		if (NULL != res->ai_canonname) { free(res->ai_canonname); }
		free(res->ai_addr);
		free(res);
	}
}

#endif /* HAVE_GETADDRINFO */


static int set_blocking(SOCKET sock, int blocking) {
#ifdef WIN32
	u_long flags;
# else /* !WIN32 */
	int flags;
#endif /* WIN32/! */

#if defined(O_NONBLOCK) && !defined(WIN32)
	flags = fcntl(sock, F_GETFL, 0);
	if (-1 == flags) { flags = 0; }
	if (blocking) {
		flags &= ~O_NONBLOCK;
	} else {
		flags |= O_NONBLOCK;
	}
	return fcntl(sock, F_SETFL, flags);
#else /* WIN32 || !O_NONBLOCK */
	flags = blocking ? 0 : 1;
	return platform_ioctlsocket(sock, FIONBIO, &flags);
#endif
}

static bf_listener *bf_listener_alloc(const char *bind_addr) {
	bf_listener *listener = calloc(1, sizeof(bf_listener));
	if (NULL == listener) { goto out_of_memory; }
	if (NULL == bind_addr) { bind_addr = "*"; }
	listener->addr = strdup((NULL != bind_addr) ? bind_addr : "*");
	if (NULL == listener->addr) { goto out_of_memory; }
	listener->poller = bf_poll_new();
	if (NULL == listener->poller) { goto out_of_memory; }
	return listener;

out_of_memory:
	bf_listener_destroy(listener);
	errno = ENOMEM;
	return NULL;
}

static int bf_gai(
		const char *host,
		const char *serv,
		int flags,
		struct addrinfo **ptr_ai_head)
{
	struct addrinfo hints;
	int ret;

	memset(&hints, 0, sizeof(hints));
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = flags;
	if (NULL != serv) {
		hints.ai_flags |= AI_NUMERICSERV;
	}
	if (BFCONF_EXISTS(getaddrinfo_use_addrconfig)) {
		hints.ai_flags |= AI_ADDRCONFIG;
	}

	ret = getaddrinfo(host, serv, &hints, ptr_ai_head);
	return (0 != ret || NULL != *ptr_ai_head) ? ret : EAI_NONAME;
}


bf_listener *bf_listener_new(const char *bind_addr, int bind_port, int backlog)
{
	struct sockaddr_storage name;
	struct addrinfo *ai_head = NULL;
	struct addrinfo *ai;
	socklen_t slen;
	SOCKET sock = INVALID_SOCKET;
	int ok = 0;
	int ret = ENOENT;
	char addr[256];
	char bind_serv[32];
	char serv[32];
	char *p = NULL;
	bf_listener *listener;

	listener = bf_listener_alloc(bind_addr);
	if (NULL == listener) {
		BFTRACE("bf_listener_new: out of memory");
		return NULL;
	}

	if (bind_port != 0) {
		if (bind_port < 0 || bind_port > 65535) {
			activity_write(BFHERE "bf_listener_new: bad port: %d", bind_port);
			errno = EINVAL;
			return NULL;
		}
		snprintf(bind_serv, sizeof(bind_serv), "%d", bind_port);
		p = bind_serv;
	} else {
		strncpy(bind_serv, "*", 2);
	}

	listener->port = bind_port;
	ret = bf_gai(bind_addr, p, AI_PASSIVE, &ai_head);
	if (NULL == bind_addr) { bind_addr = "*"; }
	if (0 != ret) {
		activity_write(BFHERE "bf_listener_new: [%s/%s] getaddrinfo: %d: %s",
				bind_addr, bind_serv, ret, gai_strerror(ret));
		bf_listener_destroy(listener);
		errno = ENOENT;
		return NULL;
	}

	for (ai = ai_head; NULL != ai; ai = ai->ai_next) {
		sock = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
		if (INVALID_SOCKET == sock) {
			ret = errno;
			activity_write(BFHERE "bf_listener_new: [%s/%s] socket(%d,%d,%d): %d: %s",
					bind_addr, bind_serv,
					ai->ai_family, ai->ai_socktype, ai->ai_protocol,
					ret, strerror(ret));
			continue;
		}

#ifdef WIN32
		{
			BOOL yes = TRUE;
			setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (const char *)&yes, sizeof(yes));
		}
#elif defined(SO_REUSEADDR)
		{
			int yes = 1;
			setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes));
		}
#endif

		/* Attempt to bind the socket.  We don't consider it an error if
		 * the bind fails with an address in use message if we have already
		 * succeeded in binding with a different address.
		 */
		if (0 != bind(sock, ai->ai_addr, ai->ai_addrlen)) {
			ret = GetNetworkError();
			if (!IsNetworkErrorAddressInUse() || !ok) {
				activity_write(BFHERE "bf_listener_new: [%s/%s] bind: %d: %s",
						bind_addr, bind_serv, ret, strerror(ret));
			}
			platform_closesocket(sock);
			continue;
		}

		/* Find out what address we got */
		slen = (socklen_t)(sizeof(name));
		if (0 != getsockname(sock, (struct sockaddr *)&name, &slen)) {
			ret = errno;
			activity_write(BFHERE "bf_listener_new: [%s/%s] getsockname: %d: %s",
					bind_addr, bind_serv, ret, strerror(ret));
			platform_closesocket(sock);
			continue;
		}

		ret = getnameinfo((struct sockaddr *)&name, slen,
					addr, sizeof(addr),
					serv, sizeof(serv),
					NI_NUMERICHOST | NI_NUMERICSERV);
		if (0 != ret) {
			D(activity_write(BFHERE "[%s/%s] getnameinfo: %d: %s",
					bind_addr, bind_serv, ret, gai_strerror(ret)));
			platform_closesocket(sock);
			ret = ENOENT;
			continue;
		}

		/* If we are using an ephemeral port, then we need to note which one
		 * it gave us and make sure that we get the same port on every
		 * interface.
		 */
		if (0 == bind_port) {
			ret = atoi(serv);
			if (0 == listener->port) {
				listener->port = ret;
			} else if (ret != listener->port) {
				ret = EINVAL;
				D(activity_write(BFHERE "[%s/*] bind -> %d != expected %d; discarding...",
						bind_addr, ret, listener->port));
				platform_closesocket(sock);
				continue;
			}
		}
		ok = 1;
		D(activity_write(BFHERE "bf_listener_new: [%s/%s] bind -> %s/%s", bind_addr, bind_serv, addr, serv));

		/* Listen for connections */
		if (backlog <= 0 || backlog > SOMAXCONN) { backlog = SOMAXCONN; }
		if (0 != listen(sock, backlog)) {
			ret = errno;
			D(activity_write(BFHERE "[%s/%s] listen: %d: %s",
					addr, serv, ret, strerror(ret)));
			platform_closesocket(sock);
			continue;
		}
		activity_write(BFHERE "bf_listener_new: [%s/%s] listening", addr, serv);

		/* Add this socket to the listener's inventory */
		if (0 == bf_poll_add(listener->poller, sock, BF_POLL_ACCEPT)) {
			ret = ENOMEM;
			activity_write(BFHERE "bf_listener_new: [%s/%s] listen: out of memory", addr, serv);
			platform_closesocket(sock);
			continue;
		}
	}
	freeaddrinfo(ai_head);

	if (listener->poller->len <= 0) {
		bf_listener_destroy(listener);
		errno = (0 != ret) ? ret : ENOENT;
		return NULL;
	}

	D(activity_write(BFHERE "bf_listener_new: listening on socket %d", (int)sock));
	return listener;
}

static bf_poll_action close_socket_cb(
		bf_poll *poller,
		SOCKET sock,
		int readyOps,
		void *arg)
{
	platform_closesocket(sock);
	return BF_POLL_ACTION_CONTINUE;
}

void bf_listener_destroy(bf_listener *listener) {
	if (NULL == listener) { return; }
	if (NULL != listener->addr) {
		free(listener->addr);
	}
	if (NULL != listener->poller) {
		bf_poll_iterate_all(listener->poller, &close_socket_cb, NULL);
		bf_poll_destroy(listener->poller);
	}
	free(listener);
}


typedef struct _bf_listener_accept_state {
	bf_listener *listener;
	struct sockaddr_storage peer;
	socklen_t peerlen;
	SOCKET sock;
} bf_listener_accept_state;

static void accept_trace_ok(bf_listener_accept_state *state, SOCKET sock) {
	char peer_host[256];
	char peer_serv[32];
	int ret;

	ret = getnameinfo((struct sockaddr *)&state->peer, state->peerlen,
				peer_host, sizeof(peer_host),
				peer_serv, sizeof(peer_serv),
				NI_NUMERICHOST | NI_NUMERICSERV);
	if (0 != ret) {
		D(activity_write(BFHERE "[%s/%d] getnameinfo: %d: %s",
				state->listener->addr, state->listener->port,
				ret, gai_strerror(ret)));
		return;
	}

	D(activity_write(BFHERE "[%s/%d] accept: %s/%s",
				state->listener->addr, state->listener->port,
				peer_host, peer_serv));
}

static void accept_trace_error(bf_listener_accept_state *state, int err) {
	D(activity_write(BFHERE "[%d] accept: %d: %s", state->listener->port, err, strerror(err)));
}

static bf_poll_action bf_listener_accept_cb(bf_poll *poller, SOCKET sock, int readyOps, void *arg)
{
	int err;
	bf_listener_accept_state *state = arg;

	if (!BF_POLL_READY(BF_POLL_ACCEPT, readyOps)) { return BF_POLL_ACTION_CONTINUE; }
	state->peerlen = sizeof(state->peer);

	state->sock = accept(sock, (struct sockaddr *)&state->peer, &state->peerlen);


#ifdef ZOS
		if (state->sock == -1) {
			err = GetNetworkError();
			activity_write(BFHERE "[%s/%d] accept Error: %d: %s",state->listener->addr, state->listener->port, err, strerror(err));
			if (122 ==  err ){
			   BFTRACE(" accept Error:122, will exit. Pls check the network status, then start bfagent again.");
			   exit(-1);
			}
	}		
#endif

	if (INVALID_SOCKET != state->sock) {
		if (NULL != activity_file) { accept_trace_ok(state, sock); }
		return BF_POLL_ACTION_BREAK;
	}

	err = GetNetworkError();
	if (NULL != activity_file) { accept_trace_error(state, err); }
	platform_closesocket(sock);
	bf_poll_cancel(poller, sock);
	SetNetworkError(err);
	return BF_POLL_ACTION_ABORT;
}

SOCKET bf_listener_accept(bf_listener *listener, int timeout) {
	int ret;
	bf_listener_accept_state state;

	ret = bf_poll_wait(listener->poller, timeout);
	if (ret <= 0) {
		if (0 == ret) { SetNetworkErrorTimedOut(); }
		return INVALID_SOCKET;
	}

	state.listener = listener;
	ret = bf_poll_iterate_ready(listener->poller, &bf_listener_accept_cb, &state);
	if (ret <= 0) {
		if (0 == ret) { SetNetworkErrorTimedOut(); }
		return INVALID_SOCKET;
	}

	D(activity_write(BFHERE "bf_listener_accept: accepted on socket %d", (int)state.sock));
	return state.sock;
}

typedef struct _bf_socket_connect_state {
	const char *host;
	int port;
	bf_poll *poller;
	SOCKET connected;
	int socks;
	int ret;
} bf_socket_connect_state;

static bf_poll_action bf_socket_connect_finish(bf_poll *poller, SOCKET sock, int readyOps, void *arg) {
	bf_socket_connect_state *state = arg;
	struct sockaddr_storage peer;
	char host[128], serv[16];
	socklen_t peerlen;
	char dummy;
	int err;

	if (BF_POLL_READY(BF_POLL_ERROR, readyOps)) {
		D(BFTRACE("bf_socket_connect_finish: error on a socket; cancelling..."));
		bf_poll_cancel(poller, sock);
		return BF_POLL_ACTION_CONTINUE;
	}
	if (!BF_POLL_READY(BF_POLL_CONNECT, readyOps)) {
		return BF_POLL_ACTION_CONTINUE;
	}

	bf_poll_cancel(poller, sock);
	peerlen = sizeof(peer);
	if (0 != getpeername(sock, (struct sockaddr *)&peer, &peerlen)) {
		err = GetNetworkError();
		if (IsNetworkErrorNotConnected() && recv(sock, &dummy, 1, 0) < 0) {
			err = GetNetworkError();
			D(activity_write(BFHERE "bf_socket_connect_finish: %d: %s", err, strerror(err)));
		} else {
			D(activity_write(BFHERE "bf_socket_connect_finish: getpeername: %d: %s", err, strerror(err)));
		}
		state->ret = err;
		return BF_POLL_ACTION_CONTINUE;
	}

	err = getnameinfo(
			(struct sockaddr *)&peer, peerlen,
			host, sizeof(host),
			serv, sizeof(serv),
			NI_NUMERICHOST | NI_NUMERICSERV);
	if (0 != err) {
		D(activity_write(BFHERE "bf_socket_connect_finish: getnameinfo: %d: %s", err, gai_strerror(err)));
		SetNetworkErrorHostNotFound();
		state->ret = GetNetworkError();
		return BF_POLL_ACTION_CONTINUE;
	}

	D(activity_write(BFHERE "bf_socket_connect_finish: connected to %s/%s", host, serv));
	state->connected = sock;
	return BF_POLL_ACTION_BREAK;
}


static SOCKET bf_socket_connect_start(bf_socket_connect_state *state, struct addrinfo *ai) {
	int sock = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);

	if (INVALID_SOCKET == sock) {
		state->ret = errno;
		D(activity_write(BFHERE "[%s/%d] socket(%d,%d,%d): %d: %s",
				state->host, state->port,
				ai->ai_family, ai->ai_socktype, ai->ai_protocol,
				state->ret, strerror(state->ret)));
		return 0;
	}

	set_blocking(sock, 0);
	if (0 == connect(sock, ai->ai_addr, ai->ai_addrlen)) {
		/* Immediate connection even though we are non-blocking.  This
		 * probably means that the specified host is a loopback device.
		 */
		state->connected = sock;
		return 1;
	}

	if (IsNetworkErrorInProgress()) {
		if (0 != bf_poll_add(state->poller, sock, BF_POLL_CONNECT)) { return 0; }
		state->ret = ENOMEM;
		activity_write(BFHERE "[%s/%d] bf_poll_add: out of memory",
				state->host, state->port);
	} else {
		state->ret = GetNetworkError();
		activity_write(BFHERE "[%s/%d] connect: %d: %s",
				state->host, state->port, state->ret, strerror(state->ret));
	}

	platform_closesocket(sock);
	return 0;
}

SOCKET bf_socket_connect(const char *host, int port, int timeout) {
	struct addrinfo *ai_head = NULL;
	struct addrinfo *ai;
	char serv[32];
	bf_socket_connect_state state;
	time_t deadline = time(NULL) + timeout;
	int first_only = BFCONF_EXISTS(ipv6_connect_first_addr_only);

	if (NULL == host || port < 1 || port > 65535) {
		errno = EINVAL;
		return INVALID_SOCKET;
	}

	state.poller = bf_poll_new();
	if (NULL == state.poller) {
		activity_write(BFHERE "connect: %s/%d: out of memory", host, port);
		errno = ENOMEM;
		return INVALID_SOCKET;
	}

	/* resolve hostname and port */
	snprintf(serv, sizeof(serv), "%d", port);
	state.ret = bf_gai(host, serv, AI_NUMERICSERV, &ai_head);
	if (0 != state.ret) {
		activity_write(BFHERE "[%s/%d] getaddrinfo: %d: %s",
				host, port, state.ret, gai_strerror(state.ret));
		bf_poll_destroy(state.poller);
		errno = ENOENT;
		return INVALID_SOCKET;
	}

	state.host = host;
	state.port = port;
	state.connected = INVALID_SOCKET;
	state.ret = ENOENT;
	for (ai = ai_head; NULL != ai; ai = ai->ai_next)
	{
		if (bf_socket_connect_start(&state, ai) || first_only) { break; }
	}
	freeaddrinfo(ai_head);

	if (INVALID_SOCKET == state.connected) {
		while (bf_poll_count(state.poller) > 0) {
			D(BFTRACE("bf_socket_connect: waiting for completion events..."));
			state.ret = bf_poll_wait(state.poller, timeout);
			D(activity_write(BFHERE "bf_socket_connect: bf_poll_wait returned %d", state.ret));
			if (state.ret <= 0) { break; }
			bf_poll_iterate_ready(state.poller, &bf_socket_connect_finish, &state);
			if (INVALID_SOCKET != state.connected) { break; }
			timeout = (int)(deadline - time(NULL));
			if (timeout <= 0) {
				SetNetworkErrorTimedOut();
				state.ret = GetNetworkError();
				break;
			}
		}
	}

	bf_poll_iterate_all(state.poller, &close_socket_cb, NULL);
	bf_poll_destroy(state.poller);

	if (INVALID_SOCKET != state.connected) {
		set_blocking(state.connected, 1);
		D(activity_write(BFHERE "bf_socket_connect: connected on socket %d", (int)state.connected));
	} else if (0 != state.ret) {
		SetNetworkError(state.ret);
	} else {
		SetNetworkErrorHostNotFound();
	}
	return state.connected;
}

#ifndef ZOS
int bf_socket_fill(SOCKET sock, void *buf, int min, int max, int timeout) {
	char *p = buf;
	time_t deadline = time(NULL) + timeout;
	int total = 0;
	int ret;

	for (;;) {
		if (!bf_poll_can_read(sock, timeout)) {
			D(BFTRACE("bf_socket_fill: bf_poll_can_read returned 0; returning -1"));
			return -1;
		}

		ret = (int)recv(sock, p, max - total, 0);
		if (ret <= 0) {
			if (0 == ret) {
				BFTRACE("bf_socket_fill: premature close");
				SetNetworkErrorConnectionReset();
			}
			return -1;
		}
		D(activity_write(BFHERE "bf_socket_fill: recv: %d", ret));

		total += ret;
		if (total >= min) { return total; }

		timeout = (int)(deadline - time(NULL));
		if (timeout <= 0) {
			SetNetworkErrorTimedOut();
			return -1;
		}

		p = &p[ret];
	}
}
#endif

#ifndef ZOS
int bf_socket_recv(SOCKET sock, void *buf, int sz, int timeout) {
	char *p = buf;
	int ret;
	if (!bf_poll_can_read(sock, timeout)) {
		D(BFTRACE("bf_socket_recv: bf_poll_can_read returned 0; returning -1"));
		ret = -1;
	} else {
		ret = (int)recv(sock, p, sz, 0);
		D(activity_write(BFHERE "bf_socket_recv: recv: %d", ret));
	}
	return ret;
}
#endif

int bf_socket_send(SOCKET sock, const void *buf, int sz, int timeout) {
	const char *p = buf;
	int wrote;
	int left = sz;
	int err = 0;
	int ret = -1;

#ifndef WIN32
	/* SIGPIPE is annoying... Let's get an errno instead of dying, thanks... */
	signal(SIGPIPE, SIG_IGN);
#endif /* !WIN32 */

	if (sz <= 0) {
		D(BFTRACE("bf_socket_send: sz <= 0"));
		if (sz == 0) { return 0; }
		SetNetworkErrorInvalidArgument();
		err = GetNetworkError();
		goto done;
	}
	while (left > 0) {
#ifndef ZOS
		if (!bf_poll_can_write(sock, timeout)) {
			D(BFTRACE("bf_socket_send: bf_poll_can_write returned 0; returning -1"));
			SetNetworkErrorTimedOut();
			err = GetNetworkError();
			goto done;
		}
#endif
		wrote = (int)send(sock, p, left, 0);
		D(activity_write(BFHERE "bf_socket_send: send: %d", wrote));
		if (wrote < 0) {
			err = GetNetworkError();
			ret = wrote;
			goto done;
		}
		left -= wrote;
		p = &p[wrote];
	}
	ret = sz;

done:
#ifndef WIN32
	signal(SIGPIPE, SIG_DFL);
#endif /* !WIN32 */
	if (0 != err) { SetNetworkError(err); }
	return ret;
}



#ifdef BF_IPV6_UNIT_TEST

#define TEST_TIMEOUT (20)

/* Will SIGSEGV/SIGBUS if you actually try to use it, but we won't.
 * Just need for it not to be NULL for the places where it gets
 * explicitly tested to disable logging calls when they are expensive.
 */
FILE *activity_file = (FILE *)42;

void activity_write(const char *file, int line, const char *fmt, ...) {
	const char *p;
	va_list va;
	time_t now;
	char buf[256];

	while (NULL != ( p = strpbrk(file, "/\\") )) { file = ++p; }
	time(&now);
	strftime(buf, 256, "%c", localtime(&now));

	va_start(va, fmt);
	fprintf(stderr, BFTRACE_FORMAT, buf, (int)getpid(), file, line);
	vfprintf(stderr, fmt, va);
	fprintf(stderr, "\n");
	fflush(stderr);
	va_end(va);
}

static const char *cmd;
static void usage() {
	fprintf(stderr,
			"Usage:\n\t%s -c port [host]\n\t%s -s port [host]\n"
			"\n"
			"For the client (-c), the 'host' specifies the target address\n"
			"and defaults to 'localhost'.  For the server, it specifies\n"
			"the bind address for the listener and defaults to an unbound\n"
			"listener.\n"
			"\n",
			cmd,
			cmd);
	exit(2);
}

static uint8_t buf[16384];
static uint8_t *bufptr = buf;
static int bufleft = 0;
static int testsock = -1;

static bf_poll_action handle_io(bf_poll *poller, int sock, int mask, void *arg) {
	uint8_t *p;
	int *ret = arg;
	int got, wrote;
	int i, o;

	if (0 == sock) {
		D(BFTRACE("handle_io: stdin -> sock"));
		i = 0;
		o = testsock;
	} else if (testsock == sock) {
		D(BFTRACE("handle_io: sock -> stdout"));
		i = sock;
		o = 1;
	} else {
		activity_write(BFHERE "handle_io: unexpected fd %d\n", sock);
		exit(-1);
	}

	got = read(i, buf, sizeof(buf));
	D(activity_write(BFHERE "handle_io: read: got=%d", got));
	if (got <= 0) {
		*ret = (0 != sock) ? 7 : 0;
		D(activity_write(BFHERE "handle_io: aborting with ret=%d", *ret));
		return BF_POLL_ACTION_ABORT;
	}
	p = buf;
	do {
		wrote = write(o, buf, got);
		D(activity_write(BFHERE "handle_io: write: wrote=%d", wrote));
		if (wrote <= 0) {
			*ret = (0 != sock) ? 1 : 8;
			D(activity_write(BFHERE "handle_io: aborting with ret=%d", *ret));
			return BF_POLL_ACTION_ABORT;
		}
		p = &p[wrote];
		got -= wrote;
	} while (got > 0);

	return BF_POLL_ACTION_CONTINUE;
}

/* Note: the bf_socket functions use bf_poll, so you have to link that in as
 * well when compiling this unit test, as in:
 *   gcc -DBF_IPV6_UNIT_TEST -o bf_ipv6 bf_ipv6.c bf_poll.c
 * We also use platform_closesocket in this code, but that is just a
 * macro for close on UNIX/Linux, which is all we're supporting here.
 *
 * Why?  Well, one of the fundamental differences between Windows and
 * UNIX/Linux is how they present fundamental kernel objects like files,
 * sockets, pipes, and so forth.  On Windows, these objects each have a
 * distinct type and can not always be used interchangably, while on
 * UNIX/Linux, all of these things show up as a simple integer file
 * descriptor.  Due to this difference, the WinSock implementation of
 * select(2) only works with sockets, not file descriptors, so stdin and
 * stdout can not be added to a bf_poll on that platform.  Making this test
 * work on Windows would therefore be harder than I currently think it is
 * worth.
 *
 * This test creates a simple TCP/IP connection driver that then acts as
 * glue between stdin/stdout and the socket, relaying data in both
 * directions and terminating as soon as either side closes, times out,
 * or returns an error.  The exit codes are:
 *  0) stdin closed
 *  1) stdout write failed
 *  2) usage error
 *  3) listen failed
 *  4) accept failed
 *  5) connect failed
 *	6) bf_poll_wait failed (select returned an error)
 *  7) socket closed gracefully by remote host
 *  8) socket I/O failure (write failed, etc.)
 *  9) timed out (no activity for 20 seconds)
 *
 * This code is only partially non-blocking in that it does not mask reads
 * and buffer write output.  Instead, it blocks during writes and does not
 * return to the select(2) call until the blocking write is done.  For the
 * purposes of this unit test, that should not be terribly important.
 */
int main(int argc, char *argv[]) {
	bf_listener *listener = NULL;
	bf_poll *poller = NULL;
	const char *host = NULL;
	int ret = 0;
	int x;
	int server = 0;
	int port = 0;

	cmd = argv[0];
	if (argc < 3 || argc > 4) { usage(); }
	if (0 == strcmp("-s", argv[1])) {
		server = 1;
	} else if (0 != strcmp("-c", argv[1])) {
		usage();
	}

	port = atoi(argv[2]);
	if (port <= 0 || port > 65535) { usage(); }
	if (argc == 4) { host = argv[3]; }

	if (server) {
		listener = bf_listener_new(host, port, 1);
		if (NULL == listener) {
			ret = 3;
			goto done;
		}
		testsock = bf_listener_accept(listener, TEST_TIMEOUT);
		if (-1 == testsock) {
			ret = 4;
			goto done;
		}
		bf_listener_destroy(listener);
		listener = NULL;
	} else {
		if (NULL == host) { host = "localhost"; }
		testsock = bf_socket_connect(host, port, TEST_TIMEOUT);
		if (-1 == testsock) {
			ret = 5;
			goto done;
		}
	}
	poller = bf_poll_new();
	bf_poll_add(poller, 0, BF_POLL_READ);
	bf_poll_add(poller, testsock, BF_POLL_READ);

	do {
		ret = bf_poll_wait(poller, TEST_TIMEOUT);
		if (ret <= 0) {
			D(activity_write(BFHERE "bf_poll_wait: ret=%d", ret));
			ret = (ret < 0) ? 6 : 9;
			break;
		}
		ret = 0;
		x = bf_poll_iterate_ready(poller, &handle_io, &ret);
		D(activity_write(BFHERE "bf_poll_iterate_ready: x=%d ret=%d", x, ret));
	} while (x >= 0 && 0 == ret);

done:
	if (-1 != testsock) { close(testsock); }
	if (NULL != poller) { bf_poll_destroy(poller); }
	if (NULL != listener) { bf_listener_destroy(listener); }
	return ret;
}

#endif /* BF_IPV6_UNIT_TEST */

