/*
Copyright (C)2008-2011, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"
#include "bf_poll.h"

#undef D
#ifdef DEBUG_BF_POLL
#  define D(x) x
# else /* !DEBUG_BF_POLL */
#  define D(x) /* no-op */
#endif /* DEBUG_BF_POLL/! */


#define BF_POLL_CANCEL  (-1)

bf_poll *bf_poll_new(void) {
	bf_poll *poller = calloc(1, sizeof(bf_poll));
	if (NULL == poller) { return NULL; }
	poller->data = calloc(16, sizeof(bf_poll_entry));
	if (NULL == poller->data) {
		free(poller);
		errno = ENOMEM;
		return NULL;
	}
	poller->cap = 16;
	return poller;
}

static int bf_poll_prep_add(bf_poll *poller) {
	bf_poll_entry *newData;
	if (poller->len < poller->cap) { return 1; }
	poller->cap <<= 1;
	newData = realloc(poller->data, poller->cap * sizeof(bf_poll_entry));
	if (NULL == newData) {
		poller->cap = poller->len + 1;
		newData = realloc(poller->data, poller->cap * sizeof(bf_poll_entry));
		if (NULL == newData) { return 0; }
	}
	poller->data = newData;
	return 1;
}

void bf_poll_destroy(bf_poll *poller) {
	if (NULL != poller) {
		free(poller->data);
		free(poller);
	}
}

static void bf_poll_clean(bf_poll *poller) {
	int i = poller->len - 1;
	while (i >= 0 && BF_POLL_CANCEL == poller->data[i].interestOps) { --i; }
	poller->len = i + 1;
}

static void bf_poll_prep(bf_poll *poller) {
	int i = 0;
	bf_poll_clean(poller);
	while (i < poller->len) {
		if (BF_POLL_CANCEL == poller->data[i].interestOps) {
			poller->data[i] = poller->data[--poller->len];
			bf_poll_clean(poller);
		} else {
			poller->data[i++].readyOps = 0;
		}
	}
}

int bf_poll_add(bf_poll *poller, SOCKET sock, int interestOps) {
	int i;

	if (INVALID_SOCKET == sock) { return -2; }

	for (i=0; i<poller->len; ++i) {
		bf_poll_entry *entry = &poller->data[i];
		if (entry->sock != sock) { continue; }
		if (BF_POLL_CANCEL == entry->interestOps) {
			D(activity_write(BFHERE "bf_poll_add: restored sock=%d ops=(cancelled)->%d",
						(int)sock, interestOps));
			entry->interestOps = interestOps;
			return -1;
		}
		D(activity_write(BFHERE "bf_poll_add: updated sock=%d ops=%d->%d",
					(int)sock, entry->interestOps, interestOps));
		return 2;
	}

	if (!bf_poll_prep_add(poller)) {
		D(BFTRACE("Out of memory while trying to add a socket to a poller"));
		return 0;
	}

	D(activity_write(BFHERE "bf_poll_add: added sock=%d ops=%d",
				(int)sock, interestOps));
	poller->data[poller->len].sock = sock;
	poller->data[poller->len].interestOps = interestOps;
	poller->len++;
	return 1;
}

int bf_poll_cancel(bf_poll *poller, SOCKET sock) {
	int i;
	bf_poll_entry *entry;

	if (poller->len == 0) { return 0; }
	for (i=0; i<poller->len; ++i) {
		entry = &poller->data[i];
		if (entry->sock == sock) {
			if (BF_POLL_CANCEL == entry->interestOps) { return -1; }
			entry->interestOps = BF_POLL_CANCEL;
			return 1;
		}
	}
	return 0;
}

int bf_poll_iterate_ready(bf_poll *poller, bf_poll_callback cb, void *arg) {
	int i;
	int count = 0;

	for (i=0; i<poller->len; ++i) {
		bf_poll_entry *entry = &poller->data[i];
		if (BF_POLL_CANCEL == entry->interestOps || 0 == entry->readyOps) { continue; }
		++count;
		switch ((*cb)(poller, entry->sock, entry->readyOps, arg)) {
			case BF_POLL_ACTION_BREAK:
				return count;
			case BF_POLL_ACTION_ABORT:
				return -count;
			default:
				continue;
		}
	}
	return count;
}

int bf_poll_iterate_all(bf_poll *poller, bf_poll_callback cb, void *arg) {
	int i;
	int count = 0;

	for (i=0; i<poller->len; ++i) {
		bf_poll_entry *entry = &poller->data[i];
		if (BF_POLL_CANCEL == entry->interestOps) { continue; }
		++count;
		switch ((*cb)(poller, entry->sock, entry->readyOps, arg)) {
			case BF_POLL_ACTION_BREAK:
				return count;
			case BF_POLL_ACTION_ABORT:
				return -count;
			default:
				continue;
		}
	}
	return count;
}


static struct timeval *init_deadline(
		struct timeval *deadline,
		struct timeval *tv,
		int timeout)
{
	if (timeout < 0) { return NULL; }
	platform_gettimeofday(deadline);
	deadline->tv_sec += timeout;
	tv->tv_sec = timeout;
	tv->tv_usec = 0;
	return tv;
}

static int check_deadline(struct timeval *deadline, struct timeval *timeleft) {
	if (NULL == timeleft) { return 1; }
	platform_gettimeofday(timeleft);
	if (timeleft->tv_usec > deadline->tv_usec) {
		if (timeleft->tv_sec >= deadline->tv_sec) { return 0; }
		timeleft->tv_usec = deadline->tv_usec + 1000000 - timeleft->tv_usec;
		timeleft->tv_sec = deadline->tv_sec - timeleft->tv_sec - 1;
	} else {
		if (timeleft->tv_sec > deadline->tv_sec) { return 0; }
		timeleft->tv_usec = deadline->tv_usec - timeleft->tv_usec;
		timeleft->tv_sec = deadline->tv_sec - timeleft->tv_sec;
	}
	return 1;
}

int bf_poll_count(bf_poll *poller) {
	int i;
	int count = 0;

	for (i=0; i<poller->len; ++i) {
		bf_poll_entry *entry = &poller->data[i];
		if (BF_POLL_CANCEL != entry->interestOps && 0 != entry->interestOps) { ++count; }
	}
	return count;
}

int bf_poll_wait(bf_poll *poller, int timeout) {
	struct timeval deadline;
	struct timeval tv;
	struct timeval *timeleft;
	fd_set rfds, wfds, efds, *r, *w;
	fd_set *e = &efds;
	int nfds = 0;
	int i, ret;

	bf_poll_prep(poller);

	timeleft = init_deadline(&deadline, &tv, timeout);
	for (;;) {
		/* select(2) isn't supposed to modify the FD sets when it returns an
		 * error, including EINTR, but I do not trust this.  We will reset
		 * the FD set on retries as well, just to be safe.
		 */
		FD_ZERO(e);
		r = NULL;
		w = NULL;
		for (i=0; i<poller->len; ++i) {
			bf_poll_entry *entry = &poller->data[i];
			if (BF_POLL_CANCEL == entry->interestOps || 0 == entry->interestOps) { continue; }
			if (0 != (entry->interestOps & BF_POLL_READ)) {
				if (NULL == r) {
					r = &rfds;
					FD_ZERO(r);
				}
				FD_SET(entry->sock, r);
				D(activity_write(BFHERE "bf_poll_wait: checking BF_POLL_READ on socket %d", (int)entry->sock));
			}
			if (0 != (entry->interestOps & BF_POLL_WRITE)) {
				if (NULL == w) {
					w = &wfds;
					FD_ZERO(w);
				}
				FD_SET(entry->sock, w);
				D(activity_write(BFHERE "bf_poll_wait: checking BF_POLL_WRITE on socket %d", (int)entry->sock));
			}
			FD_SET(entry->sock, e);
			D(activity_write(BFHERE "bf_poll_wait: checking BF_POLL_ERROR on socket %d", (int)entry->sock));
#ifdef WIN32
			nfds = 1;
# else /* !WIN32 */
			if (entry->sock >= nfds) { nfds = entry->sock + 1; }
#endif /* WIN32/! */
		}

		if (0 == nfds) {
			if (timeout <= 0) {
				errno = EINVAL;
				return -1;
			}
			D(BFTRACE("bf_poll_wait: no sockets to watch; sleeping instead..."));
			platform_sleep(timeleft->tv_sec + ((timeleft->tv_usec > 0) ? 1 : 0));
			return 0;
		}

		ret = select(nfds, r, w, e, timeleft);
		if (ret <= 0) {
			if (0 == ret || 0 == timeout) {
				D(BFTRACE("bf_poll_wait: select timed out"));
				return 0;
			}
			if (IsNetworkErrorSelectInterrupted()) {
				if (check_deadline(&deadline, timeleft)) {
					D(BFTRACE("bf_poll_wait: select was interrupted"));
					continue;
				}
				D(BFTRACE("bf_poll_wait: select was interrupted and timeout has elapsed"));
				SetNetworkErrorTimedOut();
				return 0;
			}
			return -1;
		}

		for (i=0; i<poller->len; ++i) {
			bf_poll_entry *entry = &poller->data[i];
			if (0 != (BF_POLL_READ & entry->interestOps) && FD_ISSET(entry->sock, r)) {
				entry->readyOps |= BF_POLL_READ;
				D(activity_write(BFHERE "bf_poll_wait: got BF_POLL_READ on socket %d", (int)entry->sock));
			}
			if (0 != (BF_POLL_WRITE & entry->interestOps) && FD_ISSET(entry->sock, w)) {
				entry->readyOps |= BF_POLL_WRITE;
				D(activity_write(BFHERE "bf_poll_wait: got BF_POLL_WRITE on socket %d", (int)entry->sock));
			}
			if (FD_ISSET(entry->sock, e)) {
				entry->readyOps |= BF_POLL_ERROR;
				D(activity_write(BFHERE "bf_poll_wait: got BF_POLL_ERROR on socket %d", (int)entry->sock));
			}
		}
		D(activity_write(BFHERE "bf_poll_wait: returning %d", ret));
		return ret;
	}
}

#ifndef ZOS

static inline int bf_poll_can(SOCKET sock, int interestOps, int timeout)
{
	bf_poll poller;
	bf_poll_entry entry;

	if (INVALID_SOCKET == sock) {
		errno = EBADF;
		return 0;
	}

	entry.sock = sock;
	entry.interestOps = interestOps;
	poller.len = 1;
	poller.cap = 1;
	poller.data = &entry;

	return bf_poll_wait(&poller, timeout);
}

int bf_poll_can_read(SOCKET sock, int timeout) {
	return bf_poll_can(sock, BF_POLL_READ, timeout);
}
int bf_poll_can_write(SOCKET sock, int timeout) {
	return bf_poll_can(sock, BF_POLL_WRITE, timeout);
}
int bf_poll_has_exception(SOCKET sock, int timeout) {
	return bf_poll_can(sock, BF_POLL_ERROR, timeout);
}
#endif
