/*
Copyright (C)2008-2011, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#ifndef BF_POLL_H
#define BF_POLL_H

/*
 * This file and the accompanying .c file provide an abstraction for
 * working with select/poll in a convenient and platform-independent
 * way.  The functionality is similar to that provided by the Selector
 * class in Java's NIO code, with some inspiration also taken from
 * Perl's IO::Select module.
 *
 * For simplicity, this code currently only supports select(2) and the
 * WinSock equivalent.  However, it is deliberately structured such
 * that it should be relatively easy to add support for poll(2) and
 * other similar mechanisms.
 */

/* Bit masks for read/write/error activity.
 * Note that BF_POLL_ACCEPT is synonymous with BF_POLL_READ and
 * and that BF_POLL_CONNECT is synonymous with BF_POLL_WRITE.
 * BF_POLL_ERROR is checked implicitly when either BF_POLL_READ or
 * BF_POLL_WRITE is checked.
 */
#define BF_POLL_READ    (1<<0)
#define BF_POLL_ACCEPT  BF_POLL_READ
#define BF_POLL_WRITE   (1<<1)
#define BF_POLL_CONNECT BF_POLL_WRITE
#define BF_POLL_ERROR   (1<<2)

#define BF_POLL_READY(ops,mask)     (0 != (mask & (ops|BF_POLL_ERROR)))

/* Legal return values for bf_poll_iterate's callback function */
typedef enum _bf_poll_action {
	BF_POLL_ACTION_CONTINUE,
	BF_POLL_ACTION_BREAK,
	BF_POLL_ACTION_ABORT
} bf_poll_action;

/* Container for registrations */
typedef struct _bf_poll_entry {
	SOCKET sock;
	int interestOps;
	int readyOps;
} bf_poll_entry;

/* Container for the poller's state data */
typedef struct _bf_poll {
	int len;
	int cap;
	bf_poll_entry *data;
} bf_poll;

/*
 * Callback functions used by iterators.
 *  poller - the poller that is being iterated
 *  sock - the socket that has activity
 *  mask - the activity that the socket had
 *  arg - the caller-defined argument supplied to bf_poll_iterate
 *  return:
 *		BF_POLL_ACTION_CONTINUE - The callback succeeded and the iterator
 *				should continue with any remaining ready sockets.
 *		BF_POLL_ACTION_BREAK - The callback succeeded, but the iterator
 *				should not process any additional ready sockets at
 *				this time.
 *		BF_POLL_ACTION_ABORT - The callback failed, and the iterator
 *				should not process any additional ready sockets at
 *				this time.
 */
typedef bf_poll_action (*bf_poll_callback)(bf_poll *poller, SOCKET sock, int mask, void *arg);

/*
 * Creates a new poller with no sockets registered.
 * return:
 *		The new poller object.  It is the caller's responsibility to destroy
 *		the poller with bf_poll_destroy when done with it.
 */
bf_poll *bf_poll_new(void);

/*
 * Destroys a poller, freeing its resources.
 * Note: It is the still the caller's responsibility to close the sockets.
 * See bf_poll_add for details.
 */
void bf_poll_destroy(bf_poll *poller);

/*
 * Add a socket to the poller or update its interestOps mask.
 *  poller - the poller to use
 *  sock - the socket to be added/updated
 *  interestOps - the socket operations of interest
 * return:
 *		1 = added
 *		2 = updated
 *		0 = out of memory
 *		-1 = re-added (had been cancelled)
 *      -2 = ignored (invalid socket)
 * Note: It is the caller's responsibility to close the socket at a later
 * time.  The most straight-forward way to handle this is to close any
 * socket for which bf_poll_cancel is called and to use bf_poll_iterate_all
 * to close off the others before destroying the poller.
 */
int bf_poll_add(bf_poll *poller, SOCKET sock, int interestOps);

/*
 * Cancels any interest in the specified socket.  
 *  poller - the poller with which the socket is thought to be registered
 *  sock - the socket of interest
 * return:
 *		1 = successful cancel
 *		0 = socket was never registered or had already been cleaned out
 *				by a call to bf_poll_wait
 *		-1 = already cancelled by an earlier call
 * Note: It is still the caller's responsibility to close the socket.
 * See bf_poll_add for details.
 */
int bf_poll_cancel(bf_poll *poller, SOCKET sock);


/*
 * Returns the count of sockets that are registered with this poller.
 * Sockets that are registered with 0 interestOps or that have been
 * cancelled are not included in the total.
 */
int bf_poll_count(bf_poll *poller);

/*
 * Waits for activity.
 *  poller - the poller to wait against
 *  timeout - the maximum time to wait for activity
 * return:
 *		>0 = socket activity is available
 *		0 = timeout
 *		-1 = catastrophic failure
 */
int bf_poll_wait(bf_poll *poller, int timeout);

/*
 * Iterates over the sockets that have available activity.
 *  poller - the poller to iterate
 *  cb - the callback function to call for each ready socket
 *  arg - caller-defined data to be supplied to the callback
 * return:
 *    0 = No activity on any of the sockets, so no callbacks were issued.
 *    >0 = The number of callbacks issued, all of which returned
 *			BF_POLL_ACTION_CONTINUE, except possibly for the last one,
 *			which may have returned BF_POLL_ACTION_BREAK to stop the
 *			iteration.  If the caller needs to know which, then it can
 *			use the data pointed to by "arg" to make note of this before
 *			returning BF_POLL_ACTION_BREAK from the callback.
 *    <0 = the negative of the number of callbacks issued.  This only
 *			happens when a callback return BF_POLL_ACTION_ABORT.
 */
int bf_poll_iterate_ready(bf_poll *poller, bf_poll_callback cb, void *arg);

/*
 * Iterates over the sockets that have not been cancelled, regardless
 * of whether or not they are ready.  All arguments and the return value
 * behave identically to those of bf_poll_iter_ready.
 */
int bf_poll_iterate_all(bf_poll *poller, bf_poll_callback cb, void *arg);

/* These functions are all equivalent to creating a new poller, adding
 * the specified socket to it with the appropriate interest ops, calling
 * bf_poll_wait, and destroying the poller; however, this implementation
 * is optimized for working with a single socket, so it is much more
 * efficient.
 *  sock - the socket to watch
 *  timeout - the maximum time to wait for activity
 * return: as for bf_poll_wait
 */
int bf_poll_can_read(SOCKET sock, int timeout);
int bf_poll_can_write(SOCKET sock, int timeout);
int bf_poll_has_exception(SOCKET sock, int timeout);
#define bf_poll_can_accept(sock,timeout) bf_poll_can_read(sock,timeout)
#define bf_poll_can_connect(sock,timeout) bf_poll_can_write(sock,timeout)


/****************************************************************
 * Workarounds for WinSock differences
 ****************************************************************/

#undef GetNetworkError
#undef SetNetworkError
#undef IsNetworkErrorInProgress
#undef SetNetworkErrorInProgress
#undef IsNetworkErrorNotConnected
#undef SetNetworkErrorNotConnected
#undef IsNetworkErrorConnectionReset
#undef SetNetworkErrorConnectionReset
#undef IsNetworkErrorInvalidArgument
#undef SetNetworkErrorInvalidArgument
#undef IsNetworkErrorTimedOut
#undef SetNetworkErrorTimedOut
#undef IsNetworkErrorHostNotFound
#undef SetNetworkErrorHostNotFound
#undef IsNetworkErrorAddressInUse
#undef SetNetworkErrorAddressInUse
#undef IsNetworkErrorSelectInterrupted
#undef SetNetworkErrorSelectInterrupted

#ifdef WIN32
#  define GetNetworkError() WSAGetLastError()
static inline void SetNetworkError(int err) {
	WSASetLastError(err);
	errno = err;
}
static inline BOOL IsNetworkErrorInProgress(void) {
	int err = WSAGetLastError();
	return WSAEWOULDBLOCK == err || WSAEINPROGRESS == err;
}
#  define SetNetworkErrorInProgress() SetNetworkError(WSAEINPROGRESS)
#  define IsNetworkErrorNotConnected() (WSAENOTCONN == WSAGetLastError())
#  define SetNetworkErrorNotConnected() SetNetworkError(WSAENOTCONN)
#  define IsNetworkErrorConnectionReset() (WSAECONNRESET == WSAGetLastError())
#  define SetNetworkErrorConnectionReset() SetNetworkError(WSAECONNRESET)
#  define IsNetworkErrorInvalidArgument() (WSAEINVAL == WSAGetLastError() || WSAENOTSOCK == WSAGetLastError())
#  define SetNetworkErrorInvalidArgument() SetNetworkError(WSAEINVAL)
#  define IsNetworkErrorTimedOut() (WSAETIMEDOUT == WSAGetLastError())
#  define SetNetworkErrorTimedOut() SetNetworkError(WSAETIMEDOUT)
#  define IsNetworkErrorHostNotFound() (WSAHOST_NOT_FOUND == WSAGetLastError())
#  define SetNetworkErrorHostNotFound() SetNetworkError(WSAHOST_NOT_FOUND)
#  define IsNetworkErrorAddressInUse() (WSAEADDRINUSE == WSAGetLastError())
#  define SetNetworkErrorAddressInUse() SetNetworkError(WSAEADDRINUSE)
#  define IsNetworkErrorSelectInterrupted() (WSAEINTR == WSAGetLastError())
#  define SetNetworkErrorSelectInterrupted() SetNetworkError(WSAEINTR)
# else /* !WIN32 */
#  define GetNetworkError() (errno)
#  define SetNetworkError(err) do { errno = err; } while (0)
#  define IsNetworkErrorInProgress() (EINPROGRESS == errno)
#  define SetNetworkErrorInProgress() SetNetworkError(EINPROGRESS)
#  define IsNetworkErrorNotConnected() (ENOTCONN == errno || EINVAL == errno)
#  define SetNetworkErrorNotConnected() SetNetworkError(ENOTCONN)
#  define IsNetworkErrorConnectionReset() (ECONNRESET == errno || EBADF == errno)
#  define SetNetworkErrorConnectionReset() SetNetworkError(ECONNRESET)
#  define IsNetworkErrorInvalidArgument() (EINVAL == errno || ENOTSOCK == errno)
#  define SetNetworkErrorInvalidArgument() SetNetworkError(EINVAL)
#  define IsNetworkErrorTimedOut() (ETIMEDOUT == errno)
#  define SetNetworkErrorTimedOut() SetNetworkError(ETIMEDOUT)
#  define IsNetworkErrorHostNotFound() (ENOENT == errno)
#  define SetNetworkErrorHostNotFound() SetNetworkError(ENOENT)
/* i5/OS seems to inexplicably throw EAPAR, and refuse to translate it,
 * if IPv6 is not configured.  This gives the annoying message
 * "Error 3418 occurred." if the error is not ignored.  To silence it,
 * we treat EAPAR the same way we treat EADDRINUSE; that is, we ignore
 * it if we already have at least one listener.
 */
#  ifdef ISERIES_AGENT
#    ifndef EAPAR
#      define EAPAR 3418
#    endif /* !EAPAR */
#    define IsNetworkErrorAddressInUse() (EADDRINUSE == errno || EAPAR == errno)
#   else /* !ISERIES_AGENT */
#    define IsNetworkErrorAddressInUse() (EADDRINUSE == errno)
#  endif /* ISERIES_AGENT/! */
#  define SetNetworkErrorAddressInUse() SetNetworkError(EADDRINUSE)
#  define IsNetworkErrorSelectInterrupted() (EINTR == errno || EAGAIN == errno)
#  define SetNetworkErrorSelectInterrupted() SetNetworkError(EINTR)
#endif /* WIN32/! */

#endif /* BF_POLL_H */

