/*
Copyright (C)2011, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"
#include "bf_rand.h"


#undef D
#ifdef DEBUG_BF_RAND
#  define D(x) x
# else /* !DEBUG_BF_RAND */
#  define D(x) /* no-op */
#endif /* DEBUG_BF_RAND/! */

/* ================================================================
 * bf_rand
 * ----------------------------------------------------------------
 * These functions implement a Linear Congruential Pseudo-Random Number
 * Generator using the same algorithm that is employed by the Java class
 * java.util.Random.  This will be used by the PRNG algorithm for XSTREAM
 * file transfers.
 * ---------------------------------------------------------------- */

#define BF_RAND_POLY  ((((int64_t)0x5DEECE66) << 4) + 0xD)
#define BF_RAND_MASK  ((((int64_t)1) << 48) - 1)

/* Random(long) */
void bf_rand_init(bf_rand *state, int64_t seed) {
	*state = (bf_rand)((seed ^ BF_RAND_POLY) & BF_RAND_MASK);
}

/* So we can init using the shared secret */
void bf_rand_init_bytes(bf_rand *state, const uint8_t *bytes, size_t len) {
	uint64_t x = 0;
	const uint8_t *p = bytes;
	const uint8_t *stop = &p[(int)len];
	while (p < stop) { x = BF_RAND_POLY ^ (x >> 37) ^ (x << 11) ^ *p++; }
	bf_rand_init(state, (int64_t)x);
	D(activity_write(BFHERE "bf_rand_init_bytes: %08X%08X", (unsigned int)(x >> 32), (unsigned int)x));
}

/* Random.next(int) */
int32_t bf_rand_next(bf_rand *state, int bits) {
	int32_t ret;
	*state = (*state * BF_RAND_POLY + 0xB) & BF_RAND_MASK;
	ret = (int32_t)(*state >> ((48 - bits) & 63));
	D(activity_write(BFHERE "bf_rand_next(%p,%d): %08X", state, bits, (unsigned int)(uint32_t)ret));
	return ret;
}

uint8_t bf_rand_next_byte(bf_rand *state) {
	uint8_t byte = (uint8_t)bf_rand_next(state, 8);
	D(activity_write(BFHERE "bf_rand_next_byte: %02X", byte));
	return byte;
}

/* Random.nextInt() */
int32_t bf_rand_next_int(bf_rand *state) {
	int32_t ret = bf_rand_next(state, 32);
	D(activity_write(BFHERE "bf_rand_next_int: %08X", (unsigned int)(uint32_t)ret));
	return ret;
}

/* Random.nextLong() */
int64_t bf_rand_next_long(bf_rand *state) {
	int64_t ret = (((int64_t)bf_rand_next(state, 32)) << 32) + bf_rand_next(state, 32);
	D(activity_write(BFHERE "bf_rand_next_long: %08X%08X", (unsigned int)(ret >> 32), (unsigned int)ret));
	return ret;
}

/* Random.nextInt(int) */
int32_t bf_rand_next_int_mod(bf_rand *state, int32_t n) {
	if (n <= 0) {
		/* Java throws an exception for this.  We'll just take a one's
		 * compliment to ensure a positive value.
		 */
		n = ~n;
	}
	if ((n & -n) == n) { /* power of 2 */
		return (int32_t)((n * (int64_t)bf_rand_next(state, 31)) >> 31);
	} else {
		int32_t bits, val;
		do {
			bits = bf_rand_next(state, 31);
			val = bits % n;
		} while (bits - val + (n-1) < 0);
		return val;
	}
}


#ifdef BF_RAND_UNIT_TEST
void activity_write(const char *file, int line, const char *fmt, ...) {
	const char *p;
	va_list va;
	time_t now;
	char buf[256];

	while (NULL != ( p = strpbrk(file, "/\\") )) { file = ++p; }
	time(&now);
	strftime(buf, 256, "%c", localtime(&now));

	va_start(va, fmt);
	fprintf(stderr, BFTRACE_FORMAT, buf, (int)getpid(), file, line);
	vfprintf(stderr, fmt, va);
	fprintf(stderr, "\n");
	fflush(stderr);
	va_end(va);
}

static const char *cmd;

static void usage(void) {
	printf("Usage: %s secret_bytes_as_hex\n", cmd);
	printf("%s",
			"\n"
			"secret_bytes_as_hex\n"
			"\tThe seed for the pseudo-random number generator\n"
			"\n"
			"This unit test acts as a filter, accepting bytes from stdin,\n"
			"XOR'ing them against the bytes generated by the PRNG, and\n"
			"writing them to stdout.  This involves binary data, and it is\n"
			"up to the caller to arrange for it to be displayed in a\n"
			"sensible way, as terminals will generally not be capable of\n"
			"displaying arbitrary binary data intelligibly.\n"
			"\n"
			"Since the PRNG algorithm is fixed and XOR is a symmetric math\n"
			"operation (If A^B=C, then C^B=A), running this program with\n"
			"a given seed and piping its output to input of this program\n"
			"run again with the same random seed should reproduce the\n"
			"original input.  In other words:\n"
			"\techo hello | ./bf_rand 42424242 | ./bf_rand 42424242\n"
			"should display 'hello'\n"
			"\n");
	exit(1);
}

static int hex_char(char c) {
	if (c >= '0' && c <= '9') { return c - '0'; }
	if (c >= 'A' && c <= 'F') { return c - 'A' + 10; }
	if (c < 'a' || c > 'f') {
		printf("Invalid hex char 0x%02X\n", (unsigned int)(unsigned char)c);
		usage();
	}
	return c - 'a' + 10;
}


static uint8_t *dehex(const char *src, size_t *len) {
	uint8_t *buf, *b;
	const uint8_t *s = (const uint8_t *)src;

	if (NULL == src) {
		printf("unhex: NULL src\n");
		usage();
	}

	*len = strlen(src);
	if (1 == (*len & 1)) {
		printf("unhex: bad strlen: %d\n", (int)*len);
		usage();
	}

	*len >>= 1;
	buf = malloc((*len > 0) ? *len : 1);
	if (NULL == buf) {
		printf("unhex: Out of memory\n");
		usage();
	}

	b = buf;
	while ('\0' != *s) {
		int x = hex_char(*s++);
		*b++ = (uint8_t)((x << 4) + hex_char(*s++));
	}
	return buf;
}

static uint8_t iobuf[BFBUFSIZ];
int main(int argc, char *argv[]) {
	bf_rand state;
	uint8_t *secret;
	size_t secretlen;
	char *stop = NULL;
	int i, count;

	cmd = argv[0];

	if (argc >= 2) {
		secret = dehex(argv[1], &secretlen);
		if (argc >= 3) {
			unsigned long ulcount = strtoul(argv[2], &stop, 10);
			count = (int)ulcount;
			if ('\0' != *stop || stop == argv[2] || ulcount != (unsigned long)(unsigned int)count) {
				printf("bad count: %s; stop=%s; ulcount=%lu count=%d\n",
						argv[2], stop, ulcount, count);
				usage();
			}
		} else {
			count = 0;
		}
	} else {
		secret = dehex("", &secretlen);
	}

	bf_rand_init_bytes(&state, secret, secretlen);
	if (count > 0) {
		for (i=0; i<count; i++) {
			printf("%02X", bf_rand_next_byte(&state));
			if ((i & 31) == 31) { printf("\n"); }
		}
		printf("\n");
		return 0;
	}

	for (;;) {
		int got = (int)read(0, iobuf, BFBUFSIZ);
		uint8_t *p, *stop;
		
		if (got <= 0) { return (got < 0) ? 2 : 0; }

		p = iobuf;
		stop = &p[got];
		while (p < stop) {
			*p = (uint8_t)(*p ^ bf_rand_next_byte(&state));
			++p;
		}
		p = iobuf;
		do {
			int wrote = (int)write(1, p, got);
			if (wrote <= 0) { return 3; }
			got -= wrote;
			p = &p[wrote];
		} while (got > 0);
	}
}

#endif /* BF_RAND_UNIT_TEST */



