/*
Copyright (C)2007-2013, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#define BFCONF_C
#include "bf_agent.h"

/* Access to bfagent.conf settings.  See bfconf.h for more information. */
#define MINIMUM_READ_TIMEOUT 5
#define DEFAULT_READ_TIMEOUT 1800
#define MAXIMUM_READ_TIMEOUT 86400

#define MINIMUM_XSTREAM_TIMEOUT 5
#define DEFAULT_XSTREAM_TIMEOUT 20
#define MAXIMUM_XSTREAM_TIMEOUT 86400

static hash_ctx *ctx = NULL;
static hash_table *ht = NULL;

static char *bfconf_file = NULL;
static int bfconf_errno = 0;

static const char *magic_login = "magic_login";
static const char *update_path = "update_path";
static const char *password_prop = "_password";
static const char *locator_prop = "_locator";

extern int agent_re_exec;

bfconf_data_t bfconf_data;
int read_timeout = DEFAULT_READ_TIMEOUT;
int xstream_conn_timeout = DEFAULT_XSTREAM_TIMEOUT;
int xstream_recv_timeout = DEFAULT_XSTREAM_TIMEOUT;
int xstream_send_timeout = DEFAULT_XSTREAM_TIMEOUT;

static void init_timeout(int *timeout, const char *key, int min, int def, int max, int optional) {
	char *p = bfconf_get(key, NULL);

	if (NULL != p) {
		*timeout = atoi(p);
		free(p);
		if (optional && *timeout <= 0) {
			*timeout = 0;
		} else if (*timeout < min) {
			*timeout = min;
		} else if (*timeout > max) {
			*timeout = max;
		}
	} else {
		*timeout = def;
	}
}

void bfconf_refresh(void) {
	FILE    *f;
	char    *buf, *key, *value;
	char    *sptr;
	size_t 	len;

	/* Start with a clean slate */
	hash_clear(ht);
	bfconf_errno = 0;

	f = platform_fopen(bfconf_get_path(), "r");
	if (f == NULL) {
		bfconf_errno = errno;
		return;
	}
	buf = calloc(1, BFBUFSIZ);

	while (platform_fgets(buf, BFBUFSIZ, f) != NULL) {
		sptr = NULL;
		key = bf_next_token(buf, " \t\r\n", &sptr);
		if (NULL == key || *key == '\0' || *key == '#') { continue; }
		LowerString(key);
		len = strlen(key) + 1;	/* include the null in the key */
		if (hash_contains_key(ht, key, len)) { continue; }  /* first one wins */
		value = bf_next_token(NULL, "\t\r\n", &sptr);
		hash_put(ht, strdup(key), len, strdup((NULL != value) ? value : ""));
	}

	free(buf);
	fclose(f);

	/* There are several properties that we want to test for whether or
	 * the property exists without necessarily caring about the value
	 * that it contains at the time.  This caches these in a bit field
	 * so that they can be checked more conveniently and efficiently.
	 */
#undef REFRESH_EXISTS
#define REFRESH_EXISTS(x) bfconf_data.x = hash_contains_key_str(ht, #x)
	REFRESH_EXISTS(activity_log);
	REFRESH_EXISTS(cc_suppress_server_root);
	REFRESH_EXISTS(cygwin);
	REFRESH_EXISTS(disable_cleanup_during_exit);
	REFRESH_EXISTS(disable_telnet_support);
	REFRESH_EXISTS(disable_transcode);
	REFRESH_EXISTS(enable_agent_dll);
	REFRESH_EXISTS(enable_credential_retention);
	REFRESH_EXISTS(enable_deprecated_path_mapping);
	REFRESH_EXISTS(fail_if_dot_source_fails);
	REFRESH_EXISTS(fail_if_log_file_fails);
	REFRESH_EXISTS(getaddrinfo_use_addrconfig);
	REFRESH_EXISTS(ignore_clearcase_view);
	REFRESH_EXISTS(ipv6_connect_first_addr_only);
	REFRESH_EXISTS(leave_tmp_file);
	REFRESH_EXISTS(map_drive_is_failure);
	REFRESH_EXISTS(nologonshell);
	REFRESH_EXISTS(no_lang_env);
	REFRESH_EXISTS(no_preparse_command);
	REFRESH_EXISTS(no_pty);
	REFRESH_EXISTS(password_encrypt_config);
	REFRESH_EXISTS(password_encrypt_module);
	REFRESH_EXISTS(shell_compatible_undef_vars);
	REFRESH_EXISTS(shellarg);
	REFRESH_EXISTS(win_reexec_after_auth);
	REFRESH_EXISTS(xstream_allow_ssl_mismatch);
	REFRESH_EXISTS(xstream_debug);
	REFRESH_EXISTS(xstream_disable_md5);
	REFRESH_EXISTS(xstream_listen_randomize);

	/* The read_timeout setting controls how long we will wait for a request
	 * before giving up on it.  This, along with the checks for HTTP-style
	 * requests, was added to address the problem of the agent process
	 * sticking around indefinitely when contacted by certain port scanning
	 * software that sent a non-agent request.  Since the request did not
	 * say "go", the agent would continue to wait for the request to complete.
	 * There are now several checks to validate the request format as quickly
	 * as possible, such as by rejecting non-UTF-8 characters, rejecting tags
	 * that are too long, and timing out a request that takes too long.  This
	 * enforces some sanity limits on the read timeout value.
	 */
	init_timeout(&read_timeout, "read_timeout",
			MINIMUM_READ_TIMEOUT, DEFAULT_READ_TIMEOUT, MAXIMUM_READ_TIMEOUT, 1);
	init_timeout(&xstream_conn_timeout, "xstream_conn_timeout",
			MINIMUM_XSTREAM_TIMEOUT, DEFAULT_XSTREAM_TIMEOUT, MAXIMUM_XSTREAM_TIMEOUT, 1);
	init_timeout(&xstream_recv_timeout, "xstream_recv_timeout",
			MINIMUM_XSTREAM_TIMEOUT, DEFAULT_XSTREAM_TIMEOUT, MAXIMUM_XSTREAM_TIMEOUT, 1);
	init_timeout(&xstream_send_timeout, "xstream_send_timeout",
			MINIMUM_XSTREAM_TIMEOUT, DEFAULT_XSTREAM_TIMEOUT, MAXIMUM_XSTREAM_TIMEOUT, 1);

#ifdef WIN32
	/* For simplicity, the pipe between the parent agent process and its
	 * re-exec is always forced to carry UTF-8.
	 */
	if (agent_re_exec) { bfconf_data.disable_transcode = 1; }
#endif
}

static void bfconf_init(void) {
	if (ctx == NULL) {
		ctx = hash_ctx_new();
		ht = hash_new(ctx, 89);
		memset(&bfconf_data, '\0', sizeof(bfconf_data));
	} else {
		/* Free any strings */
	}
	bfconf_refresh();
}


void bfconf_set_path(const char *path) {
	if (NULL != bfconf_file)
		free(bfconf_file);
	bfconf_file = NULL != path ? strdup(path) : platform_default_config_file();
	bfconf_init();
}

const char *bfconf_get_path(void) {
	if (NULL == bfconf_file)
		bfconf_set_path(NULL);
	return bfconf_file;
}

int bfconf_get_errno(void) {
	return bfconf_errno;
}

char *bfconf_get(const char *key, const char *def) {
	const char *value;

	if (ctx == NULL) bfconf_init();
	value = hash_get_str(ht, key);
	if (value == NULL) value = def;
	return value == NULL ? NULL : strdup(value);
}

int bfconf_exists(const char *key) {
	if (ctx == NULL) bfconf_init();
	return hash_contains_key_str(ht, key);
}

/* Implementation for cmd sysconf */
int FuncDumpConf(void)
{
	const char *key, *value;
	char *buf = malloc(BFBUFSIZ);
	hash_iter *hi;
	int debug = request_has(TAG_DEBUG);
	size_t len;
	size_t cpy_count = 0;

	ClobberPasswords();

	if (ctx == NULL) bfconf_init();
	if (bfconf_errno != 0) {
		bfconf_refresh();
		if (bfconf_errno != 0) {
			send_msg("EXEC", "ConfFail", "si",
					bfconf_file, bfconf_errno);
			return 1;
		}
	}

	/* In debug mode, dump the settings.  The magic_login value
	 * is suppressed for security reasons, and the update_path
	 * setting is suppressed entirely so it can be reported last.
	 * That was needed for 3.8 console compatibility -- probably
	 * not important at this point.
	 */
	if (debug) {
		hi = hash_iter_new(ht);
		while (hash_iter_next(hi)) {
			key = hi->entry->key;
			value = hi->entry->value;

			if (0 == strcmp(key, magic_login)) { value = "*****"; }
			if (NULL != strstr(key, password_prop) && NULL == strstr(key, locator_prop)) { value = "*****"; }
			else if (0 == strcmp(key, update_path)) { continue; }
			else if (value == NULL) { value = ""; }

			len = strlen(key);
			memcpy(buf, key, len);
			buf[len++] = '=';
			cpy_count = (strlen(value) < BFBUFSIZ) ? (strlen(value) + 1) : BFBUFSIZ;
			strncpy(&buf[len], value, cpy_count);
			send_log("CONF", buf);
		}
		hash_iter_destroy(hi);
	}

	/* Provide the update_path setting. */
	value = platform_update_path();
	format_message(buf, "%s=%s", update_path, value);
	send_log("CONF", buf);
	free(buf);

	send_msg("EXEC", "ShowDone", NULL);
	return 0;
}

