/*
Copyright (C)2008-2013, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

/* This file provides the interface for interacting with the password
 * encryption modules.
 */

#include "bf_agent.h"
#ifdef WIN32
	#include <excpt.h>
	#include <strsafe.h>
#elif defined(HAVE_DLFCN_H)
	#include <dlfcn.h>
#else
	#include "bf_dlfcn.h"
#endif

static const char *PASSWORD_ENCRYPT_MODULE_PROPERTY = "password_encrypt_module";
static const char *DIGEST_ALGORITHM = "digest_algorithm";
static const char *BFDEBUG_SECURITY_CRYPT = "BFDEBUG_SECURITY_CRYPT";

static const passwd_ops *myops = NULL;
static int initialized = 0;
static int DEBUG = 0;

typedef const passwd_ops *(*InitializationProcedure)(const char *);

#ifndef WIN32
static void *dllhandle = NULL;

static int unix_loader_init(void) {
	InitializationProcedure functionhandle;
	char *params = NULL, *dll_path = NULL, *bfpwcrypt_path = NULL;
	char *function_name = NULL;
	char *temp = NULL;
	char *word, *lastword, *module_name, *error;
	const char *sep;
	int ret = 0;
	size_t cpy_count = 0;

	/* Already initialized or no module is specified in the conf file */
	if (initialized) { return 0; }
	params = bfconf_get(PASSWORD_ENCRYPT_MODULE_PROPERTY, NULL);
	if (NULL == params) { return 0; }

	/* Remove the path to get just "split" */
	temp = strdup(params);
	sep = ";";
	dll_path = strtok(temp, sep);
	bfpwcrypt_path = strtok(NULL, sep);
	if (NULL != activity_file && bfpwcrypt_path != NULL) {
		activity_write(BFHERE "bfpwcrypt_path: [%s]", bfpwcrypt_path);
	}
	
	/* Remove the path to get just "module.dll" */
	temp = strdup(dll_path);
	sep = "\\/";
	lastword = NULL;
	for (word = strtok(temp, sep); word; word = strtok(NULL, sep)) {
		lastword = word;
	}

	/* If dll_path is NULL or matches the regular expression /^[\/\\]*$/,
	 * then lastword will be NULL, here.
	 */
	if (NULL == lastword) {
		send_msg("AUTH", "AuthModuleConfMalformed", NULL);
		ret = -1;
		goto done;
	}
			
	/* Now remove ".dll" from the end. */
	module_name = strtok(lastword, ".");

	/* If dll_path is matches the regular expression /^[\/\\]*\.[\.\/\\]*$/,
	 * then module_name will be NULL, here.
	 */
	if (NULL == module_name) {
		send_msg("AUTH", "AuthModuleConfMalformed", NULL);
		ret = -2;
		goto done;
	}
	if (NULL != activity_file) {
		activity_write(BFHERE "module [%s]", module_name);
		if (!strcmp(module_name, "bfcrypt")) {
			activity_write(BFHERE "Debugging info for password decrypt operations can be found in ./bfcrypt.txt.");
		}
	}

	/* This allows us to load "module_init" per the struct design. */
	function_name = malloc(strlen(module_name) + 6);
	cpy_count = (strlen(module_name) < BFBUFSIZ) ? (strlen(module_name) + 1) : BFBUFSIZ;
	strncpy(function_name, module_name, cpy_count);
	strncat(function_name, "_init", 5);
	if (NULL != activity_file) {
		activity_write(BFHERE "loading [%s]", dll_path);
	}

	/* Try to open the .so library configured. */
	dllhandle = dlopen(dll_path, RTLD_NOW);
	if (!dllhandle) {
		dllhandle = dlopen(dll_path, RTLD_GLOBAL| RTLD_LAZY);
		if (!dllhandle) {
			error = dlerror();
			ret = -3;
			if (NULL != activity_file && strcmp(module_name, "bfcrypt") == 0) {
				activity_write(BFHERE "Ensure bfpwcrypt.conf has been added to the agent directory.  Search documentation for more information about bfpwcrypt.conf.");
			}
			send_msg("AUTH", "AuthModuleLoadFail", "iss",
					ret,
					module_name,
					(NULL != error) ? error : dll_path );
			goto done;
		}
	}
		
	/* Load the _init method. */
	functionhandle = (InitializationProcedure)dlsym(dllhandle, function_name);   
	if (!functionhandle) {
		error = dlerror();
		ret = -4;
		send_msg("AUTH", "AuthModuleNoInit", "iss",
				ret,
				module_name,
				(NULL != error) ? error : function_name );
		dlclose(dllhandle);
		dllhandle = NULL;
		goto done;
	}

	/* Call the function */
	if (NULL != activity_file) {
		activity_write(BFHERE "calling [%s: %s()]",
				dll_path, function_name);
	}
	myops = (*functionhandle)(bfpwcrypt_path);
	if (NULL == myops) {
		ret = -5;
		send_msg("AUTH", "AuthModuleInitFail", "iss",
				ret,
				module_name,
				function_name);
		dlclose(dllhandle);
		dllhandle = NULL;
		goto done;
	}

	initialized = 1;

done:
	if (NULL != temp) { free(temp); }
	if (NULL != function_name) { free(function_name); }
	if (NULL != dll_path) { free(dll_path); }
	return ret;
}
#endif


#ifdef WIN32
static int filter(unsigned int code, struct _EXCEPTION_POINTERS *ep) {

	LPTSTR lpszFunction = TEXT("win_loader_init");
	LPVOID lpMsgBuf;
	LPVOID lpDisplayBuf;
	DWORD dw = GetLastError(); 

	FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER | 
		FORMAT_MESSAGE_FROM_SYSTEM |
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		dw,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		(LPTSTR) &lpMsgBuf,
		0, NULL );

	/* Display the error message and exit the process */

	lpDisplayBuf = (LPVOID)LocalAlloc(LMEM_ZEROINIT, (lstrlen((LPCTSTR)lpMsgBuf)+lstrlen((LPCTSTR)lpszFunction)+40)*sizeof(TCHAR)); 
	StringCchPrintf((LPTSTR)lpDisplayBuf, 
	LocalSize(lpDisplayBuf) / sizeof(TCHAR),
	TEXT("%s failed with error %d: %s"), 
	lpszFunction, dw, lpMsgBuf); 

	if (DEBUG && NULL != activity_file) {
		BFTRACE("***** WIN32 EXCEPTION *****");
		activity_write(BFHERE "Message = %s", lpDisplayBuf);
		activity_write(BFHERE "Code = %u", code);
		activity_write(BFHERE "Exception address = %u, code = %u, flags = %u, info = %u",
				ep->ExceptionRecord->ExceptionAddress,
				ep->ExceptionRecord->ExceptionCode,
				ep->ExceptionRecord->ExceptionFlags,
				ep->ExceptionRecord->ExceptionInformation);
		activity_write(BFHERE "Context flags = %u", ep->ContextRecord->ContextFlags);
	}
	return 0;
} 

/**
 * This will load the DLL from the configuration.  If it's already been initialized, it should just return.
 **/
static int win_loader_init(void) {
	HMODULE dllhandle;
	InitializationProcedure init_func;
	char function_name[256];
	char *params = NULL, *bfpwcrypt_path = NULL;
	char *module_name=NULL, *module_from_conf=NULL, *sep=NULL, *temp=NULL, *word=NULL, *lastword=NULL;
	
	__try { 
		if (DEBUG) {
			BFTRACE("Loading password encryption module.");
		}
		if (!initialized) {
			if (NULL != (params = bfconf_get(PASSWORD_ENCRYPT_MODULE_PROPERTY, NULL))) {

				/* Remove the path to get just "split" */
				temp = strdup(params);
				sep = ";";
				module_from_conf = strtok(temp, sep);
				bfpwcrypt_path = strtok(NULL, sep);
				if (DEBUG && NULL != bfpwcrypt_path) {
					activity_write(BFHERE "bfpwcrypt_path: %s", bfpwcrypt_path);
				}
	
				/* remove the path to get just "module.dll" */
				sep = "\\/:;=-";
				temp = strdup(module_from_conf);
				for (word = strtok(temp, sep); word; word = strtok(NULL, sep))
					lastword = word;
				
				/* now remove ".dll" from the end. */
				module_name = strtok(strdup(lastword), ".");
				if (DEBUG && NULL != module_name) {
					activity_write(BFHERE "module: %s", module_name);
				}

				/* this allows us to load "module_init" per the struct design. */
				strncpy(function_name, module_name, (sizeof(function_name) - 1));
				function_name[sizeof(function_name)-1] = '\0';
				strncat(function_name, "_init", 5);
				
				if (DEBUG) {
					activity_write(BFHERE "Loading module: %s", module_from_conf);
					if (NULL != strstr(module_from_conf, "bfcrypt.dll")) {
						BFTRACE("Debugging info for bfcrypt.dll password decrypt operations can be found in bfcrypt.txt");
					}
				}

				/* load the library and initialize the function. */
				dllhandle = GetModuleHandle(module_from_conf);
				if (dllhandle == NULL) {
					dllhandle = LoadLibrary(module_from_conf);
				}

				if (dllhandle != NULL) {
					init_func = (InitializationProcedure)GetProcAddress(dllhandle, function_name);

					if (init_func == NULL) {
						send_msg("PWCrypt", "PWCryptFailedInitializingFunction", "si", function_name, GetLastError());
						if (DEBUG) {
							activity_write(BFHERE "Failed initializing the function %s.  Error = %d", function_name, GetLastError());
						}
						FreeLibrary(dllhandle);
						return GetLastError();
					}
				}
				else {
					send_msg("PWCrypt", "PWCryptFailedInitializingDLL", "si", module_from_conf, GetLastError());
					if (DEBUG) {
						activity_write(BFHERE "Failed loading DLL, error code = %d", GetLastError());
					}
					return GetLastError();
				}

				if (DEBUG) {
					activity_write(BFHERE "Loading procedure %s.", function_name);
				}

				/* Call the function */
				myops = init_func(bfpwcrypt_path);
				if (NULL == myops) {
					if (DEBUG) {
						activity_write(BFHERE "Error loading function: %d.", GetLastError());
						if (strcmp(module_from_conf, "bfcrypt.dll") == 0) {
							BFTRACE("Ensure bfpwcrypt.conf has been added to the agent directory.  Search documentation for more information about bfpwcrypt.conf.");
						}
					}
					FreeLibrary(dllhandle);
					return GetLastError();
				}

				initialized = 1;
			} else {
				if (DEBUG) {
					activity_write(BFHERE "Password encryption property %s is not configured.", PASSWORD_ENCRYPT_MODULE_PROPERTY);
				}
			}
			if (module_from_conf != NULL) free (module_from_conf);
		}
	} __except(filter(GetExceptionCode(), GetExceptionInformation())) { 
		if (DEBUG) { BFTRACE("in except."); }
	} 
	return 0;
}
#endif


static char *formCipherText(const char *module, char *info, unsigned char *cipherdata, int cipherdata_length) {
	char *buffer, *p, *b64;
	size_t module_length = strlen(module);
	size_t info_length = (NULL != info) ? strlen(info) : 0;
	size_t b64_length, total_length;
	char *cryptDebugFlag = NULL;
	int cryptDebug = 0;
	size_t cpy_count = 0;

	cryptDebugFlag = getenv( BFDEBUG_SECURITY_CRYPT );
	if( NULL != cryptDebugFlag ){
		cryptDebug = 1;
	}

	if(cryptDebug){
		printf("Cipher length: %d\n", cipherdata_length);
	}

	if (DEBUG) {
		activity_write(BFHERE "Cipher length: %d", cipherdata_length);
	}

	b64 = base64Encode(cipherdata, cipherdata_length);
	if (NULL == b64) {
		if (DEBUG) { BFTRACE("formCipherText: Base-64 encoder failed"); }
		return NULL;
	}
	
	if(cryptDebug){
		printf( "Cipher string after encoding: %s\n", b64);
	}
	if (DEBUG) { activity_write(BFHERE "Cipher string after encoding: %s", b64); }

	total_length = strlen(module);
	if (NULL != info) { total_length += strlen(info); }
	b64_length = strlen(b64);
	total_length += b64_length;
	total_length += 8;  /* {, :, }, and \0 with room to spare */

	buffer = calloc(1, total_length);
	if (NULL == buffer) {
		if (DEBUG) { BFTRACE("formCipherText: out of memory"); }
		return NULL;
	}

	p = buffer;
	*p++ = '{';
	memcpy(p, module, module_length);
	p = &p[(int)module_length];
	if (NULL != info) {
		*p++ = ':';
		memcpy(p, info, info_length);
		p = &p[(int)info_length];
	}


	*p++ = '}';
	cpy_count = (b64_length < BFBUFSIZ) ? (b64_length + 1) : BFBUFSIZ;
	strncpy(p, b64, cpy_count);

	free(b64);

    if(cryptDebug ){
		printf("current content:[%s]     TotalLength:[%d] B64Length:[%d]\n",buffer,total_length,b64_length);
	}

	if (DEBUG) { activity_write(BFHERE "Header: %s", buffer); }
	return buffer;
}

/**
 * This method will use the plaintext as input and either encode or encrypt
 * the data specified.   It will encrypt the plaintext if the password
 * encryption module is defined by the password_encrypt setting in the
 * agent's bfagent.conf file (see decrypt_pwd for more information).
 * The ciphertext result returned should be freed after use.
 **/
int encrypt_pwd(const char *plaintext, char **ciphertext) {
	const char *module = NULL;
	char *info = NULL,*sha_params = NULL;
	unsigned char *cipherdata = NULL;
	int cipherdata_length = 0;
	int rc = 0;

	if (!initialized) { 
#ifdef WIN32
		win_loader_init(); 
#elif !defined(ZOS)
	unix_loader_init();
#endif
	}

	/* if the password encrypt module is not initialized, use bf_encrypt to encode passwords. */
	if (NULL == myops) {
		BFTRACE("encrypt_pwd: bf_encrypt");
		*ciphertext = bf_encrypt(plaintext);
		if (NULL != *ciphertext) { return 0; }
		BFTRACE("encrypt_pwd: bf_encrypt failed");
		return -1;
	}
	
	/* call out to the encryption DLL */
	sha_params = bfconf_get(DIGEST_ALGORITHM, NULL);
	if (NULL == sha_params) {
		activity_write(BFHERE "digest_algorithm is not set, will use SHA1");
	}else{
		activity_write(BFHERE "digest_algorithm: [%s]", sha_params);
	}


	rc = myops->encrypt_password(plaintext, &cipherdata, &cipherdata_length, &info,sha_params);
	if (0 != rc) {
		activity_write(BFHERE "encrypt_pwd: encryption failed: rc=%d", rc);
		return rc;
	}
	
	if (ciphertext != NULL) {
		module = myops->getName();
		if (module == NULL || '\0' == *module) { module = "bfcrypt"; }
		*ciphertext = formCipherText(module, info, cipherdata, cipherdata_length);
		if (NULL == *ciphertext) {
			BFTRACE("encrypt_pwd: formCipherText failed (out of memory?)");
			rc = -1;
		}
	}

	if (NULL != info) { free(info); }
	if (NULL != cipherdata) { free(cipherdata); }
	return rc;
}


static void freeEncryptedPassword(
		char **pModule,
		char **pInfo,
		unsigned char **pData)
{
	if (NULL != *pModule) {
		free(*pModule);
		*pModule = NULL;
	}
	if (NULL != *pInfo) {
		free(*pInfo);
		*pInfo = NULL;
	}
	if (NULL != *pData) {
		free(*pData);
		*pData = NULL;
	}
}


/**
 * Splits an encrypted password "{module_name:info}cipherdatabase64" into its
 * various parts.  Each of these is independently malloc'd, and it is up to
 * the caller to free them.
 **/
static int parseEncryptedPassword(
		const char *ciphertext,
		char **pModule,
		char **pInfo,
		unsigned char **pData,
		int *pDataLen)
{
	const char *leftBrace, *rightBrace;
	char *colon;
	size_t len;

	/* Clean these out by default so we can handle error cases safely */
	*pModule = NULL;
	*pInfo = NULL;
	*pData = NULL;
	*pDataLen = 0;

	/* If any of these fail, then this is not the correct format for an
	 * encrypted password.
	 */
	if (NULL == ciphertext) { return 0; }
	leftBrace = strchr(ciphertext, '{');
	if (NULL == leftBrace) { return 0; }
	rightBrace = strchr(leftBrace, '}');
	if (NULL == rightBrace) { return 0; }
	if ('\0' == rightBrace[1]) { return 0; }

	/* Make a copy of module_name:info */
	len = (size_t)(rightBrace - leftBrace);
	*pModule = malloc((int)len);
	if (NULL == *pModule) {
		BFTRACE("parseEncryptedPassword: module name: out of memory");
		goto fail;
	}
	if (len > 0) { memcpy(*pModule, &leftBrace[1], len-1); }
	(*pModule)[(int)len] = '\0';

	/* The :info part is optional. */
	colon = strchr(*pModule, ':');
	if (NULL != colon) {
		*colon = '\0';
		*pInfo = strdup(&colon[1]);
		if (NULL == *pInfo) {
			BFTRACE("parseEncryptedPassword: info: out of memory");
			goto fail;
		}
	}

	*pData = base64Decode(&rightBrace[1], pDataLen);
	if (NULL != *pData) { return 1; }
	BFTRACE("parseEncryptedPassword: base64Decode failed");

fail:
	freeEncryptedPassword(pModule, pInfo, pData);
	return 0;
}


/**
 * Make sure the implementation name exists and matches the name returned
 * by the crypto DLL.  If it does not match, it's not going to be able to
 * decrypt it.
 **/
static int isValidModuleName(const char *module_name) {
	if (NULL == module_name || '\0' == *module_name || NULL == myops) { return 0; }
	if (0 == strcmp(myops->getName(), module_name)) { return 1; }
	if (0 == strcmp("UNKNOWN", module_name)) { return 1; }
	return 0;
}


/**
 * This method will use the ciphertext to determine if it's plaintext, encoded
 * or encrypted.   If encrypted, then we need to have an encryption DLL from
 * the password_encrypt setting in bfagent.conf.  It's recommended to fully
 * qualify the path to ensure loading the correct module for security reasons.
 * If encoded, then we will call bf_decrypt.  If plaintext, it will return the
 * input after copying it to plaintext.  The plaintext result returned should
 * be freed after use.
 **/
int decrypt_pwd(char **plaintext, const char *ciphertext, const int sha_back) {
	char *module_name = NULL;
	char *info = NULL,*sha_params = NULL;
	unsigned char *cipherdata = NULL;
	int cipherdata_len = 0;
	int rc = 0;


	if (!initialized) { 
#ifdef WIN32
		win_loader_init(); 
#elif !defined(ZOS)
		unix_loader_init();
#endif
	}

	if (!parseEncryptedPassword(ciphertext, &module_name, &info, &cipherdata, &cipherdata_len)) {
		*plaintext = bf_decrypt(ciphertext);
		if (NULL != *plaintext) {
			BFTRACE("Password decoded.");
			return 0;
		}
		*plaintext = strdup(ciphertext);
		BFTRACE("Password already in plaintext form.");
		return 0;
	}

	if (isValidModuleName(module_name)) {
		if (1 != sha_back) {
			sha_params = bfconf_get(DIGEST_ALGORITHM, NULL);
			if (NULL == sha_params) {
				activity_write(BFHERE "digest_algorithm is not set, will use SHA1");
			} else {
				activity_write(BFHERE "digest_algorithm: [%s]", sha_params);
				if (2 == sha_back) {
					sha_params = NULL;
				}
			}
			activity_write(BFHERE "Final passed digest_algorithm : [%s]", sha_params);
			rc = myops->decrypt_password(plaintext, cipherdata, cipherdata_len, info, sha_params);
		} else {
			activity_write(BFHERE "Use SHA1 because SHA2 failed");
			rc = myops->decrypt_password(plaintext, cipherdata, cipherdata_len, info, "SHA1back");
		}
		if (0 != rc) {
			BFTRACE("Password decryption module returned non-zero rc");
		} else {
			BFTRACE("Password decrypted.");
		}
	} else {
		BFTRACE("Encrypted password uses an unsupported module name");
		*plaintext = NULL;
		rc = -1;
	}
	freeEncryptedPassword(&module_name, &info, &cipherdata);
	return rc;
}

