/*
Copyright (C)2003-2011, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

/* Despite the name of this file, it has nothing to do with encryption.
 * This is merely an obfuscation algorithm, and nothing generated by
 * the bf_encrypt function should be treated as secure.  For a full
 * description of the data format, see the static method bf_decrypt(String)
 * in the JavaDocs for the com.buildforge.services.common.text.TextUtils
 * class in the Java API.
 *
 * This file also exposes the OpenSSL library's Base-64 implementation.
 */

#define	PBUFSIZE	512
#define	KEYLEN	8
#define	MINLEN	16

#undef D
#ifdef CRYPT_DEBUG
#  define D(x) x
# else
#  define D(x) /* ignore */
#endif /* CRYPT_DEBUG/! */

#ifdef ZOS

extern int encryptmode;
#endif

/* returns a pointer to an allocated string with the decoded string */
char *bf_decrypt(const char *ct) {
	int i, len, ctlen;
	unsigned int value, offset;
	const char *p;
	char ptb[PBUFSIZE], ctb[PBUFSIZE];

	if (ct == NULL) {
		D(printf("bf_decrypt: null ciphertext\n"));
		return NULL;
	}
	ctlen = (int)(strlen(ct) >> 1);
	if (ctlen < 24 || ctlen >= PBUFSIZE - 1) {
		D(printf("bf_decrypt: bad ctlen=%d\n", ctlen));
		return NULL;
	}
	for (i=0, p=ct; i<ctlen; i++, p+=2) {
		sscanf(p, "%02x", &value);
		ctb[i] = (char)value;
	}

	len = 255 - (ctb[8] & 0xFF);
	if (len != 255) {
		if (ctlen < 9 + len) {
			D(printf("bf_decrypt: !UTF8 ctlen=%d, but len=%d\n", ctlen, len));
			return NULL;
		}
		for (i=0; i<len; i++) {
			value = ctb[i+9] & 0xFF;
			if (value < 0x20 || value >= 0x7F) {
				D(printf("bf_decrypt: !UTF8 value=%u\n", value));
				return NULL;
			}
			offset = value + 0x011D - (ctb[i & 7] & 0xFF);
			ptb[i] = (char)(offset % 0x5F + 0x20);
		}
	} else {
		if (ctlen < 25) {
			D(printf("bf_decypt: UTF8 bad ctlen=%d\n", ctlen));
			return NULL;
		}
		len = 255 - (ctb[9] & 0xFF);
		if (ctlen < 10 + len) {
			D(printf("bf_decrypt: UTF8 ctlen=%d, but len=%d\n", ctlen, len));
			return NULL;
		}
		for (i=0; i<len; i++) {
			ptb[i] = (char)((ctb[i + 10] ^ ctb[i & 7]) & 0xFF);
			if (ptb[i] == '\0') {
				D(printf("bf_decrypt: UTF8 decoded to NUL at i=%d\n", i));
				return NULL;
			}
		}
	}

	ptb[len] = '\0';
#ifdef ZOS
	if (!encryptmode)
		convert_atoe_l(ptb, len);	
#endif
	D(printf("bf_decrypt: decoded to [%s]\n", ptb));
	return strdup(ptb);
}

static const char *HEX = "0123456789abcdef";
char *bf_encrypt(const char *pt) {
	unsigned char k[8];
	unsigned char *buf, *p;
	int i, x;
	size_t len = strlen(pt);
#ifdef ZOS
	char *zp;
#endif

	if (len > 116) { return NULL; }
	p = buf = malloc((len << 1) + 64);

#ifdef ZOS
	zp = malloc(len + 1);
	memcpy(zp, pt, len + 1);
	pt = zp;
#endif

	/* seed rand() */
#ifdef HAVE_SSL
	RAND_pseudo_bytes((unsigned char *)&x, sizeof(int));
# else /* !HAVE_SSL */
	x = (time(NULL) + getpid()) ^ rand();
#endif /* HAVE_SSL/! */
	srand(x);

	for (i=0; i<8; ++i) {
		k[i] = (unsigned char)rand();
		*p++ = HEX[(k[i] >> 4) & 0x0F];
		*p++ = HEX[k[i] & 0x0F];
	}
	*p++ = '0';
	*p++ = '0';
	x = 255 - (int)len;
	*p++ = HEX[(x >> 4) & 0x0F];
	*p++ = HEX[x & 0x0F];

	for (i=0; i<(int)len; ++i) {
		int x = *pt++ ^ k[i & 7];
		*p++ = HEX[(x >> 4) & 0x0F];
		*p++ = HEX[x & 0x0F];
	}

	for (i=(int)len; i<16; ++i) {
		x = (int)rand();
		*p++ = HEX[(x >> 4) & 0x0F];
		*p++ = HEX[x & 0x0F];
	}

	*p = '\0';

#ifdef ZOS
	free(zp);
#endif
	return (char *)buf;
}

unsigned char *base64Decode(const char* data, int* data_length) {
#ifdef HAVE_SSL
    BIO *mem, *b64;
    int length = 512;
    unsigned char *buffer = NULL;
	int ok = 1;

	buffer = (unsigned char *)malloc(length);
    memset(buffer, 0, length);
    b64 = BIO_new(BIO_f_base64());
	BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
    mem = BIO_new_mem_buf((char*)data, strlen(data));
	mem = BIO_push(b64, mem);
	memset(buffer, 0, length);
	*data_length = BIO_read(mem, buffer, length);
	if (BIO_flush(mem) <= 0) { ok = 0; }
	BIO_free_all(mem);
	if (ok) { return buffer; }
	free(buffer);
#endif /* HAVE_SSL */
	return NULL;
}

char *base64Encode(const unsigned char* data, const int data_length) {
#ifdef HAVE_SSL
    BIO *mem, *b64;
    char* outbuff;
	int output_size;
	int ok = 1;
	
    outbuff = (char *)malloc(data_length);
    memset(outbuff, 0, data_length);
	b64 = BIO_new(BIO_f_base64());
    mem = BIO_new(BIO_s_mem());
    BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
	mem = BIO_push(b64, mem);
	BIO_write(mem, data, data_length);
    if (BIO_flush(mem) <= 0) { ok = 0; }
	output_size = BIO_get_mem_data(mem, &outbuff);
	if (BIO_set_close(mem, BIO_NOCLOSE) <= 0) { ok = 0; }
	outbuff[output_size] = '\0';
	BIO_free_all(mem);
	if (ok) { return outbuff; }
	free(outbuff);
#endif /* HAVE_SSL */
	return NULL;
}


#ifdef CRYPT_TEST
int main(int argc, char *argv[]) {
	char *p = bf_decrypt((argc >= 2) ? argv[1] : NULL);
	if (NULL == p) { return 1; }
	printf("%s\n", p);
	free(p);
	return 0;
}
#endif


