/*
Copyright (C)2003-2013, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

/*
 * This file implements the -s/-x command line options for the agent.
 * In other words, it places the agent into the background and listens
 * on the port specified in bfagent.conf for connections and does a
 * fork/exec to provide agent functionality to each such connection it
 * receives.
 *
 * This is simply not possible on Windows, because Windows doesn't have
 * fork().  The closest equivalent in that world would be to create a
 * new thread instead, but 1) this has significant security problems,
 * and 2) we currently do things like change the current working directory
 * and the environment of the agent process, which would not be safe to
 * do in a threaded environment.  Until these are addressed, we must
 * continue to use the current model based on bfdispatch for Windows.
 */

#include "bf_agent.h"

static bf_listener *listener = NULL;
int bfagent(void);

static void bfperror(const char *s, int err) {
	printf("[%d] bfdaemon: %s: %s\n",
			(int)getpid(), s, strerror(err) );
}

static void orphan(void) {
	pid_t pid = fork();
	switch (pid) {
		case -1:
			bfperror("fork", errno);
			return;
		case 0:
			usleep(100000);
			return;
	}
	BFTRACE("--- EXITING ---  (create orphan)");
	exit(0);
}

static void background(int bf_daemon) {
	if (bf_daemon == 2) {
		BFTRACE("background: skipped (Run with -x command line option)");
	} else if (NULL == activity_file || BFCONF_EXISTS(activity_log)) {
		BFTRACE("background: orphaning daemon");
		orphan();
		orphan();
		setsid();
	} else {
		BFTRACE("background: skipped (debugging on stderr)");
	}
	close(0);
	close(1);
	close(2);
}


/* Run the agent with I/O redirected to the specified fd. */
static int start_agent(int fd) {
	int rc;

	/* Release SIGCHLD trap; otherwise, grantpt(2) could fail. */
	platform_signals_release();

	/* Redirect I/O, being careful in case we're in the background
	 * and the socket ended up in one of the stdio slots.
	 */
	if (fd != 0) { dup2(fd, 0); }
	if (fd != 1) { dup2(fd, 1); }
	if (fd != 2) { dup2(fd, 2); }
	if (fd > 2) { close(fd); }

	/* Act like a normal agent process... */
	bfconf_refresh();
	rc = bfagent();
	shutdown(0, SHUT_RDWR);
	return rc;
}

/* Poll for incoming connections, accepting them as we see them... */
static void accept_connection(int sock, int bf_daemon) {
	char peeraddr[512];
	char peerport[32];
	struct linger lopts;
	struct sockaddr_storage peer;
	socklen_t peerlen;
	pid_t pid;
	int rc;

	if (sock < 0) { return; }

	peerlen = sizeof(peer);
	if (getpeername(sock, (struct sockaddr *)&peer, &peerlen) < 0) {
		rc = errno;
		snprintf(peeraddr, sizeof(peeraddr), "???:err=%s", strerror(rc));
		snprintf(peerport, sizeof(peerport), "???:err=%d", rc);
	} else {
		rc = getnameinfo((struct sockaddr *)&peer, peerlen,
				peeraddr, sizeof(peeraddr),
				peerport, sizeof(peerport),
				NI_NUMERICHOST | NI_NUMERICSERV);
		if (0 != rc) {
			snprintf(peeraddr, sizeof(peeraddr), "???:err=%s", gai_strerror(rc));
			snprintf(peerport, sizeof(peerport), "???:err=%d", rc);
		}
	}

	if (!checkIP(peeraddr)) {
		activity_write(BFHERE
				"[%s/%d] rejecting [%s/%s] (not in allow_ip)",
				listener->addr, listener->port,
				peeraddr, peerport);
		close(sock);
		return;
	}

	if (NULL != activity_file) {
		char peername[NI_MAXHOST];
		rc = getnameinfo((struct sockaddr *)&peer, peerlen,
				peername, sizeof(peername),
				NULL, 0, 0);
		if (0 != rc) { snprintf(peername, sizeof(peername), "???:err=%d:%s", rc, gai_strerror(rc)); }
		activity_write(BFHERE "[%s/%d] accepting [%s/%s] (%s)",
				listener->addr, listener->port,
				peeraddr, peerport, peername);
	}

	lopts.l_onoff = 1;
	lopts.l_linger = 20;
	setsockopt(sock, SOL_SOCKET, SO_LINGER, (char *)&lopts, sizeof(lopts));

	pid = fork();
	switch (pid) {
		case -1:
			activity_write(BFHERE "fork: %d: %s", errno, strerror(errno));
			close(sock);
			return;

		case 0:
			/* Child process; run the agent with redirected I/O ... */
			BFTRACE("=== NEW AGENT ===");
			bf_listener_destroy(listener);
			exit(start_agent(sock));
			return;
	}

	/* Parent process; continue polling ... */
	activity_write(BFHERE "[%s/%s]: attached to pid %d", peeraddr, peerport, (int)pid);
	close(sock);
}



void bfdaemon(int bf_daemon) {
	int port = 5555;
	char *p;

	p = bfconf_get("port", NULL);
	if (NULL != p) {
		port = atoi(p);
		if (port <= 0 || port > 65535) {
			BFTRACE("BFCONF port setting is invalid; falling back on default");
			port = 5555;
		}
		free(p);
	}

	p = bfconf_get("bind", NULL);
	listener = bf_listener_new(p, port, SOMAXCONN);
	if (NULL == listener) {
		bfperror("bf_listener_new", errno);
		free(p);
		return;
	}
	free(p);

	/* Time to go into the background... */
	background(bf_daemon);
	platform_signals_catch("bfdaemon");
    for (;;) {
		int sock = bf_listener_accept(listener, -1 /* Doesn't time out */);
		if (sock >= 0) { accept_connection(sock, bf_daemon); }
	}
}


