/*
Copyright (C)2008-2013, IBM, Inc. All rights reserved.
IBM, Inc. owns the copyright and other intellectual
property rights in this software. Duplication or use of the
Software is not permitted except as expressly provided in a
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

/* JCL support for z/OS */

static iconv_t cdToUTF8;
static iconv_t cdFromUTF8;
static int jmsock;
static int DBG = 0;
static time_t timeout;

extern char *auth_user;
extern char *auth_pass;

#define B64BUFFSIZE(in) ((in/3 + 1) * 4 + 1)

static int closeConverters() {
	BFTRACE("Closing iconv converters");
	iconv_close(cdToUTF8);
	iconv_close(cdFromUTF8);
	return 0;
}

static void jm_b64encode_chars(unsigned char in[3], unsigned char out[4],
		int numchars) {
	static char b64chars[] =
			"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
	unsigned char equal = '=';
	convert_etoa((char*) in);
	out[0] = b64chars[in[0] >> 2];
	out[1] = b64chars[((in[0] & 0x03) << 4) | (in[1] >> 4)];
	out[2] = (numchars > 1) ? b64chars[((in[1] & 0x0f) << 2) | (in[2] >> 6)]
			: equal;
	out[3] = (numchars > 2) ? b64chars[in[2] & 0x3f] : equal;
}

static time_t calcTimeout(time_t* now){
	time_t timed_out;
	time(now);
	timed_out = *now + timeout;
	return timed_out;
}

static int openConverters() {
	char* codeset;
	codeset = nl_langinfo(CODESET);
	activity_write(BFHERE"Codeset is %s ", codeset);
	if( (cdToUTF8 = iconv_open("UTF-8",codeset)) == (iconv_t)(-1)) {
		if( errno == EINVAL ){
			BFTRACE("An invalid pair for conversion was specified");
		}
		activity_write(BFHERE "errno from openConverters to UTF-8: %i",errno);
		BFTRACE("Failed to open iconv to UTF-8 for jobmon");
		send_msg("SUBMITJCL", "JobMonIconv", NULL);
		return -1;
	}
	if( (cdFromUTF8 = iconv_open(codeset, "UTF-8")) == (iconv_t)(-1)){
		if( errno == EINVAL ){
			BFTRACE("An invalid pair for conversion was specified");
		}
		activity_write(BFHERE "errno from openConverters from UTF-8: %i",errno);
		BFTRACE("Failed to open iconv from UTF-8 for jobmon");
		send_msg("SUBMITJCL", "JobMonIconv", NULL);
		return -1;
	}
	if(NULL!= codeset){ free(codeset);}
	return 0;
}

static size_t jobmon_convert(jm_convert_t mode, char* in, size_t insize,
		char* out, size_t outsize) {
	iconv_t cd;
	size_t len, orig_out;
	char* o = out;
	switch (mode) {
	case JM_NATIVE_TO_UTF8:
		cd = cdToUTF8;
		break;
	case JM_UTF8_TO_NATIVE:
		cd = cdFromUTF8;
		break;
	default:
		cd = ICONV_INVALID;
	};
	if (cd == ICONV_INVALID) {
		BFTRACE("ICONV_INVALID, will try to reopen");
		closeConverters();
		openConverters();
	}
	orig_out = outsize;
	while (-1 == (len = iconv(cd, &in, &insize, &o, &outsize))) {
		if (errno == EILSEQ) {
			BFTRACE("skipping invalid char");
			++in;
			--insize;
			continue;
		}
		BFTRACE("Unknown conversion errror.");
		closeConverters();
		break;
	}
	len = orig_out - outsize;
	if (NULL != o) {
		o = NULL;
	}
	if(len ==0){
		BFTRACE("no chars converted in jobmon conv routine");
	}
	return len;
}

static int openSock() {
	struct addrinfo hints;
	struct addrinfo *ai;
	char* p;
	static char port[6];
	int rc;
	p = bfconf_get("job_monitor_port", JM_DEFAULT_PORT);
	strncpy(port, p, (sizeof(port) -1));
	port[sizeof(port)-1] = '\0';
	activity_write(BFHERE"Attempting JM connection with port %s",port);
	memset(&hints,0,sizeof(hints));
	hints.ai_flags = (1<<0);
	hints.ai_socktype = SOCK_STREAM;
	getaddrinfo("127.0.0.1",port,&hints,&ai);
	BFTRACE("Creating JM socket connection to localhost");
	jmsock = socket(ai->ai_family,ai->ai_socktype,ai->ai_protocol);
	if(jmsock == -1) {
		BFTRACE("jmsock create failed");
		send_msg("SUBMITJCL", "JobMonSock", "s",port);
		rc = -1;
		goto done;
	} else {
		rc = connect(jmsock, ai->ai_addr, ai->ai_addrlen);
		if(rc!=0) {
			BFTRACE("error opening jmsock");
			send_msg("SUBMITJCL", "JobMonSock", "s",port);
			rc = -1;
			goto done;
		}
	}
done:
	free(p);
	return rc;
}

static int connectToJM() {
	char* native_cmd;
	char* native_resp;
	char* transcode_cmd;
	char* transcode_resp;
	int sent, received;
	size_t inleft, outleft;
	time_t now, timed_out;
	int rc = 0;
	int i;
	char* codeset;

	timed_out = calcTimeout(&now);

	native_cmd = calloc(1, BFBUFSIZ);
	native_resp = calloc(1, BFBUFSIZ);
	transcode_cmd = calloc(1, BFBUFSIZ);
	transcode_resp = calloc(1, BFBUFSIZ);
	strncpy(native_cmd, JM_CONNECT, (strlen(JM_CONNECT) + 1));
	inleft = strlen(native_cmd);
	outleft = BFBUFSIZ;
	jobmon_convert(JM_NATIVE_TO_UTF8, native_cmd, inleft, transcode_cmd,
			outleft);
	sent = send(jmsock, transcode_cmd, strlen(transcode_cmd), 0);
	if (sent <= 0) {
		BFTRACE("Zero bytes sent from jm connect");
		rc = -1;
		goto done;
	}
	received = recv(jmsock, transcode_resp, BFBUFSIZ, 0);
	if (received <= 0) {
		BFTRACE("No bytes received after connect");
		rc = -1;
		goto done;
	}
	inleft = received;
	outleft = BFBUFSIZ;
	jobmon_convert(JM_UTF8_TO_NATIVE, transcode_resp, inleft, native_resp,
			outleft);
	i = strncmp(native_resp, JM_CONNECT_ACK, strlen(JM_CONNECT_ACK));
	if (i != 0) {
		BFTRACE("acknowledge from jm connect not received!");
		BFTRACE(native_resp);
		BFTRACE(JM_CONNECT_ACK);
		send_msg("SUBMITJCL", "JobMonBadAck", "s", JM_CONNECT_ACK);
		rc = -1;
	} else {
		BFTRACE("Job Monitor connection successful");
	}
	time(&now);
	if(timeout && now > timed_out){
		send_msg("SUBMITJCL","JobMonTimeOut","i",timeout);
	}	
done: 
	if (NULL != native_cmd) { free(native_cmd); }
	if (NULL != native_resp) { free(native_resp); }
	if (NULL != transcode_cmd) { free(transcode_cmd); }
	if (NULL != transcode_resp) { free(transcode_resp); }
	if (NULL != codeset) { free(codeset); }
	return rc;
}

static int submitDataset(const char* dsn, char* jobid) {
	char* native_cmd;
	char* native_resp;
	char* transcode_cmd;
	char* transcode_resp;
	char* token;
	int sent, received;
	size_t inleft, outleft;
	time_t now, timed_out;
	int rc = 0;
	int i;
	size_t cpy_count = 0;

	timed_out = calcTimeout(&now);

	native_cmd = calloc(1, BFBUFSIZ);
	native_resp = calloc(1, BFBUFSIZ);
	transcode_cmd = calloc(1, BFBUFSIZ);
	transcode_resp = calloc(1, BFBUFSIZ);
	strncpy(native_cmd, JM_SUBMIT_DS, (strlen(JM_SUBMIT_DS) + 1));
	cpy_count = (strlen(dsn) < BFBUFSIZ) ? strlen(dsn) : BFBUFSIZ;
	native_cmd = strncat(native_cmd, dsn, cpy_count);
	native_cmd = strncat(native_cmd, "\n", 1);
	activity_write(BFHERE"submitds command: %s",native_cmd);
	inleft = strlen(native_cmd);
	outleft = BFBUFSIZ;
	jobmon_convert(JM_NATIVE_TO_UTF8, native_cmd,
			inleft, transcode_cmd, outleft);
	sent = send(jmsock, transcode_cmd, strlen(transcode_cmd),0);
	if(sent <= 0) {
		BFTRACE("Zero bytes sent from jm submitds");
		rc=-1;
		goto done;
	}
	/* Get the repsonse from JM */
	received = recv(jmsock, transcode_resp, BFBUFSIZ,0);
	if(received <= 0) {
		BFTRACE("No bytes received after submitds");
		rc=-1;
		goto done;
	}
	inleft = received;
	outleft = BFBUFSIZ;
	jobmon_convert(JM_UTF8_TO_NATIVE, transcode_resp,
			inleft, native_resp,outleft);
	activity_write(BFHERE "submitds resp: %s",native_resp);
	/* 
	The repsonse is in format: 115 Job J0052724 accepted.
	we need the third token, the job id
	*/
	i = strncmp(native_resp, JM_SUBMIT_ACK_START, strlen(JM_SUBMIT_ACK_START));
	if(i != 0) {
		BFTRACE("starting acknowledge from jm submitds not received!");
		BFTRACE(native_resp);
		BFTRACE(JM_SUBMIT_ACK_START);
		send_msg("SUBMITJCL", "JobMonSubFail", "s", native_resp );
		send_msg("SUBMITJCL", "JobMonBadAck", "s", JM_SUBMIT_ACK_START );
		rc = -1;
		goto done;
	}
	token = strtok(native_resp, " ");
	i = 1;
	do {
		activity_write(BFHERE "i: %i token: %s", i, token);
		if(i==3) {
			cpy_count = (strlen(token) < BFBUFSIZ) ? (strlen(token) + 1) : BFBUFSIZ;
			jobid = strncpy(jobid, token, cpy_count);
			activity_write(BFHERE "jobid from JM is %s",jobid);
			BFTRACE("Job Monitor submission successful");
			send_msg("SUBMITJCL", "JobId", "s", jobid);
			rc = 0;
			goto done;
		}
		i++;
	}while(token = strtok(NULL, " "));
	time(&now);
	if(timeout && now > timed_out){
		send_msg("SUBMITJCL","JobMonTimeOut","i",timeout);
	}	
done:
	if (NULL != native_cmd) {free(native_cmd);}
	if (NULL != native_resp) {free(native_resp);}
	if (NULL != transcode_cmd) {free(transcode_cmd);}
	if (NULL != transcode_resp) {free(transcode_resp);}
	if (NULL != token) {free(token);}
	return rc;
}

static int submitInline(const char* jcl, char* jobid) {
	char* native_cmd;
	char* native_resp;
	char* transcode_cmd;
	char* transcode_resp;
	char* transcode_line;
	char* token;
	char utf8_newline = 0x0A;
	int sent, received;
	time_t now, timed_out;
	size_t inleft, outleft;
	int rc = 0;
	int i;
	size_t cpy_count = 0;

	timed_out = calcTimeout(&now);

	native_cmd = calloc(1, BFBUFSIZ);
	native_resp = calloc(1, BFBUFSIZ);
	transcode_cmd = calloc(1, BFBUFSIZ);
	transcode_resp = calloc(1, BFBUFSIZ);
	transcode_line = calloc(1, BFBUFSIZ);
	strncpy(native_cmd, JM_SUBMIT, (strlen(JM_SUBMIT) + 1));
	activity_write(BFHERE "submit command: %s",native_cmd);
	inleft = strlen(native_cmd);
	outleft = BFBUFSIZ;
	jobmon_convert(JM_NATIVE_TO_UTF8, native_cmd,
			inleft, transcode_cmd, outleft);
	sent = send(jmsock, transcode_cmd, strlen(transcode_cmd),0);
	if(sent <= 0) {
		activity_write(BFHERE "error code from send: %i",errno);
		BFTRACE("Zero bytes sent from jm submit intial cmd");
		rc=-1;
		goto done;
	}
	/* Get the repsonse from JM */
	received = recv(jmsock, transcode_resp, BFBUFSIZ,0);
	if(received <= 0) {
		BFTRACE("No bytes received after submit initial command");
		rc=-1;
		goto done;
	}
	inleft = received;
	outleft = BFBUFSIZ;
	jobmon_convert(JM_UTF8_TO_NATIVE, transcode_resp,
			inleft, native_resp,outleft);
	activity_write(BFHERE "submit resp: %s",native_resp);
	/*
	The repsonse is 1 line prompting to start the flow of jcl
	*/
	i = strncmp(native_resp, JM_SUBMIT_START_RESP, strlen(JM_SUBMIT_START_RESP));
	if(i != 0) {
		BFTRACE("starting acknowledge from jm submit not received!");
		BFTRACE(native_resp);
		BFTRACE(JM_SUBMIT_START_RESP);
		send_msg("SUBMITJCL", "JobMonSubFail", "s", native_resp );
		send_msg("SUBMITJCL", "JobMonBadAck", "s", JM_SUBMIT_START_RESP);
		rc = -1;
		goto done;
	}
	cpy_count = (strlen(jcl) < BFBUFSIZ) ? (strlen(jcl) + 1) : BFBUFSIZ;
	native_cmd = strncpy(native_cmd, jcl, cpy_count);
	if(native_cmd[strlen(native_cmd)-1]!= '\n'){
		BFTRACE("appending new line to jcl");
		native_cmd = strncat(native_cmd,"\n",1);
	}
	native_cmd = strncat(native_cmd, JM_SUBMIT_EOF, strlen(JM_SUBMIT_EOF));
	inleft = strlen(native_cmd);
	outleft = BFBUFSIZ;
	jobmon_convert(JM_NATIVE_TO_UTF8, native_cmd,
					inleft, transcode_cmd, outleft);
	if(transcode_cmd[strlen(transcode_cmd)-1] != utf8_newline){
		transcode_cmd = strncat(transcode_cmd,&utf8_newline,1);
	}
	memset(transcode_line, '\0',BFBUFSIZ);
	for(i=0;transcode_cmd[i] !='\0';i++ ){
		strncat(transcode_line,&transcode_cmd[i],1);
		if(transcode_cmd[i] == utf8_newline){
			sent = send(jmsock, transcode_line, strlen(transcode_line),0);
			if(sent <= 0) {
				BFTRACE("Zero bytes sent from jm submit");
				rc=-1;
				goto done;
			}
			memset(transcode_line, '\0',BFBUFSIZ);
		}
	}
	/* Get the repsonse from JM */
	received = recv(jmsock, transcode_resp, BFBUFSIZ,0);
	if(received <= 0) {
		BFTRACE("No bytes received after submit");
		rc=-1;
		goto done;
	}
	inleft = received;
	outleft = BFBUFSIZ;
	jobmon_convert(JM_UTF8_TO_NATIVE, transcode_resp,
			inleft, native_resp,outleft);
	activity_write(BFHERE "submit resp: %s",native_resp);

	token = strtok(native_resp, " ");
	i = 1;
	do {
		activity_write(BFHERE "i: %i token: %s", i, token);
		if(i==3) {
			if(0 == strncmp("accepted",token,8)){ 
			  /* We didn't get a job id, means we sent bad JCL */ 
			  BFTRACE("Jobid not received, exiting"); 
			  send_msg("SUBMITJCL", "NoJESJobId", NULL); 
			  rc=-1; 
			  goto done; 
			} 
			cpy_count = (strlen(token) < BFBUFSIZ) ? (strlen(token) + 1) : BFBUFSIZ;
			jobid = strncpy(jobid, token, cpy_count);
			activity_write(BFHERE "jobid from JM is %s",jobid);
			BFTRACE("Job Monitor submission successful");
			send_msg("SUBMITJCL", "JobId", "s", jobid);
			rc = 0;
			goto done;
		}
		i++;
	}while(token = strtok(NULL, " "));
	time(&now);
	if(timeout && now > timed_out){
		send_msg("SUBMITJCL","JobMonTimeOut","i",timeout);
	}
done:
	if (NULL != native_cmd) {free(native_cmd);}
	if (NULL != native_resp) {free(native_resp);}
	if (NULL != transcode_cmd) {free(transcode_cmd);}
	if (NULL != transcode_resp) {free(transcode_resp);}
	if (NULL != token) {free(token);}
	return rc;
}

static int checkJobStatus(const char* jid, char* jesrc) {
	char* native_cmd;
	char* native_resp;
	char* native_buff;
	char* transcode_cmd;
	char* transcode_resp;
	char* token;
	char* rc_idx;
	char* msg_ptr;
	char* full_buff;
	int output_done = 0; /*This output block is complete*/
	int job_complete = 0; /*We have an rc from JES*/
	int sent, received;
	time_t now, timed_out;
	size_t inleft, outleft;
	int rc = -1;
	size_t cpy_count = 0;

	timed_out = calcTimeout(&now);
	/* Give the JM time to submit the job*/
	sleep(2);
	native_cmd = calloc(1, BFBUFSIZ);
	native_resp = calloc(1, BFBUFSIZ);
	native_buff = calloc(1, BFBUFSIZ);
	transcode_cmd = calloc(1, BFBUFSIZ);
	transcode_resp = calloc(1, BFBUFSIZ);
	full_buff = calloc(1, BFBUFSIZ);

	while (!job_complete) {
		time(&now);
	    if(timeout && now > timed_out){
    	    send_msg("SUBMITJCL","JobMonTimeOut","i",timeout);
			rc=-1;
			goto done;
	    }	
		sleep(5); /* TODO: should this be in bfagent.conf? as JM_POLLING_INTERVAL?*/
		output_done = 0;
		strncpy(native_cmd, JM_STATUS, (strlen(JM_STATUS) + 1));
		cpy_count = (strlen(jid) < BFBUFSIZ) ? strlen(jid) : BFBUFSIZ;
		native_cmd = strncat(native_cmd, jid, cpy_count);
		native_cmd = strncat(native_cmd, "\n", 1);
		activity_write(BFHERE "status command: %s",native_cmd);
		inleft = strlen(native_cmd);
		outleft = BFBUFSIZ;
		jobmon_convert(JM_NATIVE_TO_UTF8, native_cmd,
				inleft, transcode_cmd, outleft);
		sent = send(jmsock, transcode_cmd, strlen(transcode_cmd),0);
		if(sent <= 0) {
			BFTRACE("Zero bytes sent from jm status");
			rc = -1;
			goto done;
		}
		memset(full_buff,'\0',BFBUFSIZ);
		/* 
		Responses consist of a 3 digit code, then either a dash(-) or space
		a dash indicates additional output from the JM, space means
		we are all output_done
		*/
		while(!output_done) {
			/* Get the repsonse from JM */
			memset(transcode_resp, '\0', BFBUFSIZ);
			memset(native_resp, '\0', BFBUFSIZ);
			memset(native_buff, '\0', BFBUFSIZ);
			received = 0;
			received = recv(jmsock, transcode_resp, BFBUFSIZ,0);
			if(received <= 0) {
				BFTRACE("No bytes received after status");
				rc = -1;
				goto done;
			}
			inleft = received;
			outleft = BFBUFSIZ;
			jobmon_convert(JM_UTF8_TO_NATIVE, transcode_resp,
					inleft, native_resp, outleft);
			/* 
			Typically each call will return multiple lines, we want to
			tokenize them, and look for the end of output
			*/
			if(DBG){
				activity_write(BFHERE "status resp: %s",native_resp);
			}
			token = strtok(native_resp, "\n");
			do {
				msg_ptr = token +4;
				cpy_count = (strlen(msg_ptr) < BFBUFSIZ) ? strlen(msg_ptr) : BFBUFSIZ;
				full_buff = strncat(full_buff, msg_ptr, cpy_count);
				full_buff = strncat(full_buff, "\n", 1);
				/* Check for 155-Return.Code=0x00000000 to ensure the job is output_done */
				if(token[4] == 'R') {
					if( NULL != strstr(token, JM_RETURN_CODE) ) {
						rc_idx = strchr(token,'=');
						rc_idx++;
						cpy_count = (strlen(rc_idx) < BFBUFSIZ) ? (strlen(rc_idx) + 1) : BFBUFSIZ;
						jesrc = strncpy(jesrc, rc_idx, cpy_count);
						job_complete = 1;
						activity_write(BFHERE "jes rc:%s", jesrc);
						rc=0;
					}
				}
				if(token[3] == ' ') {
					output_done =1;
					break;
				}
			}while(token = strtok(NULL, "\n"));
			/*
			We have to process each line, to remove headers and check if
			the job is done.  Build a buffer, then loop once again to return
			line by line. yuck.
			*/
			if(job_complete == 1 && output_done == 1){
				send_msg("SUBMITJCL", "JobStatBeg", NULL);
				token = strtok(full_buff, "\n");
				do{
					send_msg("SUBMITJCL", "JobStat", "s", token);
				}while(token = strtok(NULL, "\n"));
			}
		}
	}/*end while(!job_complete)*/
done:
	if (NULL != native_cmd) {free(native_cmd);}
	if (NULL != native_resp) {free(native_resp);}
	if (NULL != native_buff) {free(native_buff);}
	if (NULL != transcode_cmd) {free(transcode_cmd);}
	if (NULL != transcode_resp) {free(transcode_resp);}
	if (NULL != token) {free(token);}
	if (NULL != rc_idx) {free(rc_idx);}
	if (NULL != msg_ptr) {free(msg_ptr);}
	if (NULL != full_buff) {free(full_buff);}
	return rc;
}

static int getSpoolOutput(char* jid) {
	char* native_cmd;
	char* native_resp;
	char* transcode_cmd;
	char* transcode_resp;
	char* resp_line;
	char* resp_ptr;
	int sent, received;
	size_t inleft, outleft;
	int done = 0;
	int rc = 0;
	int i;
	time_t now, timed_out;
	size_t cat_count = 0;

	timed_out = calcTimeout(&now);
	/* Give the JM 2 seconds after checking status to get spool output*/
	sleep(2);
	native_cmd = calloc(1, BFBUFSIZ);
	native_resp = calloc(1, BFBUFSIZ);
	resp_line = calloc(1, BFBUFSIZ / 4);
	transcode_cmd = calloc(1, BFBUFSIZ);
	transcode_resp = calloc(1, BFBUFSIZ);
	strncpy(native_cmd, JM_GET_SPOOL, (strlen(JM_GET_SPOOL) + 1));
	cat_count = (strlen(jid) < BFBUFSIZ) ? strlen(jid) : BFBUFSIZ;
	native_cmd = strncat(native_cmd, jid, cat_count);
	native_cmd = strncat(native_cmd, JM_GET_SPOOL_ALL, strlen(JM_GET_SPOOL_ALL));
	activity_write(BFHERE"getspool command: %s",native_cmd);
	inleft = strlen(native_cmd);
	outleft = BFBUFSIZ;
	jobmon_convert(JM_NATIVE_TO_UTF8, native_cmd,
			inleft, transcode_cmd, outleft);
	/* Responses consist of a 3 digit code, then either a dash(-) or space
	a dash indicates additional output from the JM, space means
	we are all done
	*/
	BFTRACE("About to send the getspool command");
	sent = send(jmsock, transcode_cmd, strlen(transcode_cmd),0);
	if(sent <= 0) {
		BFTRACE("Zero bytes sent from jm getspool");
		rc = -1;
		goto done;
	}
	while(!done) {
		time(&now);
	    if(timeout && now > timed_out){
	        send_msg("SUBMITJCL","JobMonTimeOut","i",timeout);
			rc=-1;
			goto done;
	    }

		sleep(1);
		/* Get the repsonse from JM */
		memset(transcode_resp, '\0', BFBUFSIZ);
		memset(native_resp, '\0', BFBUFSIZ);
		BFTRACE("About to receive spool output");
		received = recv(jmsock, transcode_resp, BFBUFSIZ,0);
		if(received <= 0) {
			BFTRACE("No bytes received after getspool");
			rc = -1;
			goto done;
		}
		BFTRACE("Transcoding spool output");
		activity_write(BFHERE "%i bytes received from spool", received);
		inleft = received;
		outleft = BFBUFSIZ;
		jobmon_convert(JM_UTF8_TO_NATIVE, transcode_resp,
				inleft, native_resp, outleft);
		/* Will dump the entire buffer to the log, only needed when deep debugging */
		if(DBG){
	  		activity_write(BFHERE "get spool resp: %s",native_resp);
		}
		/* Typically each call will return multiple lines, we will go char
 		 by char looking for new lines. Each new line triggers a send to
 		 the agent
		*/
		for(i=0; native_resp[i]!='\0' && i < received;i++) {
			strncat(resp_line, &native_resp[i], 1);
			if(native_resp[i] == '\n') {
				resp_ptr = resp_line + 4;
				send_msg("SUBMITJCL", "SpoolOut", "s", resp_ptr);
				if(resp_line[1] == '6' && resp_line[2] =='7' &&
						resp_line[3] == ' ')
				{
					done = 1;
				}
				memset(resp_line,'\0',BFBUFSIZ/4);
			}
		}
	}
done:
	if (NULL != native_cmd) {free(native_cmd);}
	if (NULL != native_resp) {free(native_resp);}
	if (NULL != transcode_cmd) {free(transcode_cmd);}
	if (NULL != transcode_resp) {free(transcode_resp);}
	if (NULL != resp_line) {free(resp_line);}
	if (NULL != resp_ptr) {free(resp_ptr);}
	return rc;
}

static int closeSock() {
	char* native_cmd;
	char* transcode_cmd;
	size_t inleft, outleft;
	int sent, received;
	int rc = 0;

	BFTRACE("Logging off JM");
	native_cmd = calloc(1, 10);
	transcode_cmd = calloc(1, 20);
	strncpy(native_cmd, JM_LOGOFF, (strlen(JM_LOGOFF) + 1));
	inleft = strlen(native_cmd);
	outleft = BFBUFSIZ;
	jobmon_convert(JM_NATIVE_TO_UTF8, native_cmd, inleft, transcode_cmd,
			outleft);
	sent = send(jmsock, transcode_cmd, strlen(transcode_cmd), 0);
	if (sent <= 0) {
		BFTRACE("Zero bytes sent from jm quit");
		rc = -1;
		goto done;
	}
	
	BFTRACE("Shutting down jobmon socket");
	rc = shutdown(jmsock, SHUT_RD);
	if (rc != 0) {
		BFTRACE("ERROR on jobmon socket shutdown!");
		rc = -1;
	}
	BFTRACE("Closing jobmon socket");
	rc = close(jmsock);
	if (rc != 0) {
		BFTRACE("ERROR on jobmon socket close!");
		rc = -1;
	}
done:
	if (NULL != native_cmd) {free(native_cmd);}
	if (NULL != transcode_cmd) {free(transcode_cmd);}
	
	return rc;
}

static int check_jes_rc(char* rc) {
	char* rc_ptr = rc;
	long rc_status; /*job termination status lianping*/
	long rc_maxcc; /*max cc if exists lianping*/
	char* status;
	char* abend; /*abend will stay as char value, since they are reported in hex anyway */
	char* maxcc;
        char* stopstr; /* lianping */

	/*RC is in the format, 0x000000000, remove the leading 0x*/
	activity_write(BFHERE"evaluating jesrc: %s",rc);
	rc_ptr = rc_ptr + 2;
	status = calloc(1, 3); /*2 -> 3 lianping */
	memcpy(status, rc_ptr, 2);
        rc_status = strtol(status,&stopstr,16);  /* lianping */
	activity_write(BFHERE"rc_status: %i", rc_status);
	rc_ptr = rc_ptr +2;
	abend = calloc(1,4); /* 3->4 lianping */
	memcpy(abend, rc_ptr, 3);
	activity_write(BFHERE "rc_abend : %s", abend);
	rc_ptr = rc_ptr +3;
	maxcc = calloc(1,4); /* 3->4 lianping */
	memcpy(maxcc, rc_ptr, 3);
        rc_maxcc = strtol(maxcc,&stopstr,16);  /* lianping */
	activity_write(BFHERE "rc_maxcc : %i", rc_maxcc);
	switch ( rc_status ) {
		case RC_ABEND:
		send_msg("SUBMITJCL", "JobAbend", "s", abend);
		rc_maxcc = 99;
		break;
		case RC_CANCEL:
		rc_maxcc = 98;
		send_msg("SUBMITJCL", "JobCancel", NULL);
		break;
		case RC_JCLERR:
		rc_maxcc = 97;
		send_msg("SUBMITJCL", "JCLError", NULL);
		break;
		case RC_NORMAL:
		send_msg("SUBMITJCL", "JobCompleted", "i", rc_maxcc);
		break;
		default:
		send_msg("SUBMITJCL", "JobRcInvalid", "isi", rc_status, abend,
				rc_maxcc);
		rc_maxcc =96;
		break;
	}
	if(NULL != rc_ptr) {free(rc_ptr);}
	if(NULL != status) {free(status);}
	if(NULL != abend) {free(abend);}
	if(NULL != maxcc) {free(maxcc);}
	return rc_maxcc;
}

/* Caller must free the returned memory */
static char *jm_b64encode(unsigned char *pw) {
	unsigned char *p;
	unsigned char b64in[4];
	unsigned char b64out[5];
	unsigned char *enc_pw;
	int i;
	int buffsize = 0;

	buffsize = B64BUFFSIZE(strlen((char *)pw));
	enc_pw = calloc(1, buffsize);
	memset(enc_pw, 0, buffsize);
	p = pw;
	memset(b64in, '\0', 4);
	memset(b64out, '\0', 5);
	i = 0;
	while (*p != '\0') {
		++i;
		b64in[i - 1] = *p;
		if (i == 3) {
			jm_b64encode_chars(b64in, b64out, i);
			enc_pw = (unsigned char*) strncat((char*) enc_pw, (char*) b64out, strlen(b64out));
			i = 0;
			memset(b64in, '\0', 4);
			memset(b64out, '\0', 5);
		}
		++p;
	}
	/*check for a final segment < 3 chars */
	if (i > 0) {
		jm_b64encode_chars(b64in, b64out, i);
		enc_pw = (unsigned char*) strncat((char*) enc_pw, (char*) b64out, strlen(b64out));
	}

	return (char *)enc_pw;
}

/* Obtains the username/password from either bfagent.conf(jcl_submit_user key) or creds used to 
   auth with the agent.
   The caller is responsible for freeing the provided username and password.
 */
static int get_jm_login(char **username, char **password) {
	char *p;
	char *cipher = NULL;
	char *jm_login = NULL;
	int rc = 0;

	*username = NULL;
	*password = NULL;

	jm_login = bfconf_get("jcl_submit_user", NULL);
	if (NULL == jm_login) {
		if (NULL != auth_user && NULL != auth_pass) {
			BFTRACE("No jcl_submit_user line found in conf: Using retained credentials");
	 		*username = strdup(auth_user);
			*password = strdup(auth_pass);
		} else {
			BFTRACE("No jcl_submit_user line found in conf");
			send_msg("SUBMITJCL", "JobMonNoUser", NULL);
			rc = -1;
			goto done;
		}
	} else {
		p = strrchr(jm_login, ':');
		if (NULL == p) {
		  /* This should probably be a different error code in the future */
			BFTRACE("No password found for jcl_submit_user line in conf");
			send_msg("SUBMITJCL", "JobMonNoUser", NULL);
			rc = -1;
			goto done;
		}

		*p++ = '\0';
		*username = strdup(jm_login);
		activity_write(BFHERE"username from file is: |%s|",*username);
		decrypt_pwd(password, p, 0);
		if(DBG){
			activity_write(BFHERE"rawlen %i",strlen(*password));
		}	
	}
done:
	if (NULL != jm_login) {free(jm_login);}
	return rc;
}

static int authToJM(const char* username, const char* password) {
	char* native_cmd = NULL;
	char* native_resp = NULL;
	char* transcode_cmd = NULL;
	char* transcode_resp = NULL;
	char* jm_ptr = NULL;
	char* jm_user = NULL;
	char* jm_pw = NULL;
	char* jm_pw_native = NULL;
	int sent, received;
	size_t inleft, outleft;
	int rc = 0;
	int i;
	size_t cat_count = 0;

	native_cmd = calloc(1, BFBUFSIZ);
	native_resp = calloc(1, BFBUFSIZ);
	transcode_cmd = calloc(1, BFBUFSIZ);
	transcode_resp = calloc(1, BFBUFSIZ);
	jm_user = calloc(1, JM_USER_LEN);
	jm_pw_native = calloc(1, JM_PW_LEN);

	if (NULL == username) {
		get_jm_login(&jm_user, &jm_pw);
	} else {
		jm_user = strdup(username);
		jm_pw = strdup(password);
	}

	jm_pw_native = jm_b64encode((unsigned char*)jm_pw);
	if(DBG){
		activity_write(BFHERE"b64len %i",strlen(jm_pw_native));
	}

	strncpy(native_cmd, JM_LOGON, (strlen(JM_LOGON) + 1));
	cat_count = (strlen(jm_user) < BFBUFSIZ) ? strlen(jm_user) : BFBUFSIZ;
	native_cmd = strncat(native_cmd, jm_user, cat_count);
	native_cmd = strncat(native_cmd, " ", 1);
	activity_write(BFHERE"Connect command: %s *****", native_cmd);
	cat_count = (strlen(jm_pw_native) < BFBUFSIZ) ? strlen(jm_pw_native) : BFBUFSIZ;
	native_cmd = strncat(native_cmd, jm_pw_native, cat_count);
	native_cmd = strncat(native_cmd, "\n", 1);
	activity_write(BFHERE"auth string num bytes %i", strlen(native_cmd));
	inleft = strlen(native_cmd);
	outleft = BFBUFSIZ;
	jobmon_convert(JM_NATIVE_TO_UTF8, native_cmd,
			inleft, transcode_cmd, outleft);
	sent = send(jmsock, transcode_cmd, strlen(transcode_cmd),0);
	if(sent <= 0) {
		BFTRACE("Zero bytes sent from jm auth");
	}
	activity_write(BFHERE"num bytes sent for auth %i", sent);
	/* Get the repsonse from JM */
	received = recv(jmsock, transcode_resp, BFBUFSIZ,0);
	if(received <= 0) {
		BFTRACE("No bytes received after auth");
	}
	inleft = received;
	outleft = BFBUFSIZ;
	jobmon_convert(JM_UTF8_TO_NATIVE, transcode_resp,
			inleft, native_resp,outleft);
	i = strncmp(native_resp, JM_LOGON_ACK_START, strlen(JM_LOGON_ACK_START));
	if(i != 0) {
		BFTRACE("starting acknowledge from jm auth not received!");
		BFTRACE(native_resp);
		BFTRACE(JM_LOGON_ACK_START);
		send_msg("SUBMITJCL", "JobMonAuthFail", "s", native_resp );
		send_msg("SUBMITJCL", "JobMonBadAck", "s", JM_LOGON_ACK_START );
		rc = -1;
		goto done;
	} else if(NULL == strstr(native_resp, JM_LOGON_ACK_END)) {
		BFTRACE("ending acknowledge from jm auth not received!");
		BFTRACE(native_resp);
		BFTRACE(JM_LOGON_ACK_END)
		send_msg("SUBMITJCL", "JobMonBadAck", "s", JM_LOGON_ACK_END);
		rc = -1;
		goto done;
	} else {
		BFTRACE("Job Monitor authentication successful");
	}
done:
	if (NULL != native_cmd) {free(native_cmd);}
	if (NULL != native_resp) {free(native_resp);}
	if (NULL != transcode_cmd) {free(transcode_cmd);}
	if (NULL != transcode_resp) {free(transcode_resp);}
	if (NULL != jm_ptr) {free(jm_ptr);}
	if (NULL != jm_user) {free(jm_user);}
	if (NULL != jm_pw) {free(jm_pw);}
	if (NULL != jm_pw_native) {free(jm_pw_native);}
	return rc;
}

/*
We should receive either a dsn or inline_jcl, to receive both is an error.
Note, username and password will both be either NULL or non-NULL
username - NULL or username to auth with
password - NULL or password to auth with
dsn - dataset name containing JCL to send to the job monitor
inline_jcl - buffer containing JCL to stream to the JM
*/
int jmSubmitJob(const char* username, const char* password, const char* dsn, const char* inline_jcl,time_t time_out) {
	int rc = 0;
	char* dataset = NULL;
	char* jcl = NULL;
	char* jobid;
	char* jesrc;
	char* p;
	time_t now, timed_out;
	size_t cpy_count = 0;

	timeout = time_out;

	p = getenv ( "BF_DEBUG_JOBMON" );
	if (p != NULL) {
		DBG=1;
	}

	if (NULL != dsn) {
		dataset = calloc(1, strlen(dsn) + 1);
		cpy_count = (strlen(dsn) < BFBUFSIZ) ? (strlen(dsn) + 1) : BFBUFSIZ;
		dataset = strncpy(dataset, dsn, cpy_count);
	}
	if (NULL != inline_jcl) {
			jcl = calloc(1, strlen(inline_jcl) + 1);
			cpy_count = (strlen(inline_jcl) < BFBUFSIZ) ? (strlen(inline_jcl) + 1) : BFBUFSIZ;
			jcl = strncpy(jcl, inline_jcl, cpy_count);
	}
	jobid = calloc(1, 20);
	jesrc = calloc(1, JM_RC_BUFFSIZE);

	if (0 != (rc = openConverters())) {
		activity_write(BFHERE"RC from open conv: %i",rc);
		goto done;
	}
	if(0 != (rc = openSock()) ) {
		activity_write(BFHERE "RC from open sock: %i",rc);
		goto done;
	}
	if(0 != (rc =connectToJM()) ) {
		activity_write(BFHERE "RC from conn: %i",rc);
		goto done;
	}
	if(0 != (rc = authToJM(username, password)) ) {
		activity_write(BFHERE "RC from auth: %i", rc);
		goto done;
	}
	if(NULL != dataset){
		BFTRACE("Submitting dataset");
		if(0 != (rc = submitDataset(dataset,jobid))) {
			activity_write(BFHERE "RC from submitjob: %i", rc);
			rc = -1;
			goto done;
		}
	}else if(NULL != jcl){
		BFTRACE("Submitting inline jcl");
		if(0 != (rc = submitInline(jcl,jobid))) {
			activity_write(BFHERE "RC from submitjob: %i", rc);
			rc = -1;
			goto done;
		}
	}
	if(0 != (rc = checkJobStatus(jobid, jesrc))) {
		activity_write(BFHERE "RC from checkjobstatus: %i", rc);
		rc = -1;
		goto done;
	}
	if(0 != (rc = getSpoolOutput(jobid))) {
		activity_write(BFHERE "RC from getspool: %i", rc);
		rc = -1;
		goto done;
	}

	rc = check_jes_rc(jesrc);
done:
	closeSock(jmsock);
	closeConverters();
	if(NULL != dataset) {free(dataset);}
	if(NULL != jobid) {free(jobid);}
	if(NULL != jesrc) {free(jesrc);}
	if(NULL != p) {free(p);}
	return rc;
}

/* Prints out a hex dump of the native input
BFTRACE("native_cmd hex");
for(i=0;native_cmd[i] != '\0'; i++){
	activity_write(BFHERE "%02X",(unsigned char)native_cmd[i]);
}
*/

