/*
Copyright (C)2003-2013, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"
#include "md5.h"

/* This file provides the program entry point, command line
 * argument parsing, and enforcement of some of the bfagent.conf
 * options, such as the "allow" setting.
 */
#ifdef WIN32
#define GETOPT_OPTIONS "e:df:vE:I:P:RS:"
#else
#define GETOPT_OPTIONS "e:df:sxvE:I:P:"
#endif /* WIN32 */

#ifdef WIN32
extern SOCKET hRemote = (SOCKET)INVALID_HANDLE_VALUE;
extern int agent_re_exec;
extern int win_sock;
#endif
extern int ofd;
extern int ifd;

extern int optind;
extern char *optarg;

int bfagent(void);
void bfdaemon(int);

const char *AGENT_VERSION = AGENT_ID_VERSION;

/* Implementation for -v */
static void report_version(void) {
#ifdef HAVE_SSL
	char *ssl_protocol = NULL;
#elif HAVE_SYSTEM_SSL /* HAVE_SSL */		
	char *gsk_ssl_protocol = NULL;
#endif /* HAVE_SYSTEM_SSL */
	printf("IBM Rational Build Forge Agent\n");
	if (*AGENT_VERSION == '_') {
		printf("Development Internal Build!\n");
	} else {
		char *ver = strdup(AGENT_VERSION);
		char *ifx = NULL;
		char *bld = NULL;
		int ifix = 0;

		if (NULL != (ifx = strchr(ver, '-'))) {
			*ifx++ = '\0';
			if (NULL != (bld = strchr(ifx, '-'))) {
				*bld++ = '\0';
				ifix = atoi(ifx);
			} else {
				bld = ifx;
			}
		}

		if (ifix > 0) {
			printf("%s iFix %03d", ver, ifix);
		} else {
			printf("%s GA", ver);
		}
		if (NULL != bld)
			printf(" (Build %s)", bld);
		printf("\n");
		free(ver);
	}
#ifdef WIN32
	printf("Platform: Win32\n");
#else /* !WIN32 */
	printf("Platform: Unix\n");
#endif /* WIN32/! */
#ifdef HAVE_SSL
	ssl_protocol = bfconf_get("ssl_protocol", "TLSv1");
	printf("SSL: %s\n", ssl_protocol);
#elif HAVE_SYSTEM_SSL /* HAVE_SSL */		
	gsk_ssl_protocol = bfconf_get("gsk_ssl_protocol", "TLSv1"); 
	printf("SYSTEM SSL: %s\n", gsk_ssl_protocol);
#endif /* WIN32 */
}

/* Implementation for -P */
static int report_smd5(const char *pass, size_t len) {
	int		ret;
	char	smd5[64];
	char	*p = strdup(pass);

#ifdef ZOS
	convert_etoa_l(p, len);
#endif /* ZOS */

	ret = smd5_create(smd5, (const md5_byte_t *)p, (int)len);
	free(p);
	if (ret != 0) {
		printf("!!! %d\n", ret);
	} else {
		printf("%s\n", smd5);
	}
	return ret;
}


#ifdef WIN32
/* The Windows equivalent of signal handling */
BOOL CtrlHandler(DWORD fdwCtrlType) {
	switch (fdwCtrlType) {
		case CTRL_C_EVENT:
			send_msg("EXEC", "WinEventCtrlC", NULL);
			return TRUE;
		case CTRL_CLOSE_EVENT:
			send_msg("EXEC", "WinEventClose", NULL);
			return TRUE;
		case CTRL_BREAK_EVENT:
			send_msg("EXEC", "WinEventBreak", NULL);
			return FALSE;
		case CTRL_SHUTDOWN_EVENT:
			send_msg("EXEC", "WinEventShutdown", NULL);
			return FALSE;
		case CTRL_LOGOFF_EVENT:
			send_msg("EXEC", "WinEventLogOff", NULL);
			return TRUE;
		default:
			return FALSE;
	} 
}
#endif /* WIN32 */

/* Implements the "allow" bfagent.conf setting */
static int addr_match(const char *addr, size_t addrlen, const char *maybe)
{
	size_t maybelen = strlen(maybe);

	/* Address string isn't long enough, so it can't possibly match. */
	if (maybelen > addrlen) { return 0; }

	/* Did not match the template from allow list. */
	if (0 != memcmp(addr, maybe, maybelen)) { return 0; }

	/* If the match address ends with a separator, then this is
	 * definitely a match; the edge case check below is not needed.
	 */
	switch(maybe[maybelen - 1]) {
		case '.':
		case ':':
		case '%':
			return 1;
	}

	/* Edge case check: If the match address doesn't have a trailing
	 * separator, then accept any separator as the next char in the
	 * address.  This prevents the allow entry 10.10.10 from matching
	 * 10.10.100.1 and so forth.
	 */
	switch (addr[maybelen]) {
		case '\0':
		case '.':
		case ':':
		case '%':
			return 1;
	}

	/* Failed the edge case check */
	return 0;
}

int checkIP(const char *addr) {
	char *  allow;
	char ** list;
	int     i;
	int     allowed = 0;
	size_t  addrlen;

	allow = bfconf_get("allow", NULL);
	if (NULL == allow) { return 1; }
	list = getArgs(allow);
	free(allow);
	if (NULL == list) { return 1; }

	if (NULL == *list) {
		allowed = 1;
	} else if ('\0' != *addr) {
		addrlen = strlen(addr);
		for (i=0; NULL != list[i]; ++i) {
			if (addr_match(addr, addrlen, list[i])) {
				allowed = 1;
				break;
			}
		}
	}

	freeArgs(list);
	return allowed;
}



#ifndef WIN32
/* Used by the agent update process, which no longer works cleanly in 7.x
 * due to the fact that the agent requires multiple files to be updated.
 */
static int isBusy(const char *updFile, int hasExt) {
	int retval = 0;
	size_t len = strlen(updFile);
	char *bsyFile = malloc(len + 8);

	memcpy(bsyFile, updFile, len + 1);
	strncpy(&bsyFile[hasExt ? (len-4) : len], ".bsy", 5);
	if (0 == access(bsyFile, 0)) { retval = 1; }
	free(bsyFile);
	return retval;
}

/*============================================================
 * Update logic for non-Windows agents.  Note that this code
 * operates entirely in the native character set, so it does
 * not need to use platform_xxx functions.
 *============================================================*/
static void handle_update_mode_agent(char *argv[]) {
	struct stat sb;
	char *updpath;
	char *binpath = NULL;
	char *oldpath = NULL;
	int err;
	mode_t mode;

	updpath = getenv("BF_AGENT_UPDATE");
	if (NULL == updpath || 0 != access(updpath,0) || isBusy(updpath,1)) { return; }
	if (agent_debug_flag) { printf("DEBUG: Agent is in update mode\n"); }

	/* build the list of pathnames */
	binpath = strdup(updpath);
	oldpath = strdup(updpath);
	strncpy(&oldpath[strlen(oldpath)-3], "old", 4);
	binpath[strlen(binpath)-4] = '\0';

	/* get the original mode bits */
	if (0 != stat(binpath, &sb)) {
		err = errno;
		activity_write(BFHERE "UPDATE stat(%s,sb): %d: %s", binpath, err, strerror(err));
		goto done;
	}

	/* move the current agent to .old */
	unlink(oldpath);
	if (0 != link(binpath, oldpath)) {
		err = errno;
		activity_write(BFHERE "UPDATE link(%s,%s): %d: %s", binpath, oldpath, err, strerror(err));
		goto done;
	}

	/* make sure that we own it */
	if (0 != chown(updpath, sb.st_uid, sb.st_gid)) {
		err = errno;
		activity_write(BFHERE "UPDATE chown(%s,%d,%d): %d: %s", updpath, sb.st_uid, sb.st_gid, err, strerror(err));
		goto done;
	}

	/* Fix perms, making sure the read and execute bits are set */
	mode = (sb.st_mode & 07777) | S_IXUSR | S_IRUSR;
	if (0 != chmod(updpath, mode)) {
		err = errno;
		activity_write(BFHERE "UPDATE chmod(%s,%04o): %d: %s", updpath, mode, err, strerror(err));
		goto done;
	}

	/* Replace the current agent with the update. */
	if (0 != unlink(binpath)) {
		err = errno;
		activity_write(BFHERE "UPDATE unlink(%s): %d: %s", binpath, err, strerror(err));
		goto done;
	}
	if (0 != link(updpath, binpath)) {
		/* uh-oh... disaster... */
		err = errno;
		activity_write(BFHERE "UPDATE link(%s,%s): %d: %s", updpath, binpath, err, strerror(err));
		if (agent_debug_flag) { printf("DEBUG: Catastrophic error during update attempt -- could not create new agent!\n"); }
		if (0 != link(oldpath, binpath)) {
			/* even worse! */
			err = errno;
			activity_write(BFHERE "UPDATE link(%s,%s): %d: %s", oldpath, binpath, err, strerror(err));
			if (agent_debug_flag) { printf("DEBUG: Catastrophic error after failed update -- could not restore old agent!\n"); }
		}
		goto done;
	}

	/* clean up the update and fire up the real agent */
	unlink(updpath);
	execv(binpath, argv);
	activity_write(BFHERE "UPDATE execv(%s, argv): %d: %s", binpath, err, strerror(err));
	if (agent_debug_flag) { printf("DEBUG: Catastrophic error during update attempt -- could not exec new agent!\n"); }

done:
	if (NULL != oldpath) { free(oldpath); }
	if (NULL != binpath) { free(binpath); }
}

static void handle_agent_update(char *argv[]) {
	char *path;
	char *binpath;
	char *updpath = NULL;
	size_t len;
	struct stat sb;
	int mode;

	handle_update_mode_agent(argv);

	binpath = bfconf_get("update_path", NULL);
	if (NULL == binpath) { return; }

	path = malloc(MAX_PATH);
	convert_utf8_to_native(binpath, CONVERT_ALL, path, MAX_PATH);
	path[MAX_PATH - 1] = '\0';
	free(binpath);
	binpath = path;

	if (isBusy(binpath, 0)) { goto done; }

	len = strlen(binpath);
	updpath = calloc(1, len + 5);
	memcpy(updpath, binpath, len);
	strncpy(&updpath[(int)len], ".upd", 5);

	/* make sure we can execute it; it might just be a file sitting there */
	if (0 != stat(updpath, &sb)) { goto done; }
	if (0 != chown(updpath, getuid(), getgid())) { goto done; }
	mode = (sb.st_mode & 0x7777) | S_IXUSR | S_IRUSR;
	if (0 != chmod(updpath, mode)) { goto done; }

	/* Looks like we're good to go! */
	setenv("BF_AGENT_UPDATE", updpath, 1);
	if (agent_debug_flag) { printf("DEBUG: Spawning new agent in update mode\n"); }
	execv(updpath, argv);

	/* Shouldn't get here. */
	if (agent_debug_flag) { printf("DEBUG: Agent update attempt failed\n"); }
	unlink(updpath); /* if the exec fails, stop trying */

done:
	if (NULL != binpath) { free(binpath); }
	if (NULL != updpath) { free(updpath); }
}

#endif /* !WIN32 */


/* Program entry point and command line argument processing */
int main(int argc, char **argv) {
	int i;
	char *ip = NULL;
	char *password = NULL;
#ifndef WIN32
	int bf_daemon = 0;
# else /* WIN32 */
	WORD	wVer;
	WSADATA wsaData;

	/* set the control handler to catch logoff events */

	/* there is a return code here, but we don't want to quit */
	/* if there is some problem registering.  Just accept and */
	/* move on.                                               */
	SetConsoleCtrlHandler( (PHANDLER_ROUTINE) CtrlHandler, TRUE ); 

	wVer = MAKEWORD(2,2);
	WSAStartup(wVer,&wsaData);
	platform_map_clear();
#endif /* !/WIN32 */

	/* parse args */
	bfconf_set_path(NULL);
	
	/* close stderr on Linux/Unix for defect RATLC01347322: BF 7.1.3.2 agent test fails when agent has excess output */
	/*  undo this close(2) for init_child_fds(...) in platform.c uses stderr*/
    /*close(2);*/

	while ((i = getopt(argc,argv,GETOPT_OPTIONS)) != EOF) {
		if (agent_debug_flag) {
			printf("DEBUG: arg: -%c\n", i);
		}
		switch (i) {
			/* Encrypt/encode a password.  Note that this is *not* suitable for
			 * use with magic_login; see case 'P' for that.
			 */
			case 'e':
			case 'E':
                        /*26537: CDL 7.1.3.3  Regression jcl_submit_user can't work properly 
                         *for the password decrypt  failed  
                         */

				if (i == 'e') {
#ifdef ZOS
				convert_etoa(optarg);	
#endif
				encrypt_pwd(optarg, &password);
				} else {					
#ifdef ZOS
				convert_atoe(password); 	
#endif
				decrypt_pwd(&password, optarg, 0);
				}

				if (NULL == password) { return 1; }

				printf("%s\n", password);
				free(password);
				return 0;

			/* Specifies the IP address of the connection.  This is used to
			 * enforce the "allow" setting in bfagent.conf.
			 */
			case 'I':
				ip = optarg;
				continue;

			case 'v':
				report_version();
				return 0;

			case 'd':
				if (!agent_debug_flag) {
					char *activity_log = bfconf_get("activity_log", NULL);
					printf("DEBUG: Writing to activity log [%s]\n",
							(NULL != activity_log) ? activity_log : "(stderr)");
					agent_debug_flag = 1;
					free(activity_log);
				}
				continue;

#ifdef WIN32
			/* Only used on Win32.  Provides the socket number to use for
			 * talking to the remote client.  This socket was inherited
			 * from the bfdispatch process that spawned us.  Not compatible
			 * with -R.
			 */
			case 'S':
				if (win_sock || agent_re_exec) { break; }
				win_sock = 1;
				hRemote = (SOCKET)atoi(optarg);
				SetHandleInformation((HANDLE)hRemote, HANDLE_FLAG_INHERIT, 0);
				continue;

			/* Only used on Win32.  Provides the handle number to use for
			 * talking to the bfagent process that spawned us.  It is a
			 * pipe that is inherited from that process.  Not compatible
			 * with -S.
			 */
			case 'R':
				if (win_sock || agent_re_exec) { break; }
				agent_re_exec = 1;
				continue;
#else
			/* Unix-only; full daemon mode (disconnects from controlling terminal) */
			case 's':
				bf_daemon = 1;
				continue;

			/* Unix-only; partial daemon mode (remains in the foreground) */
			case 'x':
				bf_daemon = 2;
				continue;
#endif

			/* Overrides the location of bfagent.conf.  Note that this may impact
			 * the processing of other options.
			 */
			case 'f':
				bfconf_set_path(optarg);
				continue;

			/* Produces a Salted MD5 of the specified password for use in the
			 * magic_login setting.
			 */
			case 'P': {
				size_t len = strlen(optarg) + 1;
				char *buf = malloc(len << 2);
				int rc;

				len = convert_native_to_utf8(optarg, len, buf, len << 2);
				rc = report_smd5(buf, len - 1);
				free(buf);
				return rc;
			}
		}

		send_hello(CODE_HELLO);
		send_msg("ERROR", "InitBadArg", NULL);
		send_result(1);
		send_goodbye();
		return 1;
	}

	if (ip != NULL && !checkIP(ip)) {
		send_hello(CODE_HELLO);
		send_msg("ERROR", "InitAccessDenied", NULL);
		send_result(1);
		send_goodbye();
		return 1;
	}

	activity_open();

#ifndef WIN32
	/* if this agent is in update mode and the update binary exists, we need to activate it */
	handle_agent_update(argv);

	/* Unix agents can run with a -s option to operate as a daemon.  On
	 * Windows, bfdispatch has this job, instead.
	 */
	if (bf_daemon) {
		if (agent_debug_flag) {
			printf("DEBUG: Starting bfagent as a daemon process.\n");
		}
		BFTRACE("=== STARTING DAEMON PROCESS ===");
		bfdaemon(bf_daemon);
	}
	else
#endif /* !WIN32 */
	{
		if (agent_debug_flag) {
			printf("DEBUG: Running agent as a one-shot.\n");
		}
		BFTRACE("=== NEW AGENT ===");
		bfagent();
	}
	platform_unmap_drives(0); 
	BFTRACE("--- EXITING ---");
	exit(0);
}

