/*
Copyright (C)2003-2013, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"
#include <fcntl.h>
#include "md5.h"

#ifdef WIN32
#include <direct.h>
#endif

/* This file provides many different parts of the agent functionality
 * that probably belong somewhere else (possible better homes are
 * given in parentheses, with * indicating that the functionality
 * should probably have a new file all to itself).
 * - Support for _LOG (exec.c)
 * - Filesystem interaction, such as determining file types, recursive
 *			mkdir/rmdir operations, temporary files, etc. (*)
 * - Conversion of binary data to/from hex (string.c)
 * - Custom user credentials (request.c or agent.c?)
 * - Support code for authentication (pam.c, which should be renamed)
 * - Support code for the activity log (agent.c?)
 */

extern ssl_state_t ssl_state;
extern int log_suppressed;
extern FILE *log_file;
char *log_timestamp = NULL;
FILE *activity_file = NULL;

#ifndef WIN32
extern int errno;
#endif

/*
 * function to close an open logfile
 */
void log_close(void)
{
	if (NULL != log_file) {
		fclose(log_file);
		log_file = NULL;
	}
	if (NULL != log_timestamp) {
		free(log_timestamp);
		log_timestamp = NULL;
	}
	log_suppressed = 0;
}

/*
 * function to open a log file if one is specified in a request
 *
 * Return value: non-zero for failure, 0 for success (or not logging)
 */
int log_open(void)
{
	char *file, *p;
	int err = 0;
	size_t cpy_count = 0;

	log_close();

	p = platform_getenv("_SUPPRESS_AGENT_LOG_OUTPUT");
	if (NULL != p) {
		platform_freeenv(p);
		log_suppressed = 1;
	} else {
		log_suppressed = 0;
	}

	p = platform_getenv("BF_LOG");
	if (NULL == p) {
		p = platform_getenv("_LOG");
		if (NULL == p) { return 0; }
	}
	if ('\0' == *p) {
		platform_freeenv(p);
		return 0;
	}
	file = strdup(p);
	platform_freeenv(p);

	p = platform_getenv("_LOG_TIMESTAMP");
	if (NULL != p) {
		if ('\0' != *p) { log_timestamp = strdup(p); }
		platform_freeenv(p);
	}

	CleanPath(file);
	if (*file != PathSeparator() && file[1] != ':') {
		char *buf = malloc(BFBUFSIZ);
		platform_getcwd(buf, BFBUFSIZ);
		p = &buf[strlen(buf)];
		*p++ = PathSeparator();
		cpy_count = (strlen(file) < BFBUFSIZ) ? (strlen(file) + 1) : BFBUFSIZ;
		strncpy(p, file, cpy_count);
		free(file);
		file = buf;
	}
	CleanPath(file);

	log_file = platform_fopen(file, "a");
	if (NULL != log_file) {
		setbuf(log_file, NULL);
		send_msg("LOG", "LogPath", "s", file);
	} else {
		err = errno;
		send_msg("LOG", "LogPathFail", "s", file);
	}
	free(file);
	return err;
}

/*
 * function to log a data block when logging is enabled
 */
void log_write(const char *msg)
{
	char *s, *d;

	if (NULL == log_file || NULL == msg) { return; }
	s = strdup(msg);
	RemoveTrailingNewline(s);
	if (log_timestamp != NULL) {
		d = expandDate(log_timestamp);
		fprintf(log_file, "%s %s\n", d, s);
		free(d);
	} else {
		fprintf(log_file, "%s\n", s);
	}
	free(s);
}

int BFCreateDirectory(char *root, char *dir, int mode)
{
	char *original_working_dir;
	int rc;

	original_working_dir = calloc(sizeof(*original_working_dir), BFBUFSIZ);
	platform_getcwd(original_working_dir, BFBUFSIZ);
	activity_write(BFHERE "BFCreateDirectory original_working_dir=[%s] root=[%s] dir=[%s] mode=%o",
			original_working_dir, root, dir, mode);
	CleanPath(root);

	errno = 0;
	if (platform_chdir(root) != 0) {
		send_msg("MKDIR", "ChdirFail", "si", root, errno);
		rc = -1;
		goto done;
	}

	rc = BFCreateDirectoryRelative(dir, mode);
	if (platform_chdir(original_working_dir) != 0) {
		send_msg("MKDIR", "ChdirFail", "si", original_working_dir, errno);
		rc = -1;
	}

done:
	free(original_working_dir);
	return rc;
}

int BFCreateDirectoryRelative(char *dir, int mode)
{
	char    *buf, *p, *usedir, *root;
	char    *sptr=NULL;
	wchar_t *tmpbuf;
	size_t  cpy_count = 0;

	if (mode == 0) { mode = 0755; }
	root = calloc(sizeof(*root), BFBUFSIZ);
	buf = calloc(sizeof(*buf), BFBUFSIZ);
	tmpbuf = calloc(sizeof(*tmpbuf), BFBUFSIZ);

	platform_getcwd(root, BFBUFSIZ);
	CleanUnixPath(root);
	cpy_count = (strlen(root) < BFBUFSIZ) ? (strlen(root) + 1) : BFBUFSIZ;
	strncpy(buf, root, cpy_count);
	usedir = strdup(dir);
	CleanUnixPath(usedir);

	p = bf_next_token(usedir, "/", &sptr);
	while (p) {
		strncat(buf, "/", 1);
		cpy_count = (strlen(p) < BFBUFSIZ) ? strlen(p) : BFBUFSIZ;
		strncat(buf, p, cpy_count);
		if (!isDir(buf)) {
			if (0 != platform_mkdir(buf, mode)) {
				activity_write(BFHERE "BFCreateDirectory FAIL buf=[%s] mode=%o errno=%d",
						buf, mode, errno);
				break;
			}
			/* force it even if umask squashed something */
			platform_chmod(buf, mode);
		}
		p = bf_next_token(NULL, "/", &sptr);
	}

	free(usedir);
	free(root);
	free(tmpbuf);
	if (isDir(buf)) {
		activity_write(BFHERE "BFCreateDirectory ok buf=[%s] mode=%o", buf, mode);
		free(buf);
		return 0;
	}
	free(buf);
	return -1;
}

/* 
 *  function to test a pathname to see if it's a directory. 
 *  doesn't alter path at all. 
 */
#ifndef S_ISDIR
#define S_ISDIR(x)	((x & S_IFDIR) == S_IFDIR)
#endif
int isDir(const char *path)
{
	struct 	stat sb;
	return (platform_stat(path, &sb) == 0 && S_ISDIR(sb.st_mode)) ? 1 : 0;
}

/* 
 *  function to test a pathname to see if it's a regular file. 
 *  doesn't alter path at all. 
 */
#ifndef S_ISREG
#define S_ISREG(x)	((x & S_IFREG) == S_IFREG)
#endif
int isFile(const char *path)
{
	struct 	stat sb;
	return (platform_stat(path, &sb) == 0 && S_ISREG(sb.st_mode)) ? 1 : 0;
}


static const char *HEX_ENCODE =
	"000102030405060708090A0B0C0D0E0F"
	"101112131415161718191A1B1C1D1E1F"
	"202122232425262728292A2B2C2D2E2F"
	"303132333435363738393A3B3C3D3E3F"
	"404142434445464748494A4B4C4D4E4F"
	"505152535455565758595A5B5C5D5E5F"
	"606162636465666768696A6B6C6D6E6F"
	"707172737475767778797A7B7C7D7E7F"
	"808182838485868788898A8B8C8D8E8F"
	"909192939495969798999A9B9C9D9E9F"
	"A0A1A2A3A4A5A6A7A8A9AAABACADAEAF"
	"B0B1B2B3B4B5B6B7B8B9BABBBCBDBEBF"
	"C0C1C2C3C4C5C6C7C8C9CACBCCCDCECF"
	"D0D1D2D3D4D5D6D7D8D9DADBDCDDDEDF"
	"E0E1E2E3E4E5E6E7E8E9EAEBECEDEEEF"
	"F0F1F2F3F4F5F6F7F8F9FAFBFCFDFEFF";

int hex_encode(char *buf, const uint8_t *src, int bytes) {
	char *p = buf;
	int i;

	for (i=0; i<bytes; ++i) {
		const char *q = &HEX_ENCODE[((int)(*src++)) << 1];
		*p++ = *q++;
		*p++ = *q;
	}
	*p = '\0';
	return (int)(p - buf);
}

#ifdef ZOS
/* IBM-1047 hex decoding map */
static short HEX_DECODE[] = {
	-2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 00-0F */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 10-1F */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 20-2F */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 30-3F */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 40-4F */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 50-5F */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 60-6F */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 70-7F */
	-1,0xA,0xB,0xC,0xD,0xE,0xF, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 80-8F */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 90-9F */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* A0-AF */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* B0-BF */
	-1,0xA,0xB,0xC,0xD,0xE,0xF, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* C0-CF */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* D0-DF */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* E0-EF */
	 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, -1, -1, -1, -1, -1, -1  /* F0-FF */
};
# else /* !ZOS */
/* ISO 8859-1 hex decoding map */
static short HEX_DECODE[] = {
	-2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 00-0F */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 10-1F */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 20-2F */
	 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, -1, -1, -1, -1, -1, -1, /* 30-3F */
	-1,0xA,0xB,0xC,0xD,0xE,0xF, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 40-4F */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 50-5F */
	-1,0xA,0xB,0xC,0xD,0xE,0xF, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 60-6F */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 70-7F */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 80-8F */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 90-9F */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* A0-AF */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* B0-BF */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* C0-CF */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* D0-DF */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* E0-EF */
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1  /* F0-FF */
};
#endif /* ZOS/! */


int hex_decode(uint8_t *result, const char *src) {
	uint8_t *q = result;
	const uint8_t *p = (const uint8_t *)src;
	int x,y;

	for (;;) {
		x = HEX_DECODE[(int)*p++];
		if (x < 0) {
			if (-2 == x) { break; }
			*q++ = '?';
			if ('\0' == *p++) { break; }
			continue;
		}

		y = HEX_DECODE[(int)*p++];
		if (y < 0) {
			*q++ = '?';
			if (-2 == y) { break; }
		}

		*q++ = (uint8_t)((x << 4) + y);
	}   
	return (int)(q - result);
}




void cleanMyCruft(char *path) {
	if (path) {
		if (BFCONF_EXISTS(leave_tmp_file)) {
			send_msg("EXEC", "LeavingTemp", "s", path);
		} else {
			platform_unlink(path);
		}
	}
}


void StoreUserData(char *data, size_t bytes) {
	char *e;
	FILE *f;

	if (data && NULL != (e = platform_getenv("_USERDATA"))) {
		if (NULL != (f = platform_fopen(e, "w"))) {
			fwrite(data, 1, bytes, f);
			fclose(f);
		}
		platform_freeenv(e);
	}
}

void FreeUserData(void) {
	char *e;

	if (NULL != (e = platform_getenv("_USERDATA"))) {
		platform_unlink(e);
		platform_freeenv(e);
	}
}

/*
 * If we are in the SS_CERTIFIED state for SSL, then the client was
 * successfully able to supply a certificate for identification.  If
 * explicitly enabled in bfagent.conf, this allows them to bypass
 * logging in.  Essentially, this is like magic_login, except that
 * it uses a strong cryptography and is allowed even for privileged
 * users.
 */
int check_cert_login(void) {
	char *p;
	int ret;

	if (ssl_state != SS_CERTIFIED) { return SU_ILLEGAL; }
	p = bfconf_get("ssl_client_auth_sufficient", NULL);
	if (NULL == p) { return SU_ILLEGAL; }
	ret = (0 == strcmp(p, "true")) ? SU_OK_NOACCOUNT : SU_ILLEGAL;
	free(p);
	return ret;
}

/* The magic_login is intended to increase the security of the agent.
 * Its behavior depends on the agent configuration, as follows:
 *
 * Running as root (Unix) or a member of the Administrators group (Windows):
 * - magic_login is ignored
 * - user must provide valid login credentials
 *
 * Running as an ordinary user:
 * - The credentials must match the magic login
 * - If no magic login is defined, then the credentials must be left blank
 *
 * Return value:
 * 		SU_OK           = Success
 * 		SU_OK_NOACCOUNT = Not applicable (no magic_login defined)
 * 		SU_ILLEGAL      = Denied - no login specified
 * 		SU_UNKNOWNUSER  = Denied - username or password mismatch
 */
int auth_magic = 0;
int check_magic_login(const char *username, const char *password) {
	char *p;
	char *magic_login = NULL;
	int ret = SU_UNKNOWNUSER;

	magic_login = bfconf_get("magic_login", NULL);
	if (username == NULL || *username == '\0'
			|| password == NULL || *password == '\0')
	{
		if (NULL == magic_login) {
			BFTRACE("AUTH user mode (Insecure!)");
			return SU_OK_NOACCOUNT;
		}
		BFTRACE("AUTH failed (magic login required)");
		free(magic_login);
		return SU_ILLEGAL;
	}
	if (NULL == magic_login) {
		BFTRACE("AUTH failed (creds specified for insecure mode)");
		return SU_UNKNOWNUSER;
	}

	p = strrchr(magic_login, ':');
	if (p != NULL) *p++ = '\0';

	if (strcmp(username, magic_login) == 0) {
		size_t len = strlen(password);
#ifdef ZOS
		char *pass = strdup(password);
		convert_etoa_l(pass, len);
		password = pass;
#endif
		if (smd5_check(p, (const md5_byte_t *)password, (int)len) == 0) {
			auth_magic = 1;
		BFTRACE("AUTH ok (magic_login)");
			ret = SU_OK;
		} else {
			BFTRACE("AUTH failed (magic_login)");
		}
#ifdef ZOS
		free(pass);
#endif
	}

	free(magic_login);
	return ret;
}


void activity_open(void) {
	char *activity_log = bfconf_get("activity_log", NULL);
	activity_close();
	if (NULL != activity_log) {
		activity_file = platform_fopen(activity_log, "a+");
		free(activity_log);
	} else if (agent_debug_flag) {
		int fd = dup(2);
		activity_file = fdopen(fd, "a+");
	}
}

static int bf_tmp_file_name_counter = 0;
void bf_temp_file_name(char *buffer) {
	md5_state_t ctx;
	md5_byte_t  digest[16];
	pid_t       pid = getpid();
	time_t      now = time(NULL);
	int         rnum;
	int         i = ++bf_tmp_file_name_counter;
	char *      p = &buffer[(int)strlen(buffer)];

#ifdef HAVE_SSL
	RAND_pseudo_bytes((unsigned char *)&rnum, sizeof(int));
# else /* !HAVE_SSL */
	rnum = rand();
#endif /* HAVE_SSL/! */

	md5_init(&ctx);
	md5_append(&ctx, (const md5_byte_t *)&pid, sizeof(pid_t));
	md5_append(&ctx, (const md5_byte_t *)&now, sizeof(time_t));
	md5_append(&ctx, (const md5_byte_t *)&rnum, sizeof(int));
	md5_append(&ctx, (const md5_byte_t *)&i, sizeof(int));
	md5_finish(&ctx, digest);

	for (i=0; i<16; ++i) {
		snprintf(p, 3, "%02X", (unsigned int)(digest[i]));
		p += 2;
	}
}

static const char NL = '\n';
void activity_write(const char *file, int line, const char *fmt, ...) {
	const char *p;
	va_list va;
	size_t len;
	time_t now;
	char buf[256];

	if (!bf_debug && NULL == activity_file) { return; }
	while (NULL != ( p = strpbrk(file, "/\\") )) { file = ++p; }
	time(&now);
	strftime(buf, 256, "%c", localtime(&now));

	va_start(va, fmt);
	if (bf_debug) {
		len = snprintf(bf_debug, BFBUFSIZ, BFTRACE_FORMAT, buf, (int)getpid(), file, line);
		len += vsnprintf(&bf_debug[len], BFBUFSIZ, fmt, va);
		send_debug(bf_debug);
		if (NULL != activity_file) {
			bf_debug[len++] = '\n';
			fwrite(bf_debug, 1, len, activity_file);
			fflush(activity_file);
		}
	} else if (NULL != activity_file) {
		fprintf(activity_file, BFTRACE_FORMAT, buf, (int)getpid(), file, line);
		vfprintf(activity_file, fmt, va);
		fwrite(&NL, 1, 1, activity_file);
		fflush(activity_file);
	}
	va_end(va);
}

void activity_close(void) {
	if (NULL != activity_file) {
		fclose(activity_file);
		activity_file = NULL;
	}
}

