/*
Copyright (C)2003-2011, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#define BF_USE_REGEX
#include "bf_agent.h"
#include "bf_rand.h"
#include "md5.h"
#include "xstream.h"

/* Implements "cmd read", which is used to read files off the agent for
 * checkpoint or file transfer purposes.  This is used by .push, .get,
 * .put, .rget, .rput, and .scan.
 */

typedef struct _bf_ignore_t {
	struct _bf_ignore_t *   next;
	char *                  ignore_substr;
	bf_regex_t *            ignore_regex;
} bf_ignore_t;

static bf_ignore_t *ignore_list = NULL;
char *cb_read_buf = NULL;
size_t cb_read_buf_sz = 0;

#ifndef ZOS
extern SOCKET xs_sock;
#endif

static void ignore_list_free(void) {
	bf_ignore_t *ig = ignore_list;
	while (NULL != ig) {
		bf_ignore_t *next = ig->next;
		if (NULL != ig->ignore_regex) { bf_regex_free(ig->ignore_regex); }
		free(ig);
		ig = next;
	}
	ignore_list = NULL;
}

static void ignore_list_init(void) {
	bf_ignore_t head;
	bf_ignore_t *ig = &head;
	char *igs;
	char *p, *q;

	if (NULL != ignore_list) { return; }

	ig->next = NULL;
	igs = request_get(TAG_IGNORE);

	for (p = igs; NULL != p && '\0' != *p; p = q) {
		q = strchr(p, '\n');
		if (NULL != q) { *q++ = '\0'; }

		if (p[0] == '\0' || p[1] != ':' || p[2] == '\0') {
			activity_write(BFHERE "Discarding malformed ignore request [%s]", p);
			continue;
		}
		ig->next = malloc(sizeof(bf_ignore_t));
		if (NULL == ig->next) {
			BFTRACE("Out of memory trying to process ignore list!");
			continue;
		}
		ig->next->ignore_substr = NULL;
		ig->next->ignore_regex = NULL;
		ig->next->next = NULL;

		switch (*p) {
			case 'S':
				ig->next->ignore_substr = &p[2];
				break;

			case 'R':
				ig->next->ignore_regex = bf_regex_compile(&p[2], RE_FLAG_PCRE);
				if (NULL != ig->next->ignore_regex) { break; }
				activity_write(BFHERE "Discarding unparsable ignore regex [%s]", p);
				free(ig->next);
				ig->next = NULL;
				continue;

			default:
				free(ig->next);
				ig->next = NULL;
				activity_write(BFHERE "Discarding unsupported ignore request [%s]", p);
				continue;
		}
		ig = ig->next;
	}

	ignore_list = head.next;
}

static int is_ignore(const char *name) {
	const char *p, *q;
	regmatch_t *posix_groups;
	bf_ignore_t *ig;
	
	for (ig = ignore_list; NULL != ig; ig = ig->next) {
		if (NULL != ig->ignore_substr) {
			p = name;
			for (;;) {
				q = strstr(p, ig->ignore_substr);
				if (NULL == q) { break; }
				if (q == name || q[-1] == '/' || q[-1] == '\\') {
					p = &q[(int)strlen(ig->ignore_substr)];
					if (*p == '\0' || *p == '/' || *p == '\\') { return 1; }
				}
				p = &q[1];
			}
		} else {
			posix_groups = bf_regex_match(name, ig->ignore_regex);
			if (NULL != posix_groups) {
				free(posix_groups);
				return 1;
			}
		}
	}
	return 0;
}

/* The CRC-32 is mostly pointless.  The reason it is exists here is that
 * Build Forge originally used MySQL 3.23 as its primary supported platform,
 * and that database was incapable of indexing on a large field like the one
 * that the full pathname was stored in.  So we stored a CRC-32 of the name
 * as well and used that to index against.  All of the databases we've used
 * since the 7.0 release have been able to index the name directly, so this
 * is no longer necessary, but kept for compatibility and because it isn't
 * expensive enough to bother getting rid of it.
 */

/*
 * CRC-32-MPEG2 (aka IEEE 802.3 / V.42)
 */
static uint32_t *crc_table = NULL;
static void crc_gen_table(void)			/* build the crc table */
{
	uint32_t crc, poly = 0xEDB88320L;
	int	i, j;

	crc_table = calloc(256, sizeof(uint32_t));

	for (i = 0; i < 256; i++) {
		crc = i;
		for (j = 8; j > 0; j--) {
			if (crc & 1) {
				crc = (crc >> 1) ^ poly;
			} else {
				crc >>= 1;
			}
		}
		crc_table[i] = crc;
	}
}

static uint32_t crc_string(const char *s)	/* calculate the crc value */
{
	register uint32_t crc = 0xFFFFFFFF;
	if (NULL != s) {
		while ('\0' != *s) {
			crc = (crc >> 8) ^ crc_table[(crc ^ *s++) & 0xFF];
		}
	}
	return crc ^ 0xFFFFFFFF;
}




/*
 * Callback functions.  The cb_read_normal callback is used when we want to
 * stream the file contents across the wire using messages 341/342.  The
 * cb_read_xstream is used when we have established an external stream
 * for carrying the data and are using 345 to communicate the file metadata.
 * The bf_checkpoint callback is used when the "checkpoint" flag is specified,
 * in which case we send 330 messages (.scan data) instead.
 */

static int cb_read_normal(const char *p, const struct stat *sb, platform_ftw_flag_t flag)
{
	int     fd = -1;
	int     err = 0;
	size_t  bytes;
	off_t   offset;
	char *  path = CleanPath(strdup(p));

	switch (flag) {
		case BF_FTW_FILE:
			if (is_ignore(path)) {
				activity_write(BFHERE "cb_read_normal: ignoring FILE: [%s]", path);
				break;
			}
			fd = platform_open2(path, O_RDONLY|O_BINARY);
			if (fd == -1) {
				activity_write(BFHERE "cb_read_normal: unable to open FILE: [%s]: %d", path, errno);
				break;
			}

			offset = (off_t)0;
			while ((bytes = read(fd, cb_read_buf, cb_read_buf_sz)) > 0) {
				if (!send_file_info(
						(off_t)offset,
						(long)bytes,
						(int)sb->st_mode,
						(long)sb->st_size,
						(long)sb->st_mtime,
						path))
				{
					err = 7;
					goto done;
				}
				if (!send_file_data((const uint8_t *)cb_read_buf, bytes)) {
					err = 8;
					goto done;
				}
				offset += bytes;
			}
			if (offset == (off_t)0) {
				if (!send_file_info((off_t)0, (long)0, (int)sb->st_mode, (long)0, (long)sb->st_mtime, path)) {
					err = 9;
					goto done;
				}
			}
			break;
		case BF_FTW_DIR:
		case BF_FTW_DIR_NO_READ:
			if (is_ignore(path)) {
				activity_write(BFHERE "cb_read_normal: ignoring DIR: [%s]", path);
				err = BF_FTW_DIR_PRUNE;
				goto done;
			}
			if (!send_dir((int)sb->st_mode, (long)sb->st_mtime, path)) {
				err = 10;
				goto done;
			}
			break;
		default:
			/* activity_write(BFHERE "cb_read_normal: ignoring file type %d", (int)flag); */
			break;
	}

done:
	if (-1 != fd) { close(fd); }
	free(path);
	return err;
}


static int cb_read_xstream(const char *p, const struct stat *sb, platform_ftw_flag_t flag)
{
	int     fd = -1;
	int     err = 0;
	char *  path = CleanPath(strdup(p));

	switch (flag) {
		case BF_FTW_FILE:
			if (is_ignore(path)) {
				activity_write(BFHERE "cb_read_xstream: ignoring FILE: [%s]", path);
				break;
			}
			fd = platform_open2(path, O_RDONLY|O_BINARY);
			if (fd == -1) {
				activity_write(BFHERE "cb_read_xstream: unable to open FILE: [%s]: %d", path, errno);
				break;
			}

			if (!send_file_info_xstream(
					(int)sb->st_mode,
					(long)sb->st_size,
					(long)sb->st_mtime,
					path) ||
				!send_cache_flush())
			{
				err = 7;
				break;
			}

#ifndef ZOS
			if (!xstream_send_file(path, sb, fd)) {
				err = 8;
				break;
			}
#endif
			break;
		case BF_FTW_DIR:
		case BF_FTW_DIR_NO_READ:
			if (is_ignore(path)) {
				activity_write(BFHERE "cb_read_xstream: ignoring DIR: [%s]", path);
				err = BF_FTW_DIR_PRUNE;
				break;
			}
			if (!send_dir((int)sb->st_mode, (long)sb->st_mtime, path)) {
				err = 10;
				break;
			}
			break;
		default:
			/* activity_write(BFHERE "cb_read_xstream: ignoring file type %d", (int)flag); */
			break;
	}

	if (-1 != fd) { close(fd); }
	free(path);
	return err;
}

static int cb_checkpoint(const char *p, const struct stat *sb, platform_ftw_flag_t flag)
{
	uint32_t        crc;
	md5_state_t     ctx;
	md5_byte_t      digest[16];
	int             fd;
	int             got;
	int             err = 0;
	off_t           block;
	off_t           blocks;
	char *          path = CleanPath(strdup(p));

	crc = crc_string(path);
	switch (flag) {
		case BF_FTW_FILE:
			if (is_ignore(path)) {
				activity_write(BFHERE "cb_checkpoint: ignoring FILE: [%s]", path);
				send_checkpoint_ignored("FILE", crc, (long)sb->st_mtime, path);
				break;
			}
			block = 0;
			blocks = sb->st_size / cb_read_buf_sz;
			if (sb->st_size % cb_read_buf_sz > 0) { ++blocks; }

			fd = platform_open2(path, O_RDONLY|O_BINARY);
			if (fd == -1) { break; }

			md5_init(&ctx);
			while ((got = (int)read(fd, cb_read_buf, cb_read_buf_sz)) > 0) {
				md5_append(&ctx, (const md5_byte_t *)cb_read_buf, got);
				if ((++block % cb_read_buf_sz) == 0) {
					activity_write(BFHERE "cb_checkpoint: md5 block %llu of %llu",
							(unsigned long long)block,
							(unsigned long long)blocks );
				}
			}
			md5_finish(&ctx, digest);

			close(fd);
			send_checkpoint("FILE", crc, (long)sb->st_mtime, digest, path);
			break;
		case BF_FTW_DIR:
		case BF_FTW_DIR_NO_READ:
			if (is_ignore(path)) {
				err = BF_FTW_DIR_PRUNE;
				activity_write(BFHERE "cb_checkpoint: ignoring DIR: [%s]", path);
				send_checkpoint_ignored("DIR", crc, (long)sb->st_mtime, path);
			} else {
				send_checkpoint("DIR", crc, (long)sb->st_mtime, NULL, path);
			}
			break;
		case BF_FTW_SYMLINK:
			if (is_ignore(path)) {
				activity_write(BFHERE "cb_checkpoint: ignoring SYMLINK: [%s]", path);
				send_checkpoint_ignored("SYMLINK", crc, (long)sb->st_mtime, path);
			} else {
				send_checkpoint("SYMLINK", crc, (long)sb->st_mtime, NULL, path);
			}
			break;
		default:
			/* activity_write(BFHERE "cb_checkpoint: ignoring file type %d", (int)flag); */
			break;
	}

	free(path);
	return err;
}

static int ftw_read(const char *path, platform_ftw_fn_t cb, size_t bufsiz) {
	int err;
	cb_read_buf = malloc(bufsiz);
	if (NULL == cb_read_buf) {
		err = ENOMEM;
	} else {
		cb_read_buf_sz = bufsiz;
		err = platform_ftw(path, cb);
		free(cb_read_buf);
		cb_read_buf = NULL;
	}
	return err;
}


/*
 * Implementation for "cmd read".
 *
 * Read a local file/tree and send it to stdout.
 */
int FuncRead(void)
{
	const char *proto;
	char	*path, *recurse, *checkpoint, *p, *userdata;
	char	*timeout;
	char    *saved_dir = NULL;
	int 	rc = 0;
	int 	err = 0;
	size_t	userdata_size;

	/*  locate the file/dir element */
	if (NULL == (path = request_get(TAG_PATH))) {
		send_missing(CMD_READ, TAG_PATH);
		return 1;
	}
	path = strdup(path); /* this will get free'd later */

	/*  locate the userdata element */
	userdata = request_get_binary(TAG_USERDATA, &userdata_size);

	/*  locate the client timeout element */
	timeout = request_get(TAG_TIMEOUT);
	if (NULL == timeout) { timeout = "0"; }

	/*  optional recursive flag */
	recurse = request_get(TAG_RECURSIVE);

	/*  optional checkpoint flag, which implies recursive */
	checkpoint = request_get(TAG_CHECKPOINT);
	if (checkpoint && !recurse) { recurse = checkpoint; }

	StoreUserData(userdata, userdata_size);

	if (0 != platform_map_drives(defMapStr, mapStr)) {
		send_msg("MAP", "MapFail", NULL);
		FreeUserData();
		/* platform_unmap_drives(0); */
		return 2;
	}

	ClobberPasswords();

	if (clearcase_view) {
		size_t len = strlen(this_view()) + strlen(clearcase_view) + strlen(path) + 3;
		char *ccpath = calloc(1, len);
		if (BFCONF_EXISTS(cc_suppress_server_root)) {
			format_message(ccpath, "%s/%s", this_view(), clearcase_view);
		} else {
			format_message(ccpath, "%s/%s/%s", this_view(), clearcase_view, path);
		}
		free(path);
		path = ccpath;
	}

	/* process the clearcase view/vob start/mount functions */
	CCStartViews(atoi(timeout));
	CCMountVobs(atoi(timeout));

	/*  assemble and clean the final pathname. clean the
	 *  pieces so we can use them individually later.
	 */
	CleanPath(path);
	p = path;
	path = env_expand_var(p, 0);
	free(p);	/* Free the previous value of path */

	/* if not recursive, path must point to a file */
	if (!checkpoint && !recurse && !isFile(path)) {
		send_msg("READ", "ReadNotFile", "s", path);
		rc = 3;
		goto done;
	}
	if (0 != platform_access(path, 0)) {
		/* We can't checkpoint or read something that isn't there. */
		send_msg("READ", "ReadNotFound", "s", path);
		rc = 4;
		goto done;
	}

	/* RATLC1334594 - Windows likes to lie to us about paths with trailing
	 * whitespace.  Here, we force the path to canonicalize so we won't
	 * end up thinking the directory is ok, then getting nothing inside of it.
	 */
	saved_dir = recurse ? malloc(MAX_PATH) : NULL;
	if (NULL != saved_dir) {
		platform_getcwd(saved_dir, MAX_PATH);
		if (0 != platform_chdir(path)) {
			/* We can't checkpoint or read something that isn't there. */
			send_msg("READ", "ReadNotFound", "s", path);
			rc = 6;
			goto done;
		}
		p = malloc(MAX_PATH);
		if (NULL != p) {
			platform_getcwd(p, MAX_PATH);
			free(path);
			path = p;
		}
		platform_chdir(saved_dir);
		free(saved_dir);
		saved_dir = NULL;
	}

	ignore_list_init();

	if (checkpoint) {
		if (NULL == crc_table) { crc_gen_table(); }
		err = ftw_read(path, &cb_checkpoint, BFBUFSIZ);
	} else {
		proto = request_get(TAG_XSTREAM);
		if (NULL == proto) {
			err = ftw_read(path, &cb_read_normal, BFXFERBLOCKSIZE);
		} 

#ifndef ZOS
else if (xstream_offer(proto)) {
			err = ftw_read(path, &cb_read_xstream, BFBUFSIZ);
#ifndef ZOS
			if (INVALID_SOCKET != xs_sock) {
				int x = 0;
				if (!xstream_send_bye()) { x = errno; }
				xstream_close();
				if (0 == err) { err = x; }
			}
#endif
		} 
#endif
else {
			err = errno;
			if (0 == err) { err = 7; }
		}
	}


done:
	/* platform_unmap_drives(0); */
	FreeUserData();
	if (err) {
		send_msg("READ", "ReadFail", "is", err, path);
		rc = 5;
	} else if (0 == rc) {
		send_msg("READ", "ReadOk", "s", path);
	}
	if (NULL != cb_read_buf) {
		free(cb_read_buf);
		cb_read_buf = NULL;
	}
	ignore_list_free();
	if (NULL != saved_dir) { free(saved_dir); }
	if (NULL != path) { free(path); }
	return rc;
}

