/*
Copyright (C)2003-2013, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

/*
 * Implementation for the protocol output generation routines declared in
 * send.h.
 */

/*
 * GLOBAL STATE
 *
 * - datablock_cache*
 * The datablock cache is enabled by the "command_output_cache" setting in
 * bfagent.conf.  It causes output to be buffered in memory before it is
 * flushed to the output file descriptor / socket, thereby allowing more
 * efficient I/O and (usually) much better throughput.
 *
 * - output_buffer*
 * Used to format individual lines for protocol output.
 *
 * - did_hello
 * Tracks whether or not the agent has sent a 200 HELLO or equivalent message
 * yet.  The point is that even if debugging is enabled, we should not spit out
 * any 101 DEBUG messages before the 200 HELLO, as this would confuse the
 * protocol.
 */

static char *datablock_cache = NULL;
static size_t datablock_cache_len = 0;
static size_t datablock_cache_size = 0;

static char *output_buffer = NULL;
static size_t output_buffer_size = 0;

int did_hello = 0;

/* From agent.c */
extern int heartbeat_interval;
extern int log_suppressed;
extern FILE *log_file;

/*
 * Flushes the output buffer -- either to the datablock cache or
 * directly to io_puts()...
 */
static int flush_output(void) {
	size_t len;

	if (NULL == output_buffer) { return 0; }

	/* If we are caching... */;
	if (NULL != datablock_cache) {
		len = strlen(output_buffer) + 1;

		/* If it would overflow the cache, then flush to make more room. */
		if (datablock_cache_len + len >= datablock_cache_size) {
			if (!send_cache_flush()) { return 0; }
		}

		/* If it fits in the cache, then cache it */
		if (len < datablock_cache_size) {
			memcpy(&datablock_cache[datablock_cache_len], output_buffer, len);
			datablock_cache_len += len - 1;
			*output_buffer = '\0';
			return 1;
		}
	}

	/* Not caching or output_buffer is too big to fit */
	/* activity_write(BFHERE "flush_output,buf len:[%d]",  (int)strlen(output_buffer) ); */
	len = io_puts(output_buffer);
	*output_buffer = '\0';
	return (int)len;
}


int send_cache_flush(void) {
	int ok = 1;
	if (NULL != datablock_cache && '\0' != *datablock_cache) {
		ok = (io_puts(datablock_cache) > 0) ? 1 : 0;
		*datablock_cache = '\0';
		datablock_cache_len = 0;
	}
	return ok;
}


int send_cache_init(int size) {
	if (size <= 0) { return 0; }

	BUFFER_ENSURE_CAPACITY(
			&datablock_cache,
			&datablock_cache_size,
			size );

out_of_memory:
	return (int)datablock_cache_size;
}


int send_checkpoint(
		const char *type,
		uint32_t crc32,
		long mtime,
		const uint8_t *md5,
		const char *path)
{
	char crc32hex[16];
	char *md5hex = NULL;
	size_t len;

	snprintf(crc32hex, sizeof(crc32hex), "%x", (unsigned int)crc32);
	if (NULL != md5) {
		md5hex = malloc(48);
		hex_encode(md5hex, md5, 16);
	}

	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			strlen(type) + (strlen(path) << 2) + 128 );
	len = format_message(output_buffer, "%d %s %08x %ld %s ",
			CODE_READ_SCAN,
			type,
			(unsigned int)crc32,
			mtime,
			(NULL != md5hex) ? md5hex : "." );
	if (NULL != md5hex) { free(md5hex); }

	buffer_append_escaped(
			&output_buffer,
			&output_buffer_size,
			&len,
			path);
	output_buffer[len++] = '\n';
	output_buffer[len] = '\0';

out_of_memory:
	return flush_output();
}


int send_checkpoint_ignored(
		const char *type,
		uint32_t crc32,
		long mtime,
		const char *path)
{
	char crc32hex[16];
	size_t len;

	snprintf(crc32hex, sizeof(crc32hex), "%x", (unsigned int)crc32);

	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			strlen(type) + (strlen(path) << 2) + 128 );
	len = format_message(output_buffer, "%d %s %08x %ld %s ",
			CODE_READ_SCAN,
			type,
			(unsigned int)crc32,
			mtime,
			"-" );

	buffer_append_escaped(
			&output_buffer,
			&output_buffer_size,
			&len,
			path);
	output_buffer[len++] = '\n';
	output_buffer[len] = '\0';

out_of_memory:
	return flush_output();
}


int send_cont(const char *msg) {
	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			strlen(msg) + 32 );
	format_message(output_buffer, "%d CONT %s\n",
			CODE_CONT, msg);
	log_write(output_buffer);

out_of_memory:
	return flush_output();
}


int send_debug(const char *message) {
	if (NULL == bf_debug || NULL == message || !did_hello) {
		return 1;
	}
	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			strlen(message) + 16 );
	format_message(output_buffer, "%d DEBUG %s\n",
			CODE_DEBUG, message);

out_of_memory:
	return flush_output();
}



int send_dir(int mode, long mtime, const char *path)
{ 
	size_t idx;

	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			(strlen(path) << 2) + 128 );
	idx = format_message(output_buffer, "%d DIR %o %ld ",
			CODE_READ_DIR, mode, mtime);
	buffer_append_escaped(&output_buffer, &output_buffer_size, &idx, path);
	output_buffer[idx++] = '\n';
	output_buffer[idx] = '\0';

out_of_memory:
	return flush_output();
}


int send_file_data(const uint8_t *bytes, size_t len)
{
	size_t idx = 4;

	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			(len << 1) + 8 );
	format_message(output_buffer, "%d ", CODE_READ_DATA);
	idx += hex_encode(&output_buffer[idx], bytes, (int)len);
	output_buffer[idx++] = '\n';
	output_buffer[idx] = '\0';

out_of_memory:
	return flush_output();
}


int send_file_info(
		off_t offset,
		long bytes,
		int mode,
		long size,
		long mtime,
		const char *path)
{
	size_t idx;

	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			(strlen(path) << 2) + 128 );
	idx = format_message(output_buffer, "%d FILE %llu %lu %o %ld %ld ",
			CODE_READ_FILE,
			(unsigned long long)offset,
			bytes,
			mode,
			size,
			mtime);
	buffer_append_escaped(&output_buffer, &output_buffer_size, &idx, path);
	output_buffer[idx++] = '\n';
	output_buffer[idx] = '\0';

out_of_memory:
	return flush_output();
}


int send_goodbye(void) {
	char buf[128];
	format_message(buf, "%d %s\n", CODE_GOODBYE, CODE_GOODBYE_MSG);
	return io_puts(buf);
}


static time_t beatmark = 0;
static char beatbuf[32];
int send_heartbeat(void) {
	time_t now;

	time(&now);
	if (now - beatmark > heartbeat_interval) {
		if (!send_cache_flush()) { return 0; }
		beatmark = now;
		format_message(beatbuf, "%d HEARTBEAT %d\n",
				CODE_HEARTBEAT, bf_LoadValue());
		return (io_puts(beatbuf) > 0) ? 1 : 0;
	}
	return 1;
}


int send_hello(int code) {
	const char *msg;
	char buf[128];

	switch (code) {
		case CODE_HELLO:            msg = CODE_HELLO_MSG;            break;
		case CODE_SECURE_HELLO:     msg = CODE_SECURE_HELLO_MSG;     break;
		case CODE_CERTIFIED_HELLO:  msg = CODE_CERTIFIED_HELLO_MSG;  break;
		default: msg = "???";
	}
	did_hello = 1;
	format_message(buf, "%d %s\n", code, msg);
	return io_puts(buf);
}


int send_log(const char *bucket, const char *msg) {
	size_t bucketlen;
	size_t msglen;
	char *p, *q, *r, *s;
	int suppress = log_suppressed;
	size_t cpy_count = 0;

	if (NULL == bucket || '\0' == *bucket) {
		bucket = "*";
		bucketlen = 1;
	} else {
		bucketlen = strlen(bucket);
		if (0==strcmp(bucket, "RESULT") ||0==strcmp(bucket, "MKDIR")|| 0==strcmp(bucket, "SCRIPT")) {
			suppress = 0;
		}
	}

	if (suppress) {
		if (NULL == log_file) { return 2; }
		if (!flush_output()) { return 0; }
	}

	while (NULL != ( p = strpbrk(msg, "\r\n") )) {
		msglen = (size_t)(p - msg);
		BUFFER_ENSURE_CAPACITY(
				&output_buffer,
				&output_buffer_size,
				bucketlen + msglen + 32);

		p = output_buffer;
		p += format_message(p, "%d %s ", CODE_LOG, bucket);
		memcpy(p, msg, msglen);
		p += msglen;
		*p++ = '\n';
		*p = '\0';
		log_write(output_buffer);

		msg = &msg[msglen];
		if ('\0' == *msg) { break; }
		msg = ('\r' == *msg && '\n' == msg[1]) ? &msg[2] : &msg[1];
	}

	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			bucketlen + strlen(msg) + 32);
	q = strdup(msg);
	msglen = strlen(msg);
	r = (char *)calloc(msglen, sizeof(char));
	s = (char *)calloc(msglen, sizeof(char));
	msglen = strlen(FORGE);
	p = (char *)strstr(msg, FORGE);
	if (NULL != p) {
		memset(q, 0, strlen(q));
		strncpy(q, msg, (p - msg));
		p = (char *)strstr((p + msglen), FORGE);
		if (NULL == p) {
			/*
			This will cause some long step log missing some content when its length is bigger than BFBUFSIZ(16384)
			*/
			cpy_count = (strlen(msg) < BFBUFSIZ) ? (strlen(msg) + 1) : BFBUFSIZ;
			strncpy(q, msg, cpy_count);
		} else {
			strncat(q, "*****", 5);
			p += msglen + 8;
			cpy_count = (strlen(p) < BFBUFSIZ) ? (strlen(p) + 1) : BFBUFSIZ;
			strncpy(s, p, cpy_count);
			p = (char *)strstr(s, FORGE);
			while (NULL != p) {
				memset(r, 0, strlen(r));
				strncpy(r, s, (p - s));
				cpy_count = (strlen(r) < BFBUFSIZ) ? strlen(r) : BFBUFSIZ;
				strncat(q, r, cpy_count);
				p = (char *)strstr((p + msglen), FORGE);
				if (NULL == p) {
					break;
				}
				strncat(q, "*****", 5);
				p += msglen + 8;
				cpy_count = (strlen(p) < BFBUFSIZ) ? (strlen(p) + 1) : BFBUFSIZ;
				strncpy(s, p, cpy_count);
				p = (char *)strstr(s, FORGE);
			}
			cpy_count = (strlen(s) < BFBUFSIZ) ? strlen(s) : BFBUFSIZ;
			strncat(q, s, cpy_count);
		}
	}
	if (NULL != s) {
		free(s);
	}
	if (NULL != r) {
		free(r);
	}

	format_message(output_buffer, "%d %s %s\n", CODE_LOG, bucket, q);
	/*
	activity_write(BFHERE "send_log, msg len:[%d],output_buffer len:[%d]",  (int)strlen(q), (int)strlen(output_buffer) );
	*/

	if (NULL != q) {
		free(q);
	}

	log_write(output_buffer);

out_of_memory:
	if (suppress) {
		*output_buffer = '\0';
		return 1;
	}
	return flush_output();
}


int send_missing(const char *cmd, const char *tag)
{
	return send_msg("ERROR", "ReqMissing", "ss", cmd, tag);
}


int send_msg(const char *bucket, const char *msgkey, const char *fmt, ...)
{
	va_list va;
	char *s;
	size_t len;

	va_start(va, fmt);
	s = buffer_format_v(&len, bucket, msgkey, fmt, va);

	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			len + 16);
	memcpy(output_buffer, s, len + 1);

out_of_memory:
	free(s);
	va_end(va);
	return flush_output();
}


/*
 *  function to output a result block 
 */
int send_result(int rc)
{
	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			64);
	format_message(output_buffer, "%d RESULT %d\n%d EOR\n",
			CODE_RESULT,
			(rc >= 0) ? rc : -rc,
			CODE_EOR);

out_of_memory:
	/* dump the output cache if there's data in it */
	if (!flush_output()) { return 0; }
	return send_cache_flush();
}


int send_script(const char *dir, char *cmd, int verbose) {
	char *p = cmd;
	char *nl;

	if (verbose) {
		while (NULL != ( nl = strchr(p, '\n') )) {
			char c = *nl;
			*nl = '\0';
			if (!send_log("SCRIPT", p)) { return 0; }
			*nl = c;
			p = &nl[1];
		}
		if (!send_log("SCRIPT", p)) { return 0; }
	}
	return send_msg("EXEC", "ExecStart", "ss",
			dir, platform_hostname());
}


int send_xstream_open(const char *proto, int port, uint8_t *secret, size_t secretlen) {
	int ret = 0;
	char *secrethex = NULL;


	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			64);

	secrethex = malloc((secretlen << 1) + 1);
	hex_encode(secrethex, secret, secretlen);

	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			(secretlen << 1) + strlen(proto) + 64);
	format_message(output_buffer, "%d XSTREAM %s %d %s\n",
			CODE_XSTREAM_OPEN, proto, port, secrethex);
	ret = 1;

out_of_memory:
	if (NULL != secrethex) { free(secrethex); }
	if (!flush_output() || !send_cache_flush()) { return 0; }
	return ret;
}


int send_file_info_xstream(
		int mode,
		long size,
		long mtime,
		const char *path)
{
	size_t idx;

	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			(strlen(path) << 2) + 128 );
	idx = format_message(output_buffer, "%d FILE %o %ld %ld ",
			CODE_READ_XSTREAM,
			mode,
			size,
			mtime);
	buffer_append_escaped(&output_buffer, &output_buffer_size, &idx, path);
	output_buffer[idx++] = '\n';
	output_buffer[idx] = '\0';

out_of_memory:
	return flush_output() && send_cache_flush();
}




