/*
Copyright (C)2003-2013, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

/* This is mostly just a wrapper around the strftime function.  The main
 * point of this is that the caller doesn't have to mess with "struct tm"
 * at all, and the format string is sanity checked to avoid undefined
 * behavior.  There is, of course, no guarantee that all of the format
 * characters supported here are accepted on all platforms, and since the
 * results are "undefined" rather than a graceful failure, we do not have
 * any way to detect this possibility in advance.  Check what your system
 * supports and change these if needed.
 */


#ifdef WIN32
static const char *SUPPORTED_STRFTIME_FORMATS = "aAbBcdHIjmMpSUwWxXyYzZ%";
# else /* !WIN32 */
static const char *SUPPORTED_STRFTIME_FORMATS = "aAbBcCdDehHIjmMnprRStTuUVwWxXyYZ%";
static const char *SUPPORTED_STRFTIME_E_FLAG = "cCxXyY";
static const char *SUPPORTED_STRFTIME_O_FLAG = "cCxXyYdeHImMSuUVwWy";
#endif /* WIN32/! */

/*
 * Formats a time structure using strftime after first sanity
 * checking the format string.
 *
 * The strftime function has undocumented behavior if an unsupported
 * format specifier is used in the format string.  The function uses
 * a whitelist of the format specifiers that are known to be supported
 * for Windows or POSIX systems as appropriate.  Any specifier that is
 * recognized, or a lone '%' at the end of the format string, is
 * doubled, resulting in a literal '%' in the output.
 *
 * buf - Where the formatted date will be stored; must be able to
 *    hold at least 256 bytes, and that is the maximum that will be
 *    written to it.  The output is guaranteed to be properly
 *    terminated.
 * format - the format string for strftime
 * when - the time structure to be formatted
 */
void formatDateBuf(char *buf, const char *format, struct tm *when) {
	const char *supported;
	const char *p = format;
	const char *mark;
	char *fmt = malloc((strlen(format) + 1) << 1);
	char *q = fmt;
	size_t cpy_count = 0;

#ifdef WIN32
	supported = SUPPORTED_STRFTIME_FORMATS;
#endif

	while (*p) {
		/* Copy non-format chars as-is */
		if (*p != '%') {
			*q++ = *p++;
			continue;
		}
		mark = p++;
		*q++ = '%';

		/* Lone % at end of string */
		if (*p == '\0') {
			*q++ = '%';
			break;
		}

#ifdef WIN32
		/* Check for # flag */
		if (*p == '#') {
			if (*++p == '\0') {
				cpy_count = (strlen(mark) < BFBUFSIZ) ? (strlen(mark) + 1) : BFBUFSIZ;
				strncpy(q, mark, cpy_count);
				break;
			}
		}
# else /* !WIN32 */
		/* Check for E or O flag */
		if (*p == 'E' || *p == 'O') {
			if (*++p == '\0') {
				*q++ = '%';
				*q++ = mark[1];
				break;
			}
			supported = (mark[1] == 'E')
					? SUPPORTED_STRFTIME_E_FLAG
					: SUPPORTED_STRFTIME_O_FLAG;
		} else {
			supported = SUPPORTED_STRFTIME_FORMATS;
		}
#endif /* WIN32/! */

		if (NULL != strchr(supported, *p++)) { ++mark; }
		while (mark < p) { *q++ = *mark++; }
	}
	*q = '\0';
	strftime(buf, 256, fmt, when);
	buf[255] = '\0';
	free(fmt);
}

/*
 * Like formatDateBuf, but allocates the buffer itself.
 *
 * format - the format string for strftime
 * when - the time structure to be formatted
 * return - the newly allocated buffer; it is the caller's
 *     responsibility to free() it
 */
char *formatDate(const char *format, struct tm *when) {
	char *buf = calloc(1, 256);
	formatDateBuf(buf, format, when);
	return buf;
}

/*
 * Like formatDateBuf, but accepts a timestamp rather than a
 * time structure.
 *
 * buf - as for formatDateBuf
 * format - the format string for strftime
 * when - a timestamp, as would be obtained by calling time(NULL)
 */
void formatDateBufStamp(char *buf, const char *format, time_t when) {
	struct tm *tmbuf = localtime(&when);
	formatDateBuf(buf, format, tmbuf);
}

/*
 * Like formatDateBufStamp, but allocates the buffer itself.
 *
 * format - the format string for strftime
 * when - a timestamp, as would be obtained by calling time(NULL)
 * return - the newly allocated buffer; it is the caller's
 *     responsibility to free() it
 */
char *formatDateStamp(const char *format, time_t when) {
	struct tm *tmbuf = localtime(&when);
	return formatDate(format, tmbuf);
}

/*
 * Like formatDateStamp, but uses gmtime instead of localtime to
 * convert the timestamp into a time structure, so the resulting
 * string is formatted for UTC instead of the local timezone.
 *
 * format - the format string for strftime
 * when - a timestamp, as would be obtained by calling time(NULL)
 * return - the newly allocated buffer; it is the caller's
 *     responsibility to free() it
 */
char *formatUTCDateStamp(const char *format, time_t when) {
	struct tm *tmbuf = gmtime(&when);
	return formatDate(format, tmbuf);
}

/* 
 * Like formatDateStamp, using the current time.
 *
 * format - the format string for strftime
 * return - the newly allocated buffer; it is the caller's
 *     responsibility to free() it
 */
char *expandDate(const char *format) {
	return formatDateStamp(format, time(NULL));
}

/* 
 * Like formatUTCDateStamp, using the current time.
 *
 * format - the format string for strftime
 * return - the newly allocated buffer; it is the caller's
 *     responsibility to free() it
 */
char *expandGMDate(const char *format) {
	return formatUTCDateStamp(format, time(NULL));
}

