/*
Copyright (C)2003-2011, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"
#include "xstream.h"

#ifndef ZOS
extern SOCKET xs_sock;
#endif

/* Implements "cmd write", which is used by those dot commands that need
 * to create a file on the agent.  In particular, this is used by .pop,
 * .get, .put, .rget, and .rput.
 */

int FuncWrite(void)
{
	char    *tmpname = NULL;
	char	*buf = malloc(BFBUFSIZ), *path, *p, *userdata;
	char	*timeout;
	int 	bytes, total, len, fd = -1, mode, append, clobber;
	time_t	stamp;
	struct 	utimbuf tims;
	int 	rc = 0;
	int     stream_fd = -1;
	off_t   offset = (off_t)0;
	size_t	userdata_size;

	/*  locate the path element */
	path = request_get(TAG_PATH);
	if (NULL == path) {
		send_missing(CMD_WRITE, TAG_PATH);
		free(buf);
		return 1;
	}

	stamp = 0;
	p = request_get(TAG_STAMP);
	if (NULL != p) { stamp = atol(p); }

	/* See if a data stream was provided */
	stream_fd = request_get_stream(TAG_STREAM);
#ifndef ZOS
	if (-1 == stream_fd && INVALID_SOCKET == xs_sock) {
		send_missing(CMD_WRITE, TAG_STREAM);
		free(buf);
		return 2;
	}
#endif

	p = request_get(TAG_OFFSET);
	if (NULL != p) {
		while (isdigit((int)*p)) {
			offset = offset * 10 + (*p++ - '0');
		}
		if (offset < 0) { offset = 0; }
	}

	/* locate the append element. if it wasn't supplied, set a default. */
	append = request_has(TAG_APPEND) ? O_APPEND : 0;

	/* locate the clobber element. if it wasn't supplied, set a default. */
	clobber = request_has(TAG_CLOBBER);

	/* locate the mode element. if it wasn't supplied, set a default. */
	mode = -1;
	p = request_get(TAG_MODE);
	if (NULL != p) { sscanf(p, "%o", (unsigned int *)&mode); }

	/* locate the client timeout element */
	timeout = request_get(TAG_TIMEOUT);
	if (NULL == timeout) { timeout = "0"; }

	/* locate the userdata element */
	userdata = request_get_binary(TAG_USERDATA, &userdata_size);

	StoreUserData(userdata, userdata_size);

	if (0 != platform_map_drives(defMapStr, mapStr)) {
		send_msg("MAP", "MapFail", NULL);
		FreeUserData();
		free(buf);
		/* platform_unmap_drives(0); */
		return 3;
	}

	ClobberPasswords();

	if (clearcase_view) {
		char *view = this_view();
		p = calloc(1, strlen(view) + strlen(clearcase_view) +
				strlen(path) + 3);
		format_message(p, "%s/%s/%s", view, clearcase_view, path);
		path = p;
	} else {
		path = strdup(path);
	}
	/* Note: Do not use "goto done;" to bail out until after this
	 * point, as it would try to free(path);
	 */

	/* process the clearcase view/vob start/mount functions */
	CCStartViews(atoi(timeout));
	CCMountVobs(atoi(timeout));

	/* assemble and clean the final pathname. clean the
	 * pieces so we can use them individually later. */
	CleanPath(path);
	path = env_expand_var((p = path), 0);
	free(p);

	/* clobber the target file?  */
	if (clobber) {
		platform_unlink(path);
		tmpname = request_get(TAG_STREAM);
		if (NULL != tmpname && -1 != stream_fd) {
			total = (int)lseek(stream_fd, 0, SEEK_END);

#ifdef WIN32
			/* Windows can't do a rename on a file that is currently
			 * open, so we have to close it first and reopen it if
			 * the rename fails.  How annoying...
			 */
			request_close_stream(TAG_STREAM);
			stream_fd = -1;
#endif /* WIN32 */

			/* Try to rename the file in place instead of copying it */
			if (0 == platform_rename(tmpname, path)) {
				BFTRACE("WRITE ok (rename)");
				send_msg("WRITE", "WriteOk", "is", total, path);
				tmpname = NULL;
				goto done;
			}
			activity_write(BFHERE "WRITE rename fail - %d", errno);

#ifdef WIN32
			/* On Windows, we closed the file and have to reopen it */
			stream_fd = platform_open2(tmpname, O_RDWR|O_BINARY);
			if (stream_fd < 0) {
				send_msg("WRITE", "WriteNoOpen", "s", tmpname);
				rc = 7;
				goto done;
			}
#endif /* WIN32 */
		} else {
			BFTRACE("WRITE rename N/A - no stream / tmpname");
		}
	} else {
		BFTRACE("WRITE rename N/A - !clobber");
	}

	/* open the target file */
	total = 0;
	if (!clobber && 0 == platform_access(path, 0)) {
		fd = platform_open2(path, O_WRONLY|O_BINARY|append);
		if (fd >= 0 && offset > 0) { lseek(fd, offset, SEEK_SET); }
	} else {
		fd = platform_open3(path, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY|append, 0600);
	}
	if (fd < 0) {
		send_msg("WRITE", "WriteNoOpen", "s", path);
		rc = 4;
		goto done;
	}

	if (-1 != stream_fd) { lseek(stream_fd, 0, SEEK_SET); }
	for (;;) {
		char *p = buf;
		if (-1 != stream_fd) {
			len = (int)read(stream_fd, buf, BFBUFSIZ);
		} else {
#ifndef ZOS
			len = xstream_recv_file(buf, BFBUFSIZ);
			if (len < 0) {
				int err = errno;
				send_msg("XSTREAM", "XStreamFail", "is", err, strerror(err));
				xstream_close();
				break;
			}
#endif
		}

		if (len <= 0) { break; }

		do {
			bytes = (int)write(fd, p, len);
			if (bytes <= 0) {
				BFTRACE("WRITE failed");
				send_msg("WRITE", "WriteFail", "si", path, errno);
				if (-1 != stream_fd) {
					lseek(stream_fd, 0, SEEK_END);
				} else {
#ifndef ZOS
					xstream_close();
#endif
				}
				rc = 6;
				goto done;
			}
			total += bytes;
			p = &p[bytes];
			len -= bytes;
		} while (len > 0);
	}
	close(fd);
	fd = -1;

	/* Set the file's permissions if a specific value was requested. */
	if (mode != -1) {
		if (0 == platform_chmod(path, mode)) {
			send_msg("WRITE", "WriteSetMode", "si",
					path, mode);
		} else {
			send_msg("WRITE", "WriteSetModeFail", "sii",
					path, mode, errno);
		}
	}

	/* Set the file's timestamp if a specific value was requested. */
	if (stamp > 0) {
		tims.actime = tims.modtime = stamp;
		if (0 == platform_utime(path, &tims)) {
			send_msg("WRITE", "WriteSetTime", "ss",
					path, ctime(&stamp));
		} else {
			send_msg("WRITE", "WriteSetTimeFail", "ssi",
					path, ctime(&stamp), errno);
		}
	}

	BFTRACE("WRITE ok");
	send_msg("WRITE", "WriteOk", "is", total, path);

	/* Note: See warning above about the point at which 'goto done;' is
	 * safe.
	 */
done:
#ifdef WIN32
	/* On Windows, if we stole the tmpfile from the request data,
	 * then we have to be responsible for closing and deleting it.
	 */
	if (NULL != tmpname) {
		if (stream_fd >= 0) { close(stream_fd); }
		cleanMyCruft(tmpname);
	}
#endif /* WIN32 */
	/* platform_unmap_drives(0); */
	if (fd >= 0) { close(fd); }
	FreeUserData();
	free(path);
	free(buf);
	return rc;
}

