/*
Copyright (C)2011, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"
#include "bf_rand.h"
#include "md5.h"

#undef D
#undef DEBUG_BUFFER
#ifdef DEBUG_XSTREAM
#  define D(x) x
#  define DEBUG_BUFFER(desc,data,len) debug_buffer(desc,data,len)
# else /* !XSTREAM_DEBUG */
#  define D(x) /**/
#  define DEBUG_BUFFER(desc,data,len) /**/
#endif /* XSTREAM_DEBUG/! */

/* ----------------------------------------------------------------
 *  Constants
 * ---------------------------------------------------------------- */
/* Defaults */
#define DEFAULT_TIMEOUT (20)
#define DEFAULT_SECRET_SIZE (32)
#define DEFAULT_MINIMUM_PORT (16384)
#define DEFAULT_MAXIMUM_PORT (32767)

/* Command codes */
#define XSTREAM_COMMAND_EOF (0x56A9)
#define XSTREAM_COMMAND_BYE (0xC6BF)

#define XSTREAM_TRASH_BYTES (8)

/* ----------------------------------------------------------------
 *  Internal data types
 * ---------------------------------------------------------------- */
typedef int (*xstream_protocol_init_fn)(const uint8_t *secret, size_t secretlen);
typedef int (*xstream_protocol_send_fn)(const uint8_t *buf, int buflen);
typedef int (*xstream_protocol_recv_fn)(uint8_t *buf, int buflen);
typedef int (*xstream_protocol_size_fn)(void);
typedef void (*xstream_protocol_free_fn)(void);

typedef struct _xstream_protocol {
	const char *name;
	const char *lower;
	xstream_protocol_init_fn init_fn;
	xstream_protocol_send_fn send_fn;
	xstream_protocol_recv_fn recv_fn;
	xstream_protocol_size_fn size_fn;
	xstream_protocol_free_fn free_fn;
} xstream_protocol;

/* ----------------------------------------------------------------
 *  Globals
 * ---------------------------------------------------------------- */
extern char *cb_read_buf;
extern size_t cb_read_buf_sz;
extern ssl_state_t ssl_state;
SOCKET xs_sock = INVALID_SOCKET;
static const uint8_t XSTREAM_MAGIC[] = { 0x1B, 'M', 'R', (uint8_t)0xBF, 'X', 'S', '\r', '\n' };

static uint8_t xs_secret[DEFAULT_SECRET_SIZE];
static bf_listener *xs_listener = NULL;
static md5_state_t xs_send_md5;
static md5_state_t xs_recv_md5;
static int xs_recv_left = 0;
static int xs_debug = 0;

extern int xstream_conn_timeout;
extern int xstream_recv_timeout;
extern int xstream_send_timeout;

static uint8_t *crypt_buf = NULL;
static int send_phase = 0;

/* ----------------------------------------------------------------
 *  Random data generator and cleared memory allocator
 * ---------------------------------------------------------------- */
#ifdef HAVE_SSL
#  define xstream_random_bytes(p,len) RAND_pseudo_bytes(p,len)
# else /* !HAVE_SSL */
static int xs_rand_init = 0;
static void xstream_random_bytes(uint8_t *p, int len) {
	uint8_t *stop = &p[len];
	int r;

	if (!xs_rand_init) {
		r = (time(NULL) + getpid()) ^ rand();
		srand(r);
		xs_rand_init = 1;
	}
	while (p < stop) { *p++ = (uint8_t)rand(); }
}
#endif /* HAVE_SSL/! */

static int calloc_or_clear(const char *desc, uint8_t **ptr, size_t sz) {
	if (NULL != *ptr) {
		memset(*ptr, 0, sz);
		return 1;
	}
	*ptr = calloc(1, sz);
	if (NULL != *ptr) { return 1; }
	activity_write(BFHERE "calloc_or_clear(%s,&ptr,%d): out of memory", desc, (int)sz);
	return 0;
}

/* ----------------------------------------------------------------
 *  Debug support
 * ---------------------------------------------------------------- */
#ifdef XSTREAM_DEBUG
static char hex_char[] = "0123456789ABCDEF";
static char debug_buf[BFBUFSIZ];
static void debug_buffer(const char *desc, const uint8_t *data, size_t len) {
	size_t sz = strlen(desc);
	const uint8_t *p = data;
	const uint8_t *stop = &data[(int)len];
	char *q = &debug_buf[(int)sz];

	memcpy(debug_buf, desc, sz);
	*q++ = ':';
	*q++ = ' ';
	while (p < stop) {
		*q++ = hex_char[(int)((*p >> 4) & 0xF)];
		*q++ = hex_char[(int)(*p++ & 0xF)];
	}
	*q = '\0';
	fprintf(stderr, "%s\n", debug_buf);
	fflush(stderr);
}

#endif /* XSTREAM_DEBUG */


static int check_ssl_mismatch(const char *proto) {
	if (SS_CERTIFIED != ssl_state && SS_SECURE != ssl_state) { return 1; }
	if (BFCONF_EXISTS(xstream_allow_ssl_mismatch)) { return 1; }
	send_msg("XSTREAM", "XStreamProtoMismatch", "s", proto);
	return 0;
}

/* ----------------------------------------------------------------
 *  PLAIN Protocol
 * ---------------------------------------------------------------- */

static int plain_init(const uint8_t *secret, size_t secretlen) { return check_ssl_mismatch("PLAIN"); }
static int plain_send(const uint8_t *buf, int buflen) { return bf_socket_send(xs_sock, buf, buflen, xstream_send_timeout); }
static int plain_recv(uint8_t *buf, int buflen) { return bf_socket_recv(xs_sock, buf, buflen, xstream_recv_timeout); }
static int plain_size(void) { return 0; }
static void plain_free(void) { /* no-op */ }

/* ----------------------------------------------------------------
 *  PRNG Protocol
 * ---------------------------------------------------------------- */
static bf_rand prng_encrypt_rand = 0;
static bf_rand prng_decrypt_rand = 0;
static int prng_init_impl(const uint8_t *secret, size_t secretlen) {
	DEBUG_BUFFER("prng_init: SEED ", secret, secretlen);
	bf_rand_init_bytes(&prng_encrypt_rand, secret, secretlen);
	prng_decrypt_rand = prng_encrypt_rand;
	return (NULL != crypt_buf) ? 1 : calloc_or_clear("crypt_buf", &crypt_buf, BFBUFSIZ);
}
static int prng_init(const uint8_t *secret, size_t secretlen) {
	return check_ssl_mismatch("PRNG") ? prng_init_impl(secret, secretlen) : 0;
}
static inline int prng_crypt(uint8_t *out, const uint8_t *buf, int buflen, bf_rand *prng) {
	uint8_t *q = out;
	const uint8_t *p = buf;
	const uint8_t *stop = &buf[buflen];

	while (p < stop) { *q++ = (uint8_t)(bf_rand_next_byte(prng) ^ *p++); }
	return buflen;
}
static int prng_send(const uint8_t *buf, int buflen) {
	int ret = prng_crypt(crypt_buf, buf, buflen, &prng_encrypt_rand);
	return (ret > 0) ? plain_send(crypt_buf, ret) : ret;
}
static int prng_recv(uint8_t *buf, int buflen) {
	int ret = plain_recv(crypt_buf, buflen);
	return (ret > 0) ? prng_crypt(buf, crypt_buf, ret, &prng_decrypt_rand) : ret;
}
static void prng_free(void) {
	if (NULL != crypt_buf) {
		free(crypt_buf);
		crypt_buf = NULL;
	}
}

/* ----------------------------------------------------------------
 *  AES-CBC Protocol
 * ---------------------------------------------------------------- */
#ifdef HAVE_SSL
static const EVP_CIPHER *aes_cbc = NULL;
static EVP_CIPHER_CTX aes_cbc_recv_ctx;
static EVP_CIPHER_CTX aes_cbc_send_ctx;
static int aes_cbc_ctx_init = 0;

static uint8_t *aes_cbc_recv_fragment = NULL;
static uint8_t *aes_cbc_recv_frag_ptr = NULL;
static int aes_cbc_recv_frag_len = 0;

static void report_openssl_error(const char *file, int line, const char *msg) {
	int saved_errno = GetNetworkError();
	char err_buf[1024];
	unsigned long err = ERR_get_error();

	if (NULL == activity_file) { return; }

	*err_buf = '\0';  /* Just in case... */
	ERR_error_string_n(err, err_buf, sizeof(err_buf));

	activity_write(file, line, "AES-CBC FAIL");
	activity_write(BFHERE "- msg    : %s", msg);
	activity_write(BFHERE "- err_buf: %s", err_buf);
	activity_write(BFHERE "- err    : %08lX", err);
	activity_write(BFHERE "- errno  : %d", saved_errno);
	SetNetworkError(saved_errno);
}

static void aes_cbc_free(void) {
	if (aes_cbc_ctx_init) {
		EVP_CIPHER_CTX_cleanup(&aes_cbc_send_ctx);
		EVP_CIPHER_CTX_cleanup(&aes_cbc_recv_ctx);
		aes_cbc_ctx_init = 0;
	}
	prng_free();
}

static int aes_cbc_init(const uint8_t *secret, size_t secretlen) {
	uint8_t *key = NULL;
	uint8_t *iv = NULL;
	int ret = 0;

	aes_cbc_free();
	if (!prng_init_impl(secret, secretlen)) { return 0; }

	aes_cbc = EVP_aes_128_cbc();
	if (NULL == aes_cbc) {
		report_openssl_error(BFHERE "EVP_aes_128_cbc");
		return 0;
	}

	/* Initialize Key, IV, and fragment buffer */
	key = calloc(1, aes_cbc->key_len);
	if (NULL == key) {
		BFTRACE("aes_cbc_init: out of memory allocating KEY");
		goto done;
	}
	prng_crypt(key, key, aes_cbc->key_len, &prng_encrypt_rand);
	DEBUG_BUFFER("aes_cbc_init: key", key, aes_cbc->key_len);

	if (aes_cbc->iv_len > 0) {
		iv = calloc(1, aes_cbc->iv_len);
		if (NULL == iv) {
			BFTRACE("aes_cbc_init: out of memory allocating IV");
			goto done;
		}
		prng_crypt(iv, iv, aes_cbc->iv_len, &prng_encrypt_rand);
		DEBUG_BUFFER("aes_cbc_init: iv", iv, aes_cbc->iv_len);
	}

	aes_cbc_recv_fragment = malloc(aes_cbc->block_size << 2);
	if (NULL == aes_cbc_recv_fragment) {
		BFTRACE("aes_cbc_init: out of memory allocating fragment buffer");
		goto done;
	}

	/* Sync the PRNG states */
	prng_decrypt_rand = prng_encrypt_rand;

	/* Init send/encrypt */
	EVP_CIPHER_CTX_init(&aes_cbc_send_ctx);
	EVP_CIPHER_CTX_init(&aes_cbc_recv_ctx);
	aes_cbc_ctx_init = 1;

	EVP_CIPHER_CTX_set_padding(&aes_cbc_send_ctx, 0); 
	if (!EVP_EncryptInit_ex(&aes_cbc_send_ctx, aes_cbc, NULL ,key, iv)) {
		report_openssl_error(BFHERE "EVP_EncryptInit_ex");
		goto done;
	}

	/* Init recv/decrypt */
	EVP_CIPHER_CTX_set_padding(&aes_cbc_recv_ctx, 0); 
	if (!EVP_DecryptInit_ex(&aes_cbc_recv_ctx, aes_cbc, NULL, key, iv)) {
		report_openssl_error(BFHERE "EVP_DecryptInit_ex");
		goto done;
	}

	send_phase = 0;
	BFTRACE("AES-CBC init ok");
	ret = 1;

done:
	if (NULL != key) { free(key); }
	if (NULL != iv) { free(iv); }
	if (0 == ret) { aes_cbc_free(); }
	return ret;
}

static int aes_cbc_send(const uint8_t *buf, int buflen) {
	int len = 0;
	DEBUG_BUFFER("aes_cbc_send: IN ", buf, buflen);
	if (!EVP_EncryptUpdate(&aes_cbc_send_ctx, crypt_buf, &len, buf, buflen)) {
		report_openssl_error(BFHERE "EVP_EncryptUpdate");
		D(fprintf(stderr, "aes_cbc_send: OUT: (encrypt failed)\n"));
		D(fflush(stderr));
		return -1;
	}
	send_phase = (send_phase + buflen) % aes_cbc->block_size;
	if (len <= 0) {
		D(fprintf(stderr, "aes_cbc_send: OUT: (no complete blocks)\n"));
		D(fflush(stderr));
		return 0;
	}
	DEBUG_BUFFER("aes_cbc_send: OUT", crypt_buf, len);
	return bf_socket_send(xs_sock, crypt_buf, len, xstream_send_timeout);
}

static int aes_cbc_recv_loop(const char *desc, uint8_t *buf, int buflen) {
	int got, len;
	do {
		got = bf_socket_recv(xs_sock, crypt_buf, buflen, xstream_recv_timeout);
		if (got <= 0) { return got; }
		DEBUG_BUFFER("aes_cbc_recv_loop: IN ", crypt_buf, got);
		if (!EVP_DecryptUpdate(&aes_cbc_recv_ctx, buf, &len, crypt_buf, got)) {
			report_openssl_error(BFHERE desc);
			D(fprintf(stderr, "aes_cbc_recv_loop: OUT: (decrypt failed)\n"));
			D(fflush(stderr));
			return -1;
		}
	} while (len <= 0);
	DEBUG_BUFFER("aes_cbc_recv_loop: OUT", buf, len);
	return len;
}

static int aes_cbc_recv(uint8_t *buf, int buflen) {
	int got;

	/* If we have leftovers in the fragment buffer, then return that. */
	if (aes_cbc_recv_frag_len > 0) {
		got = (buflen > aes_cbc_recv_frag_len) ? aes_cbc_recv_frag_len : buflen;
		memcpy(buf, aes_cbc_recv_frag_ptr, (size_t)got);
		aes_cbc_recv_frag_len -= got;
		aes_cbc_recv_frag_ptr = &aes_cbc_recv_frag_ptr[got];
		return got;
	}

	/* If we have a very small buffer, then use the fragment space. */
	if (buflen <= (aes_cbc->block_size << 1)) {
		got = aes_cbc_recv_loop("EVP_DecryptUpdate (fragment)",
				aes_cbc_recv_fragment, aes_cbc->block_size << 1);
		if (got <= 0) { return got; }
		if (buflen > got) { buflen = got; }
		memcpy(buf, aes_cbc_recv_fragment, (size_t)buflen);
		aes_cbc_recv_frag_ptr = &aes_cbc_recv_fragment[buflen];
		aes_cbc_recv_frag_len = got - buflen;
		return buflen;
	}

	/*
	 * Limit the maximum buffer size to ensure we don't overflow crypt_buf,
	 * and quantize to a block boundary so we can't overflow buf if a partial
	 * block was read from the socket in a previous call.  Finally, reserve
	 * at least one block to be even more certain we can't overflow.
	 */
	if (buflen > BFBUFSIZ) { buflen = BFBUFSIZ; }
	buflen -= buflen % aes_cbc->block_size;
	buflen -= aes_cbc->block_size;
	return aes_cbc_recv_loop("EVP_DecryptUpdate (blocks)", buf, buflen);
}

static int aes_cbc_size(void) { return aes_cbc->block_size; }

# else /* !HAVE_SSL */
static int aes_cbc_init(const uint8_t *secret, size_t secretlen) {
	send_msg("XSTREAM", "XStreamProtoNeedsSSL", "s", "AES-CBC");
	return 0;
}
static int aes_cbc_send(const uint8_t *buf, int buflen) {
	errno = ENOSYS;
	return -1;
}
static int aes_cbc_recv(uint8_t *buf, int buflen) {
	errno = ENOSYS;
	return -1;
}
static void aes_cbc_free(void) { /* no-op */ }
static int aes_cbc_size(void) { return 0; }
static int aes_cbc_left(void) { return 0; }
#endif /* HAVE_SSL/! */

/* ----------------------------------------------------------------
 *  Protocol resolution
 * ---------------------------------------------------------------- */
static const xstream_protocol protocol_registry[] = {
	/* Note: the first protocol listed is the default! */
	{ "PRNG", "prng", &prng_init, &prng_send, &prng_recv, &plain_size, &prng_free },
	{ "PLAIN", "plain", &plain_init, &plain_send, &plain_recv, &plain_size, &plain_free },
	{ "AES-CBC", "aes-cbc", &aes_cbc_init, &aes_cbc_send, &aes_cbc_recv, &aes_cbc_size, &aes_cbc_free },
	{ NULL }
};

static const xstream_protocol *xs_proto = &protocol_registry[0];

static const char *DEFAULT_ALIASES[] = {
	"default",
	"1",
	"y",
	"yes",
	"enabled",
	"on",
	"true",
	NULL
};

static int isDefaultAlias(const char *lower) {
	const char **p = DEFAULT_ALIASES;
	while (NULL != *p) {
		if (0 == strcmp(lower, *p)) { return 1; }
		++p;
	}
	return 0;
}

static const xstream_protocol *protocol_lookup(const char *name) {
	const xstream_protocol *proto = &protocol_registry[0];
	char *lower = NULL;

	if (NULL == name || '\0' == *name) { goto done; }
	lower = LowerString(strdup(name));
	if (isDefaultAlias(lower)) { goto done; }

	do {
		if (0 == strcmp(proto->lower, lower)) { goto done; }
		++proto;
	} while (NULL != proto->name);
	proto = NULL;

done:
	if (NULL != lower) { free(lower); }
	if (NULL == proto) {
		activity_write(BFHERE "XSTREAM protocol_lookup: protocol [%s] is not supported", name);
		errno = ENOSYS;
	}
	return proto;
}

/* ----------------------------------------------------------------
 *  Higher-level functions
 * ---------------------------------------------------------------- */

/* Initializes the I/O stream with a given protocol and either the
 * provided secret or a new one that is randomly generated.
 *
 * Returns 1 on success, or 0 if the protocol initialization fails
 */
static int xstream_init(const char *proto_name, uint8_t *secret, size_t secret_len) {
	const xstream_protocol *proto = protocol_lookup(proto_name);

	xs_debug = BFCONF_EXISTS(xstream_debug);

	if (NULL == proto) {
		send_msg("XSTREAM", "XStreamProtoUnknown", "s", proto_name);
		return 0;
	}

	if (NULL == secret) {
		xstream_random_bytes(xs_secret, DEFAULT_SECRET_SIZE);
		secret = xs_secret;
		secret_len = DEFAULT_SECRET_SIZE;
	}

	if (!(*proto->init_fn)(secret, secret_len)) {
		send_msg("XSTREAM", "XStreamProtoInitFail", "s", proto->name);
		return 0;
	}

	xs_proto = proto;
	md5_init(&xs_recv_md5);
	md5_init(&xs_send_md5);
	return 1;
}

static int xstream_send(const uint8_t *data, int len) {
	if (xs_debug) { activity_write(BFHERE "xstream_send len=%d", len); }
	return (*xs_proto->send_fn)(data, len);
}

static int xstream_recv(uint8_t *data, int len) {
	int got = (*xs_proto->recv_fn)(data, len);
	if (xs_debug) { activity_write(BFHERE "xstream_recv len=%d got=%d", len, got); }
	return got;
}

/* Fill a known-length data buffer.
 * Returns 0 on fail, 1 on success.
 */
static int xstream_recv_fill(void *data, int len) {
	uint8_t *p = data;
	int left = len;

	if (xs_debug) { activity_write(BFHERE "xstream_recv_fill: %d", len); }

	do {
		int got = xstream_recv(p, left);
		if (got <= 0) {
			activity_write(BFHERE "xstream_recv_fill: connection interrupted: %d/%d", len-left, len);
			errno = EBADF;
			return 0;
		}
		p = &p[got];
		left -= got;
	} while (left > 0);

	return 1;
}

/* Receive a 16-bit unsigned value in big-endian byte order.
 * Returns -1 if the connection fails.
 */
static int xstream_recv_short(void) {
	uint8_t buf[2];
	unsigned int x;

	if (!xstream_recv_fill(buf, 2)) { return -1; }
	x = buf[0];
	return (int)((x << 8) + buf[1]);
}

/* Process an XSTREAM_COMMAND_EOF packet, including verification of the
 * MD5 digest.  Returns 0 on success, -1 on failure.
 */
static int xstream_recv_eof() {
	uint8_t buf[16];
	md5_byte_t digest[16];

	D(fprintf(stderr, "xstream_recv_eof: reading MD5\n"));
	D(fflush(stderr));
	if (!xstream_recv_fill(buf, 16)) { return -1; }
	if (BFCONF_EXISTS(xstream_disable_md5)) { return 0; }

	md5_finish(&xs_recv_md5, digest);
	md5_init(&xs_recv_md5);

	if (BFCONF_EXISTS(xstream_disable_md5)) { return 0; }
	if (0 != memcmp(digest, buf, 16)) {
		int i;
		for (i=0; i<16; ++i) {
			if (digest[i] != 0x5A) {
				BFTRACE("xstream_recv_verify_md5: mismatch");
				SetNetworkErrorConnectionReset();
				return -1;
			}
		}

		BFTRACE("xstream_recv_verify_md5: MD5 was apparently disabled on remote system; ignoring...");
		return 0;
	}
	
	D(fprintf(stderr, "xstream_recv_eof: MD5 ok\n"));
	D(fflush(stderr));
	return 0;
}


/* Sends an XSTREAM_COMMAND_EOF packet, including the MD5 digest.
 * The sending MD5 is then reset for the next file (if any).
 * Returns 1 on success, 0 on failure.
 */
static int xstream_send_eof(void) {
	uint8_t buf[20];

	buf[0] = 0;
	buf[1] = 0;
	buf[2] = (uint8_t)(XSTREAM_COMMAND_EOF >> 8);
	buf[3] = (uint8_t)(XSTREAM_COMMAND_EOF);

	if (BFCONF_EXISTS(xstream_disable_md5)) {
		memset(&buf[4], 0x5A, sizeof(buf) - 4);
	} else {
		md5_finish(&xs_send_md5, (md5_byte_t *)&buf[4]);
		md5_init(&xs_send_md5);
	}
	if (xstream_send(buf, 20) < 0) {
		BFTRACE("xstream_send_eof: xstream_send failed");
		return 0;
	}
	return 1;
}

/* Sends an XSTREAM_COMMAND_BYE packet, which indicates a graceful closing
 * of the xstream connection.  The connection is permitted to include
 * additional information past this command for padding purposes, and
 * this must be ignored by the receiver.
 * Returns 1 on success, 0 on failure.
 */
int xstream_send_bye(void) {
	uint8_t buf[256];
	int sz = (*xs_proto->size_fn)();

	buf[0] = 0;
	buf[1] = 0;
	buf[2] = (uint8_t)(XSTREAM_COMMAND_BYE >> 8);
	buf[3] = (uint8_t)(XSTREAM_COMMAND_BYE);

	if (sz > 0) {
		sz <<= 1;
		if (sz > 252) { sz = 252; }
		xstream_random_bytes(&buf[4], sz);
		sz += 4;
	} else {
		sz = 4;
	}

	if (xstream_send(buf, sz) < 0) {
		BFTRACE("xstream_send_bye: xstream_send failed");
		return 0;
	}
	return 1;
}

/* Receives and processes a command packet.
 * Returns:
 *  0 successful EOF command
 *  >0 some other command was processed successfully and the receive should
 *    continue with the next packet
 *  -1 failure
 *  -2 graceful termination of the connection
 */
static int xstream_recv_command() {
	int command = xstream_recv_short();

	if (xs_debug) { activity_write(BFHERE "xstream_recv_command: command: %04X", command); }
	D(fprintf(stderr, "xstream_recv_command: xstream_recv_short returned %04X\n", command));
	D(fflush(stderr));

	switch (command) {
		case XSTREAM_COMMAND_EOF:
			D(fprintf(stderr, "XSTREAM_COMMAND_EOF\n"));
			D(fflush(stderr));
			return xstream_recv_eof();
		case XSTREAM_COMMAND_BYE:
			return -2;
		case -1:
			/* Connection lost; already reported in xstream_recv_fill */
			return -1;
		default:
			activity_write(BFHERE "xstream_recv_command: unknown command 0x%04X", (unsigned int)command);
			return -1;
	}
}

/* Receives the length bytes for the next packet.
 * Returns:
 *  0 Reached EOF for the current file and the MD5 was verified
 *  >0 The number of data bytes that follow in this file data packet
 *  -1 The connection terminated prematurely or the protocol was otherwise
 *		violated
 *  -2 Graceful connection close (the data source is exhausted)
 */
static int xstream_recv_get_blocklen() {
	int blocklen;

	if (INVALID_SOCKET == xs_sock) {
		SetNetworkErrorInvalidArgument();
		return -1;
	}
	if (xs_recv_left > 0) {
		D(fprintf(stderr, "xstream_recv_get_blocklen: had leftovers: %d\n", xs_recv_left));
		D(fflush(stderr));
		return xs_recv_left;
	}
	for (;;) {
		blocklen = xstream_recv_short();
		D(fprintf(stderr, "xstream_recv_get_blocklen: xstream_recv_short returned %d\n", blocklen));
		D(fflush(stderr));
		if (blocklen < 0) { return blocklen; }
		if (blocklen > 0) {
			if (blocklen > BFBUFSIZ-2) {
				activity_write(BFHERE "xstream_recv_get_block: corrupted protocol stream: bad blocklen: %d", blocklen);
				return -1;
			}
			if (xs_debug) { activity_write(BFHERE "xstream_recv_get_block: blocklen: %d", blocklen); }
			xs_recv_left = blocklen;
			return blocklen;
		}
		blocklen = xstream_recv_command();
		if (blocklen <= 0) { return blocklen; }
	}
}

/* Receive file data from the stream.
 * Returns:
 * 0 on successful EOF of the current file
 * >0 number of bytes of file data placed in the data buffer
 * -1 failure (protocol violation, premature closing of connection, etc.)
 * -2 graceful connection close (though we wouldn't normally ever ask for it)
 */
int xstream_recv_file(void *data, int len) {
	int blocklen, got;

	blocklen = xstream_recv_get_blocklen();
	D(fprintf(stderr, "blocklen=%d\n", blocklen));
	D(fflush(stderr));
	if (blocklen <= 0) { return blocklen; }
	if (len > blocklen) { len = blocklen; }

	got = xstream_recv(data, len);
	if (got <= 0) {
		xs_recv_left = 0;
		SetNetworkErrorInvalidArgument();
		return -1;
	}

	if (!BFCONF_EXISTS(xstream_disable_md5)) {
		md5_append(&xs_recv_md5, (const md5_byte_t *)data, got);
	}
	xs_recv_left = blocklen - got;
	if (!send_heartbeat()) {
		SetNetworkErrorConnectionReset();
		return -1;
	}
	return got;
}

/* Sends a file's data through the xstream connection. */
int xstream_send_file(const char *path, const struct stat *sb, int fd) {
	int bytes;

	if (INVALID_SOCKET == xs_sock) { return 0; }

	for (;;) {
		/* Note: First 2 bytes must be reserved for the block length. */
		bytes = (int)read(fd, &cb_read_buf[2], (int)cb_read_buf_sz - 2);
		if (xs_debug) { activity_write(BFHERE "xstream_send_file: read: %d", bytes); }
		D(fprintf(stderr, "xstream_send_file: read: %d\n", bytes));
		D(fflush(stderr));
		if (bytes <= 0) { break; }

		/* Fill in byte count, update the MD5, and send it. */
		cb_read_buf[0] = (uint8_t)(bytes >> 8);
		cb_read_buf[1] = (uint8_t)(bytes);
		if (!BFCONF_EXISTS(xstream_disable_md5)) {
			md5_append(&xs_send_md5, (const md5_byte_t *)&cb_read_buf[2], bytes);
		}
		bytes = xstream_send((const uint8_t *)cb_read_buf, bytes + 2);
		if (bytes < 0) { return 0; }
	}
	return xstream_send_eof();
}

static void xstream_adjust_ports(const char *bind_addr, int *minport, int *maxport, int *firstport) {
	uint16_t r;

	/* We have to have either a bind address or a port number to use the
	 * getaddrinfo function.  If we don't have either one, then we will
	 * pretend that a very large port range was specified.
	 */
	if (NULL == bind_addr && 0 == *minport) {
		*minport = DEFAULT_MINIMUM_PORT;
		*maxport = DEFAULT_MAXIMUM_PORT;
	}

	/* If we were not asked to randomize the port numbers, then just start
	 * with the minimum.
	 */
	if (!BFCONF_EXISTS(xstream_listen_randomize)) {
		*firstport = *minport;
		return;
	}

	/* We can't randomize the port numbers without a range. */
	if (0 == *minport) {
		*minport = DEFAULT_MINIMUM_PORT;
		*maxport = DEFAULT_MAXIMUM_PORT;
	}

	/* Pick the first port at random */
	xstream_random_bytes((uint8_t *)&r, sizeof(r));
	*firstport = ((int)r) % (*maxport - *minport + 1) + *minport;
}

static void xstream_listen_range(const char *bind_addr, int *minport, int *maxport) {
	char *port_range = bfconf_get("xstream_listen_range", NULL);
	char *pos;
	int min, max;

	*minport = *maxport = 0;

	if (NULL == port_range) { return; }

	pos = strchr(port_range, '-');
	if (NULL == pos) {
		min = atoi(port_range);
		if (min <= 0 || min > 65535) {
			activity_write(BFHERE "xstream_listen_range: Illegal value [%s]; ignored...", port_range);
		} else {
			activity_write(BFHERE "xstream_listen_range: Restricted to port [%d]", min);
			*maxport = *minport = min;
		}
	} else {
		*pos++ = '\0';
		min = atoi(port_range);
		max = atoi(pos);
		if (min <= 0 || min > 65535) {
			activity_write(BFHERE "xstream_listen_range: Illegal minimum value [%s]; setting ignored...", port_range);
		} else if (max <= 0 || max > 65535) {
			activity_write(BFHERE "xstream_listen_range: Illegal maximum value [%s]; setting ignored...", pos);
		} else {
			if (min < max) {
				activity_write(BFHERE "xstream_listen_range: %d-%d", min, max);
				*minport = min;
				*maxport = max;
			} else if (min > max) {
				activity_write(BFHERE "xstream_listen_range: %d-%d", max, min);
				*minport = max;
				*maxport = min;
			} else {
				activity_write(BFHERE "xstream_listen_range: Forced to use port [%d]", min);
				*minport = *maxport = min;
			}
		}
	}

	free(port_range);
	return;
}

int xstream_close(void) {
	int did_something = 0;
	if (NULL != xs_listener) {
		bf_listener_destroy(xs_listener);
		xs_listener = NULL;
		did_something = 1;
	}
	if (INVALID_SOCKET != xs_sock) {
		platform_closesocket(xs_sock);
		xs_sock = INVALID_SOCKET;
		did_something = 1;
	}
#ifdef HAVE_SSL
	/* TODO - Release AES-CBC resources */
#endif /* HAVE_SSL */
	return did_something;
}

int xstream_offer(const char *proto) {
	char *bind_addr = NULL;
	int minport = 0;
	int maxport = 0;
	int firstport, port;
	uint8_t r[XSTREAM_TRASH_BYTES];
	SOCKET sock;

	xstream_close();
	if (!xstream_init(proto, NULL, 0)) { return 0; }

	bind_addr = bfconf_get("xstream_bind", NULL);
	xstream_listen_range(bind_addr, &minport, &maxport);
	xstream_adjust_ports(bind_addr, &minport, &maxport, &firstport);
	port = firstport;
	do {
		xs_listener = bf_listener_new(bind_addr, port, 1);
		if (NULL != xs_listener) {
			activity_write(BFHERE "xstream_offer: listening on %s/%d",
					(NULL != bind_addr) ? bind_addr : "*",
					xs_listener->port);
			break;
		}

		/* Try the next port, wrapping around if necessary. */
		port = (port != maxport) ? (port + 1) : minport;
	} while (port != firstport);

	if (NULL != bind_addr) { free(bind_addr); }
	if (NULL == xs_listener) {
		BFTRACE("xstream_offer: unable to open a listening port");
		return 0;
	}

	DEBUG_BUFFER("xstream_offer: SEED", xs_secret, DEFAULT_SECRET_SIZE);
	send_xstream_open(
			xs_proto->name,
			xs_listener->port, 
			xs_secret,
			DEFAULT_SECRET_SIZE);
	sock = bf_listener_accept(xs_listener, xstream_conn_timeout);
	if (INVALID_SOCKET == sock) {
		int err = errno;
		send_msg("XSTREAM", "XStreamFail", "is", err, strerror(err));
		xstream_close();
		return 0;
	}

	/* Closes the listener only because we haven't set xs_sock yet. */
	xstream_close();
	xs_sock = sock;

	/* Send the handshake (random bytes followed by magic) */
	xstream_random_bytes(r, sizeof(r));
	if (xstream_send(r, sizeof(r)) < 0) {
		BFTRACE("XSTREAM error send rand");
		goto error_handshake;
	}
	if (xstream_send(XSTREAM_MAGIC, sizeof(XSTREAM_MAGIC)) < 0) {
		BFTRACE("XSTREAM error send magic");
		goto error_handshake;
	}
	return 1;

error_handshake:
	send_msg("XSTREAM", "XStreamHandshakeFail", NULL);
	xstream_close();
	return 0;
}

static int xstream_init_client(char **host, int *port) {
	char *xs = request_get(TAG_XSTREAM);
	char *protocol, *secrethex, *serv;
	char *sptr = NULL;
	uint8_t secret[32];
	int secretlen;

	if (NULL == xs) {
		send_missing(CMD_XSTREAM, TAG_XSTREAM);
		return 1;
	}

	/* xstream protocol port secret hostname */

	protocol = bf_next_token(xs, " ", &sptr);
	if (NULL == protocol) {
		BFTRACE("xstream_init_client: missing protocol");
		goto usage;
	}

	serv = bf_next_token(xs, " ", &sptr);
	if (NULL == serv || *serv < '0' || *serv > '9') {
		BFTRACE("xstream_init_client: missing serv");
		goto usage;
	}
	*port = atoi(serv);
	if (*port < 1 || *port > 65535) {
		activity_write(BFHERE "xstream_init_client: bad port: %s", serv);
		goto usage;
	}


	secrethex = bf_next_token(xs, " ", &sptr);
	if (NULL == secrethex) {
		BFTRACE("xstream_init_client: missing secret");
		goto usage;
	}
	secretlen = (int)strlen(secrethex);
	if (secretlen <= 0 || 0 != (secretlen & 1) || secretlen > 64) {
		activity_write(BFHERE "xstream_init_client: invalid secret length: %d", secretlen);
		goto usage;
	}
	secretlen = hex_decode(secret, secrethex);

	*host = bf_next_token(xs, " ", &sptr);
	if (NULL == *host) {
		BFTRACE("xstream_init_client: missing host");
		goto usage;
	}

	if (!xstream_init(protocol, secret, secretlen)) { return 3; }
	return 0;

usage:
	send_msg("XSTREAM", "XStreamUsage", NULL);
	return 2;
}


/* ----------------------------------------------------------------
 *  cmd xstream -- client-side control of an xstream connection
 * ---------------------------------------------------------------- 
 */
int FuncXStream(void)
{
	char *host;
	int port;
	int ret = xstream_close();
	uint8_t magic[XSTREAM_TRASH_BYTES + sizeof(XSTREAM_MAGIC)];

	if (ret > 0) { send_msg("XSTREAM", "XStreamClosed", NULL); }
	if (!request_has(TAG_XSTREAM)) { return 0; }

	ret = xstream_init_client(&host, &port);
	if (0 != ret) { return ret; }

	xs_sock = bf_socket_connect(host, port, xstream_conn_timeout);
	if (INVALID_SOCKET == xs_sock) {
		int err = errno;
		send_msg("XSTREAM", "XStreamFailConnect", "siis", host, port, err, strerror(err));
		return 1;
	}

	if (!xstream_recv_fill(magic, sizeof(magic))) {
		send_msg("XSTREAM", "XStreamFailCorrupted", "si", host, port);
		return 4;
	}

	if (0 != memcmp(&magic[XSTREAM_TRASH_BYTES], XSTREAM_MAGIC, sizeof(XSTREAM_MAGIC))) {
		char hex[(sizeof(magic) << 1) + 1];
		hex_encode(hex, magic, sizeof(magic));
		BFTRACE(hex);
		send_msg("XSTREAM", "XStreamFailCorrupted", "si", host, port);
		return 5;
	}

	send_msg("XSTREAM", "XStreamConnected", "si", host, port);
	return 0;
}

